{
        Program Name : ExtentSt
        Package Name : LB
        Purpose      : Extent store
        Institution  : Conroy & Conroy Co.
        Date Written : 16-Feb-2003
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2003 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          The TSExtent_Store class provides an interface to a shared store (that
        is, a managed store that contains more than just the extent store).  The
        store is logically contiguous but not necessarily physically so.  It
        uses a linked list of extent records on the shared store to keep track
        of all of the pieces of the data for the extent store.
}

unit ExtentSt ;

interface

uses { C&C... }
     Compatib, { int64 }
     _DebugIn, { PSDebug_Interface }
     MStores, { TSStore }
     _Stores, { TSStore }
     _UEHDefs ; { TUEC }

const Extent_Store_Facility  = 66 ;
      Extent_StoreErr_Cannot_Allocate_On_Store = 1 ;
      Extent_StoreErr_RW_Size_Invalid = 2 ;
      Extent_StoreErr_Access_Violation = 3 ;
      Extent_StoreErr_No_Space_On_Store = 4 ;

type TSExtent_Store_Header = record
                                Root : TStore_Address64 ;
                                Size : TStore_Address64 ; { Data size }
                                Extent_Size : longint ;
                                Extent_Record_Size : longint ;
                                Flags : longint ; { and 1 = contiguous }
                            end ;

     TAllocation_Record = array[ 0..4095 ] of TStore_Address64 ;
     PAllocation_Record = ^TAllocation_Record ;

     PSExtent_Store = ^TSExtent_Store ;
     TSExtent_Store = object( TSStore64 )
                         private { Internal data... }
                             _Contiguous : boolean ; { True if the store is physically contiguous - using no extents }
                             Size : TStore_Address64 ; { Allocated size of extent store's data, in bytes }
                             Last_Error : TUEC ;
                             Extent_Record_Size : longint ; { Size of extent records, in bytes (minimum 2 * sizeof(TStore_Address64)) }
                             Extent_Size : longint ; { How much to extend store by, in bytes - clustersize }

                         protected { Internal utility routines... }
                             procedure Set_Error( Code : longint ) ;
                             procedure Update_Header ;

                         public { API... }
                             Shared_Store : PSManaged_Store ; { Store where extent store is persisted }
                             Root : TStore_Address64 ; { Header address }
                             Data_Root : TStore_Address64 ; { Shared store address of first extent record (or of data if contiguous) }
                             Allow_Read : boolean ; { True if reading is allowed }
                             Allow_Write : boolean ; { True if writing/changing is allowed }

                             function Read_Data( var Data ;
                                 Address, _Size : TStore_Address64 ;
                                 var UEC : TUEC ) : TStore_Address64 ; virtual ;
                              { Read data from specified Address in store to
                                Data.  Read _Size bytes.  Returns success of
                                operation in UEC.  Function returns the number
                                of bytes actually read. }

                              function Write_Data( var Data ;
                                  Address, _Size : TStore_Address64 ;
                                  var UEC : TUEC ) : TStore_Address64 ; virtual ;
                              { Write data from Data to specified Address in
                                the store.  Write _Size bytes.  Returns success
                                of operation in UEC. Function returns the number
                                of bytes actually written. }

                              function Max_Storage : TStore_Size64 ; virtual ;

                              function Min_Storage : TStore_Address64 ; virtual ;

                              function Extend( Amount : TStore_Address64 ) : TStore_Address64 ;
                                  virtual ;
                              { Requests extension of store to the amount
                                specified.  Returns the amount actually added to
                                the store.  If Amount is 0, function returns 0
                                if store is extendable, and -1 if it is not. }

                              function Read_Only : boolean ; virtual ;

                              function Write_Only : boolean ; virtual ;

                              procedure Initialize ; virtual ;

                              procedure Terminate ; virtual ;

                              function Debugger : PSDebug_Interface ; virtual ;

                              function Contiguous_Store : boolean ; virtual ;

                              procedure Set_Max_Storage( Value : TStore_Address64 ;
                                  var Res : TUEC ) ; virtual ;
                              { Sets the maximum (current) size of the store, in
                                bytes.  In other words, this is the new highest
                                valid address (plus 1).  If the store is not
                                resizable, the call fails and an error code is
                                returned in Res.  This should only be used to
                                shrink the size of the store, not to extend it.
                                To extend, use the Extend method. }

                              function Extended_Size : TStore_Address64 ;
                                  virtual ;

                              procedure Define( _Root : TStore_Address64 ;
                                  SS : PSManaged_Store ) ;
                              { Points extent store to existing extent store on
                                the shared store. }

                              procedure Set_Contiguous( Value : boolean ) ;

                              function Get_Contiguous : boolean ;

                              procedure Create_New( ES, ERS, Flags : longint ;
                                  SS : PSManaged_Store ) ; virtual ;
                              { Create a new store with exten size ES, extent
                                record size ERS, and flags (1 = contiguous). }
                     end ; { TSExtent_Store }

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses { C&C... }
{$IFDEF Test}
     _App,
     App,
     dialogs,
{$ENDIF}
     _Types, { Small_String }
     CommonUt, { Boolean_To_String }
     DebugInt, { TSText_Debugger }
     Num1s ; { Num1 }

type PSExtent_Store_Debugger = ^TSExtent_Store_Debugger ;
     TSExtent_Store_Debugger = object( TSText_Debugger )
                          private
                              _Instance : PSExtent_Store ;
                              S : Small_String ;

                          public
                              constructor Init( I : PSExtent_Store ) ;

                              function Count : longint ; virtual ;

                              function Child( Ordinal : longint ) : PSDebug_Interface ;
                                  virtual ;
                      end ;

constructor TSExtent_Store_Debugger.Init( I : PSExtent_Store ) ;

begin
    inherited Init ;

    _Instance := I ;
end ;


function TSExtent_Store_Debugger.Count : longint ;

begin
    Count := 10 ;
end ;


function TSExtent_Store_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

var R : PSText_Debugger ;

begin
    new( R, Init ) ;
    case Ordinal of
        0 : S := '_Contiguous = ' + Boolean_To_String( _Instance._Contiguous ) ;
        1 : S := 'Size = ' + Num1( _Instance.Size ) ;
        2 : S := 'Last_Error = Facility: ' + Num1( _Instance.Last_Error.Facility ) +
                ', Code: ' + Num1( _Instance.Last_Error.Code ) ;
        3 : S := 'Extent_Record_Size = ' + Num1( _Instance.Extent_Record_Size ) ;
        4 : S := 'Extent_Size = ' + Num1( _Instance.Extent_Size ) ;
        5 : begin
                if( _Instance.Shared_Store = nil ) then
                begin
                    S := 'Shared_Store = nil' ;
                end else
                begin
                    dispose( R, done ) ;
                    R := PSText_Debugger( _Instance.Shared_Store^.Debugger ) ;
                    if( R = nil ) then
                    begin
                        new( R, Init ) ;
                    end ;
                    S := 'Shared_Store' ;
                end ;
            end ;
        6 : S := 'Root = ' + Num1( _Instance.Root ) ;
        7 : S := 'Data_Root = ' + Num1( _Instance.Data_Root ) ;
        8 : S := 'Allow_Read = ' + Boolean_To_String( _Instance.Allow_Read ) ;
        9 : S := 'Allow_Write = ' + Boolean_To_String( _Instance.Allow_Write ) ;
    end ;
    S := S + #0 ;
    R^.Set_Title( @S[ 1 ] ) ;
    Child := PSDebug_Interface( R ) ;
end ; { TSExtent_Store_Debugger.Child }


{ TSExtent_Store methods... }

{ Internal utility routines... }

procedure TSExtent_Store.Set_Error( Code : longint ) ;

begin
    Last_Error.Facility := Extent_Store_Facility ;
    Last_Error.Code := Code ;
end ;


procedure TSExtent_Store.Update_Header ;

var Header : TSExtent_Store_Header ;
    UEC : TUEC ;

begin
    if( Shared_Store <> nil ) then
    begin
        if( Data_Root = 0 ) then
        begin
            Data_Root := Shared_Store.Allocate( sizeof( TSExtent_Store_Header ) ) ;
        end ;
        if( Data_Root = 0 ) then
        begin
            exit ;
        end ;
        Header.Root := Data_Root ;
        Header.Size := Size ;
        Header.Extent_Size := Extent_Size ;
        Header.Extent_Record_Size := Extent_Record_Size ;
        if( _Contiguous ) then
        begin
            Header.Flags := 1 ;
        end else
        begin
            Header.Flags := 0 ;
        end ;
        Shared_Store^.Write_Data( Header, Root, sizeof( Header ), UEC )
    end ;
end ;


{ API... }

function TSExtent_Store.Read_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUEC ) : TStore_Address64 ;

var Allocation : PAllocation_Record ;
    Base : longint ;
    Index : integer ;
    Res : TStore_Address64 ;
    Temp : TStore_Address64 ;

begin
    if( not Allow_Read ) then
    begin
        UEC.Facility := Extent_Store_Facility ;
        UEC.Code := Extent_StoreErr_Access_Violation ;
        Read_Data := 0 ;
        exit ;
    end ;

    if( Address + _Size > Size ) then
    begin
        _Size := Size - Address - 1 ;
    end ;
    if( _Size <= 0 ) then
    begin
        UEC.Facility := Extent_Store_Facility ;
        UEC.Code := Extent_StoreErr_RW_Size_Invalid ;
        Read_Data := 0 ;
        exit ;
    end ;

    if( _Contiguous ) then
    begin
        Read_Data := Shared_Store^.Read_Data( Data, Data_Root + Address, _Size, UEC ) ;
        Last_Error := UEC ;
        exit ;
    end ;

    { Position to first byte to read... }
    getmem( Allocation, Extent_Record_Size ) ;
    Shared_Store^.Read_Data( Allocation^, Data_Root, Extent_Record_Size, UEC ) ;
    if( UEC.Code <> 0 ) then
    begin
        Last_Error := UEC ;
        freemem( Allocation, Extent_Record_Size ) ;
        Read_Data := 0 ;
        exit ;
    end ;
    Index := 1 ;
    while( Address > Extent_Size ) do
    begin
        Address := Address - Extent_Size ;
        inc( Index ) ;
        if( Index > Extent_Record_Size div sizeof( TStore_Address64 ) ) then
        begin
            Shared_Store^.Read_Data( Allocation^, Allocation^[ 0 ], Extent_Record_Size, UEC ) ;
            if( UEC.Code <> 0 ) then
            begin
                Last_Error := UEC ;
                freemem( Allocation, Extent_Record_Size ) ;
                Read_Data := 0 ;
                exit ;
            end ;
            Index := 1 ;
        end ;
    end ;

    { Read the data... }
    Base := 0 ;
    Res := 0 ;
    while( _Size > 0 ) do
    begin
        Temp := _Size ;
        if( Temp > Extent_Size ) then
        begin
            Temp := Extent_Size ;
        end ;
        _Size := _Size - Temp ;
        Shared_Store^.Read_Data( PChar( Data )[ Base ], Allocation^[ Index ], Temp, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Last_Error := UEC ;
            freemem( Allocation, Extent_Record_Size ) ;
            Read_Data := Res ;
            exit ;
        end ;
        Base := Base + Temp ;
        Res := Res + Temp ;
        inc( Index ) ;
        if( Index > Extent_Record_Size div sizeof( TStore_Address64 ) ) then
        begin
            Shared_Store^.Read_Data( Allocation^, Allocation^[ 0 ], Extent_Record_Size, UEC ) ;
            if( UEC.Code <> 0 ) then
            begin
                Last_Error := UEC ;
                freemem( Allocation, Extent_Record_Size ) ;
                Read_Data := Res ;
                exit ;
            end ;
            Index := 1 ;
        end ;
    end ; { while( _Size > 0 )  }

    { Clean up... }
    freemem( Allocation, Extent_Record_Size ) ;
    Read_Data := Res ;
end ; { TSExtent_Store.Read_Data }


function TSExtent_Store.Write_Data( var Data ; Address, _Size : TStore_Address64 ;
    var UEC : TUEC ) : TStore_Address64 ;

var Allocation : PAllocation_Record ;
    Base : longint ;
    Index : integer ;
    Res : TStore_Address64 ;
    Temp : TStore_Address64 ;

begin
    if( not Allow_Write ) then
    begin
        UEC.Facility := Extent_Store_Facility ;
        UEC.Code := Extent_StoreErr_Access_Violation ;
        Write_Data := 0 ;
        exit ;
    end ;

    if( Address + _Size > Size ) then
    begin
        _Size := Size - Address - 1 ;
    end ;
    if( _Size <= 0 ) then
    begin
        UEC.Facility := Extent_Store_Facility ;
        UEC.Code := Extent_StoreErr_RW_Size_Invalid ;
        Write_Data := 0 ;
        exit ;
    end ;

    if( _Contiguous ) then
    begin
        Write_Data := Shared_Store^.Write_Data( Data, Data_Root + Address, _Size, UEC ) ;
        Last_Error := UEC ;
        exit ;
    end ;

    { Position to first byte to write... }
    getmem( Allocation, Extent_Record_Size ) ;
    Shared_Store^.Read_Data( Allocation^, Data_Root, Extent_Record_Size, UEC ) ;
    if( UEC.Code <> 0 ) then
    begin
        Last_Error := UEC ;
        freemem( Allocation, Extent_Record_Size ) ;
        Write_Data := 0 ;
        exit ;
    end ;
    Index := 1 ;
    while( Address > Extent_Size ) do
    begin
        Address := Address - Extent_Size ;
        inc( Index ) ;
        if( Index > Extent_Record_Size div sizeof( TStore_Address64 ) ) then
        begin
            Shared_Store^.Read_Data( Allocation^, Allocation^[ 0 ], Extent_Record_Size, UEC ) ;
            if( UEC.Code <> 0 ) then
            begin
                Last_Error := UEC ;
                freemem( Allocation, Extent_Record_Size ) ;
                Write_Data := 0 ;
                exit ;
            end ;
            Index := 1 ;
        end ;
    end ;

    { Write the data... }
    Base := 0 ;
    Res := 0 ;
    while( _Size > 0 ) do
    begin
        Temp := _Size ;
        if( Temp > Extent_Size ) then
        begin
            Temp := Extent_Size ;
        end ;
        _Size := _Size - Temp ;
        Shared_Store^.Write_Data( PChar( Data )[ Base ], Allocation^[ Index ], Temp, UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            Last_Error := UEC ;
            freemem( Allocation, Extent_Record_Size ) ;
            Write_Data := Res ;
            exit ;
        end ;
        Base := Base + Temp ;
        Res := Res + Temp ;
        inc( Index ) ;
        if( Index > Extent_Record_Size div sizeof( TStore_Address64 ) ) then
        begin
            Shared_Store^.Read_Data( Allocation^, Allocation^[ 0 ], Extent_Record_Size, UEC ) ;
            if( UEC.Code <> 0 ) then
            begin
                Last_Error := UEC ;
                freemem( Allocation, Extent_Record_Size ) ;
                Write_Data := Res ;
                exit ;
            end ;
            Index := 1 ;
        end ;
    end ; { while( _Size > 0 )  }

    { Clean up... }
    freemem( Allocation, Extent_Record_Size ) ;
    Write_Data := Res ;
end ; { TSExtent_Store.Write_Data }


function TSExtent_Store.Max_Storage : TStore_Size64 ;

begin
    Max_Storage := Size ;
end ;


function TSExtent_Store.Min_Storage : TStore_Address64 ;

begin
    Min_Storage := Shared_Store.Min_Storage ;
end ;


function TSExtent_Store.Extend( Amount : TStore_Address64 ) : TStore_Address64 ;

var Old : TStore_Address ;
    UEC : TUEC ;

begin
    if( Amount = 0 ) then { Query whether store can extend }
    begin
        if( ( Shared_Store = nil ) or ( not Allow_Write ) ) then
        begin
            Extend := -1 ;
        end else
        begin
            Extend := Shared_Store.Extend( 0 ) ;
        end ;
        exit ;
    end ;

    { Extend store to Amount bytes... }
    Old := Size ;
    Set_Max_Storage( Amount, UEC ) ;
    if( UEC.Code = 0 ) then
    begin
        Extend := Amount ;
    end else
    begin
        Extend := 0 ;
    end ;
end ;


function TSExtent_Store.Read_Only : boolean ;

begin
    Read_Only := not Allow_Read ;
end ;


function TSExtent_Store.Write_Only : boolean ;

begin
    Write_Only := not Allow_Write ;
end ;


procedure TSExtent_Store.Initialize ;

begin
    inherited Initialize ;

    _Contiguous := False ;
    Shared_Store := nil ;
    Root := 0 ;
    Data_Root := 0 ;
    Extent_Record_Size := 2 * sizeof( TStore_Address64 ) ; { One extent per allocation record, plus Next pointer }
    Extent_Size := 512 ;
    Allow_Read := True ;
    Allow_Write := True ;
end ;


procedure TSExtent_Store.Terminate ;

begin
    Shared_Store := nil ;
    Root := 0 ;
    Data_Root := 0 ;
    Extent_Size := 0 ;
    Allow_Read := False ;
    Allow_Write := False ;
end ;


function TSExtent_Store.Debugger : PSDebug_Interface ;

begin
    Debugger := new( PSExtent_Store_Debugger, Init( @self ) ) ;
end ;


function TSExtent_Store.Contiguous_Store : boolean ;

begin
    Contiguous_Store := True ; { Logically contiguous }
end ;


procedure TSExtent_Store.Set_Max_Storage( Value : TStore_Address64 ;
    var Res : TUEC ) ;

var Allocation : PAllocation_Record ;
    Current : TStore_Address64 ;
    Index : integer ;
    New_Size, Old_Size, Temp : TStore_Address64 ;
    Deallocations : array[ 0..4095 ] of TStore_Address64 ;
    D_Index : integer ; { Last used element in Deallocations }
    { List of extents to deallocate.  We accumulate them in case we cannot
      update the last allocation record. }

begin
    if( not Allow_Write ) then
    begin
        Res.Facility := Extent_Store_Facility ;
        Res.Code := Extent_StoreErr_Access_Violation ;
        exit ;
    end ;

    fillchar( Res, sizeof( Res ), 0 ) ;

    if( _Contiguous ) then
    begin
        Temp := Shared_Store^.Reallocate( Data_Root, Size, Value ) ;
        if( Temp = 0 ) then { Failure to reallocate }
        begin
            Res.Facility := Extent_Store_Facility ;
            Res.Code := Extent_StoreErr_Cannot_Allocate_On_Store ;
        end else
        begin
            Data_Root := Temp ;
            Size := Value ;
            Update_Header ;
        end ;
        exit ;
    end ;

    { Determine rounded sizes... }
    Old_Size := Size + Extent_Size - 1 ;
    Old_Size := ( Old_Size div Extent_Size ) * Extent_Size ;
    New_Size := Value + Extent_Size - 1 ;
    New_Size := ( New_Size div Extent_Size ) * Extent_Size ;

    { Do the resize... }
    if( Value < Size ) then { Shrink store size }
    begin
        Temp := Data_Root ;
        Size := Value ;
        if( Size = 0 ) then
        begin
            Data_Root := 0 ;
        end ;
        Update_Header ;
        if( Old_Size <> New_Size ) then
        begin
            { Position to last extent to keep... }
            getmem( Allocation, Extent_Record_Size ) ;
            Shared_Store^.Read_Data( Allocation^, Temp, Extent_Record_Size, Res ) ;
            if( Res.Code <> 0 ) then
            begin
                Last_Error := Res ;
                freemem( Allocation, Extent_Record_Size ) ;
                exit ;
            end ;
            Current := Temp ;
            Index := 1 ;
            while( Old_Size > Extent_Size ) do
            begin
                Old_Size := Old_Size - Extent_Size ;
                inc( Index ) ;
                if( Index > Extent_Record_Size div sizeof( TStore_Address64 ) ) then
                begin
                    Current := Allocation^[ 0 ] ;
                    Shared_Store^.Read_Data( Allocation^, Current, Extent_Record_Size, Res ) ;
                    if( Res.Code <> 0 ) then
                    begin
                        Last_Error := Res ;
                        freemem( Allocation, Extent_Record_Size ) ;
                        exit ;
                    end ;
                    Index := 1 ;
                end ;
            end ;
            { At this point, Allocation contains the (new) last Allocation
              record, and Index is the (new) last Index in that record. }

            { Truncate... }
            fillchar( Deallocations, Extent_Record_Size, 0 ) ;
            D_Index := -1 ;
            while( Allocation^[ Index ] <> 0 ) do { Until we run out of extents }
            begin
                inc( Index ) ;
                if( Index > Extent_Record_Size div sizeof( TStore_Address64 ) ) then
                begin
                    if( ( Current <> 0 ) and ( Allocation^[ 1 ] = 0 ) ) then { Entire allocation record is empty }
                    begin
                        Shared_Store^.Deallocate( Current, Extent_Record_Size ) ;
                    end else
                    begin
                        Allocation^[ 0 ] := 0 ;
                        Shared_Store^.Write_Data( Allocation^, Current, Extent_Record_Size, Res ) ;
                        if( Res.Code <> 0 ) then
                        begin
                            Last_Error := Res ;
                            freemem( Allocation, Extent_Record_Size ) ;
                            exit ;
                        end ;
                    end ;
                    while( D_Index >= 0 ) do { Now free the extents }
                    begin
                        Shared_Store^.Deallocate( Deallocations[ D_Index ], Extent_Size ) ;
                        dec( D_Index ) ;
                    end ;
                    fillchar( Deallocations, Extent_Record_Size, 0 ) ;
                    D_Index := -1 ;

                    { Move to next allocation record... }
                    Current := Allocation^[ 0 ] ;
                    Shared_Store^.Read_Data( Allocation^, Current, Extent_Record_Size, Res ) ;
                    if( Res.Code <> 0 ) then
                    begin
                        Last_Error := Res ;
                        freemem( Allocation, Extent_Record_Size ) ;
                        exit ;
                    end ;
                    Index := 1 ;
                end ;
                if( Allocation^[ Index ] = 0 ) then
                begin
                    break ; { We are done when we hit a null extent since allocation records are not sparse }
                end ;
                inc( D_Index ) ;
                Deallocations[ D_Index ] := Allocation^[ Index ] ;
            end ; { while( Allocation^[ Index ] <> 0 ) }

            if( ( Current <> 0 ) and ( Allocation^[ 1 ] = 0 ) ) then { Entire allocation record is empty }
            begin
                Shared_Store^.Deallocate( Current, Extent_Record_Size ) ;
            end else
            begin
                Allocation^[ 0 ] := 0 ;
                Shared_Store^.Write_Data( Allocation^, Current, Extent_Record_Size, Res ) ;
                if( Res.Code <> 0 ) then
                begin
                    Last_Error := Res ;
                    freemem( Allocation, Extent_Record_Size ) ;
                    exit ;
                end ;
            end ;
            while( D_Index >= 0 ) do { Now free the extents }
            begin
                Shared_Store^.Deallocate( Deallocations[ D_Index ], Extent_Size ) ;
                dec( D_Index ) ;
            end ;

            { Clean up... }
            freemem( Allocation, Extent_Record_Size ) ;
        end ; { if( Old_Size <> New_Size ) }
    end else
    begin { Extend store size }
        if( Old_Size <> New_Size ) then
        begin
            { Handle empty store... }
            if( Old_SIze = 0 ) then
            begin
                { Create new data record... }
                Temp := Shared_Store^.Allocate( Extent_Size ) ;
                if( Temp = 0 ) then { Allocation failed }
                begin
                    Set_Error( Extent_StoreErr_No_Space_On_Store ) ;
                    Res := Last_Error ;
                    exit ;
                end ;
                Data_Root := Temp ;
            end ;

            { Position to last extent... }
            getmem( Allocation, Extent_Record_Size ) ;
            Shared_Store^.Read_Data( Allocation^, Data_Root, Extent_Record_Size, Res ) ;
            if( Res.Code <> 0 ) then
            begin
                Last_Error := Res ;
                freemem( Allocation, Extent_Record_Size ) ;
                exit ;
            end ;
            Current := Data_Root ;
            while( Allocation^[ 0 ] <> 0 ) do
            begin
                Current := Allocation^[ 0 ] ;
                Shared_Store^.Read_Data( Allocation^, Current, Extent_Record_Size, Res ) ;
                if( Res.Code <> 0 ) then
                begin
                    Last_Error := Res ;
                    freemem( Allocation, Extent_Record_Size ) ;
                    exit ;
                end ;
            end ;

            { Position to last extent in this allocation record... }
            Index := 1 ;
            while( Allocation^[ Index ] <> 0 ) do
            begin
                inc( Index ) ;
                if( Index > Extent_Record_Size div sizeof( TStore_Address64 ) ) then
                begin { This means that the last allocation record is exactly full }
                    { Create new allocation record }
                    Temp := Shared_Store^.Allocate( Extent_Record_Size ) ;
                    if( Temp = 0 ) then { Allocation failed }
                    begin
                        Set_Error( Extent_StoreErr_No_Space_On_Store ) ;
                        Res := Last_Error ;
                        freemem( Allocation, Extent_Record_Size ) ;
                        exit ;
                    end ;
                    Shared_Store^.Fill( Temp, Extent_Record_Size, 0 ) ;

                    { Update previous allocation record... }
                    Allocation^[ 0 ] := Temp ;
                    Shared_Store^.Write_Data( Allocation^, Current, Extent_Record_Size, Res ) ;
                    Current := Temp ;

                    { End loop... }
                    fillchar( Allocation^, Extent_Record_Size, 0 ) ;
                    Index := 0 ;
                end ;
            end ;
            { At this point, Allocation contains the last Allocation
              record, and Index is the last Index in that record. }

            { Extend... }
            while( Old_Size < New_Size ) do { Until we are the requested size }
            begin
                Old_Size := Old_Size - Extent_Size ;
                inc( Index ) ;
                if( Index > Extent_Record_Size div sizeof( TStore_Address64 ) ) then
                begin
                    { Create new allocation record }
                    Temp := Shared_Store^.Allocate( Extent_Record_Size ) ;
                    if( Temp = 0 ) then { Allocation failed }
                    begin
                        Set_Error( Extent_StoreErr_No_Space_On_Store ) ;
                        Res := Last_Error ;
                        freemem( Allocation, Extent_Record_Size ) ;
                        exit ;
                    end ;
                    Shared_Store^.Fill( Temp, Extent_Record_Size, 0 ) ;
                    Allocation^[ 0 ] := Temp ;

                    { Update current allocation record... }
                    Shared_Store^.Write_Data( Allocation^, Current, Extent_Record_Size, Res ) ;
                    if( Res.Code <> 0 ) then
                    begin
                        Last_Error := Res ;
                        freemem( Allocation, Extent_Record_Size ) ;
                        exit ;
                    end ;

                    fillchar( Allocation^, Extent_Record_Size, 0 ) ;
                    Index := 1 ;
                end ;

                { Create new data record... }
                Temp := Shared_Store^.Allocate( Extent_Size ) ;
                if( Temp = 0 ) then { Allocation failed }
                begin
                    Set_Error( Extent_StoreErr_No_Space_On_Store ) ;
                    Res := Last_Error ;
                    freemem( Allocation, Extent_Record_Size ) ;
                    exit ;
                end ;
                Allocation^[ Index ] := Temp ;
            end ; { while( Old_Size < New_Size ) }

            { Update current allocation record... }
            Shared_Store^.Write_Data( Allocation^, Current, Extent_Record_Size, Res ) ;
            if( Res.Code <> 0 ) then
            begin
                Last_Error := Res ;
            end ;

            { Clean up... }
            freemem( Allocation, Extent_Record_Size ) ;
        end ;
        Size := Value ;
        Update_Header ;
    end ;
end ; { TSExtent_Store.Set_Max_Storage }


function TSExtent_Store.Extended_Size : TStore_Address64 ;

begin
    Extended_Size := Size + Shared_Store.Extended_Size ;
end ;


procedure TSExtent_Store.Define( _Root : TStore_Address64 ;
    SS : PSManaged_Store ) ;

var Header : TSExtent_Store_Header ;
    UEC : TUEC ;

begin
    if( SS <> nil ) then
    begin
        Shared_Store := SS ;
    end ;
    Root := _Root ;
    Shared_Store^.Read_Data( Header, Root, sizeof( Header ), UEC ) ;
    Data_Root := Header.Root ;
    Size := Header.Size ;
    Extent_Size := Header.Extent_Size ;
    Extent_Record_Size := Header.Extent_Record_Size ;
    _Contiguous := ( Header.Flags and 1 ) = 1 ;
end ;


procedure TSExtent_Store.Set_Contiguous( Value : boolean ) ;

var Allocation : PAllocation_Record ;
    Buffer : pointer ;
    Index : integer ;
    _Size : TStore_Address64 ;
    Rounded_Size : TStore_Address64 ;
    New_Root, Next, Saved : TStore_Address64 ;
    Position : TStore_Address64 ;
    UEC : TUEC ;

begin
    Set_Error( 0 ) ;
    if( Size = 0 ) then
    begin
        _Contiguous := Value ;
        Update_Header ;
    end ;

    if( _Contiguous <> Value ) then
    begin
        { Allocate buffers... }
        getmem( Allocation, Extent_Record_Size ) ;
        getmem( Buffer, Extent_Size ) ;
        fillchar( UEC, sizeof( UEC ), 0 ) ;

        { Round size up to nearest clustersize... }
        Rounded_Size := ( Size + Extent_Size - 1 ) ;
        Rounded_Size := Rounded_Size div Extent_Size ;
        Rounded_Size := Rounded_Size * Extent_Size ;
        _Size := Rounded_Size ;
        Saved := Data_Root ;
        Position := 0 ;

        { Convert... }
        if( Value ) then { Converting to contiguous }
        begin
            { Allocate a contiguous data area... }
            Data_Root := Shared_Store^.Allocate( _Size ) ;
            if( Data_Root = 0 ) then
            begin
                Data_Root := Saved ;
                Set_Error( Extent_StoreErr_No_Space_On_Store ) ;
                exit ;
            end ;

            { Copy extents to the contiguous area... }
            Shared_Store^.Read_Data( Allocation^, Saved, Extent_Record_Size, UEC ) ;
            if( UEC.Code <> 0 ) then
            begin
                Last_Error := UEC ;
                Shared_Store^.Deallocate( Data_Root, Rounded_Size ) ;
                Data_Root := Saved ;
                exit ;
            end ;
            Next := Allocation^[ 0 ] ;
            Index := 1 ;
            while( _Size > 0 ) do
            begin
                Shared_Store^.Read_Data( Buffer^, Allocation^[ Index ], Extent_Size, UEC ) ;
                if( UEC.Code <> 0 ) then
                begin
                    Last_Error := UEC ;
                    break ;
                end ;
                Shared_Store^.Write_Data( Buffer^, Data_Root + Position, Extent_Size, UEC ) ;
                if( UEC.Code <> 0 ) then
                begin
                    Last_Error := UEC ;
                    break ;
                end ;
                _Size := _Size - Extent_Size ;
                inc( Index ) ;
                Position := Position + Extent_Size ;
                if(
                    ( Index > Extent_Record_Size div sizeof( TStore_Address64 ) )
                    and
                    ( _Size > 0 )
                  ) then
                begin
                    { Move to next allocation record... }
                    Shared_Store^.Read_Data( Allocation^, Next, Extent_Record_Size, UEC ) ;
                    Next := Allocation^[ 0 ] ;
                    Index := 1 ;
                end ;
            end ; { while( _Size < 0 ) }
            if( UEC.Code <> 0 ) then { Some error - so undo the previous work }
            begin
                Shared_Store^.Deallocate( Data_Root, Rounded_Size ) ;
                Data_Root := Saved ;
                exit ;
            end ;

            { Free the extents... }
            { Note: we don't deallocate as we go in case something goes horribly
              wrong. }
            _Size := Rounded_Size ;
            Shared_Store^.Read_Data( Allocation^, Saved, Extent_Record_Size, UEC ) ;
            if( UEC.Code = 0 ) then
            begin
                Shared_Store^.Deallocate( Saved, Extent_Record_Size ) ;
                Next := Allocation^[ 0 ] ;
                Index := 1 ;
                while( _Size > 0 ) do
                begin
                    Shared_Store^.Deallocate( Allocation^[ Index ], Extent_Size ) ;
                    _Size := _Size - Extent_Size ;
                    inc( Index ) ;
                    if(
                        ( Index > Extent_Record_Size div sizeof( TStore_Address64 ) )
                        and
                        ( _Size > 0 )
                      ) then
                    begin
                        if( Next = 0 ) then { This shouldn't happen }
                        begin
                            break ; { But don't continue in any case }
                        end ;

                        { Move to next allocation record... }
                        Shared_Store^.Read_Data( Allocation^, Next, Extent_Record_Size, UEC ) ;
                        Shared_Store^.Deallocate( Next, Extent_Record_Size ) ;
                        Next := Allocation^[ 0 ] ;
                        Index := 1 ;
                    end ;
                end ; { while( _Size < 0 ) }
            end ; { if( UEC.Code = 0 ) }
        end else { Convert from contiguous to extents... }
        begin
            { Create extents and allocation records... }
            fillchar( Allocation^, Extent_Record_Size, 0 ) ;
            Index := 1 ;
            Next := Shared_Store^.Allocate( Extent_Record_Size ) ;
            New_Root := Next ;
            while( _Size > 0 ) do
            begin
                if( Allocation^[ Index ] = 0 ) then
                begin
                    Set_Error( Extent_StoreErr_No_Space_On_Store ) ;
                    break ;
                end ;
                Shared_Store^.Read_Data( Buffer^, Saved + Position, Extent_Size, UEC ) ;
                if( UEC.Code <> 0 ) then
                begin
                    Last_Error := UEC ;
                    break ;
                end ;
                Shared_Store^.Write_Data( Buffer^, Allocation^[ Index ], Extent_Size, UEC ) ;
                if( UEC.Code <> 0 ) then
                begin
                    Last_Error := UEC ;
                    break ;
                end ;
                inc( Index ) ;
                Position := Position + Extent_Size ;
                _Size := _Size - Extent_Size ;
                if(
                    ( Index > Extent_Record_Size div sizeof( TStore_Address64 ) )
                    and
                    ( _Size > 0 )
                  ) then
                begin
                    Allocation^[ 0 ] :=
                        Shared_Store^.Allocate( Extent_Record_Size ) ;
                    Shared_Store^.Write_Data( Allocation^, Next, Extent_Record_Size, UEC ) ;
                    if( UEC.Code <> 0 ) then
                    begin
                        Last_Error := UEC ;
                        break ;
                    end ;
                    Index := 1 ;
                    Next := Allocation^[ 0 ] ;
                    fillchar( Allocation^, Extent_Record_Size, 0 ) ;
                end ; { if }
            end ; { while( _Size > 0 ) }

            { Free old data... }
            Shared_Store^.Deallocate( Data_Root, Rounded_Size ) ;
            Data_Root := New_ROot ;
        end ; { if( Value ) }
        _Contiguous := Value ;
        Update_Header ;

        { Clean up... }
        freemem( Allocation, Extent_Record_Size ) ;
        freemem( Buffer, Extent_Size ) ;
    end ; { if( _Contiguous <> Value ) }
end ; { TSExtent_Store.Set_Contiguous }


function TSExtent_Store.Get_Contiguous : boolean ;

begin
    Get_Contiguous := _Contiguous ;
end ;


procedure TSExtent_Store.Create_New( ES, ERS, Flags : longint ;
    SS : PSManaged_Store ) ;

begin
    if( SS <> nil ) then
    begin
        Shared_Store := SS ;
    end ;
    Extent_Size := ES ;
    Extent_Record_Size := ERS ;
    _Contiguous := ( ( Flags and 1 ) = 1 ) ;
    Update_header ;
end ;


{$IFDEF Test}
procedure Error ;

begin
    Showmessage( 'Error' ) ;
    halt ;
end ;


procedure Test_Unit ;

var Buffer, Work : array[ 0..1024 ] of word ;
    Loop : integer ;
    Test : TSExtent_Store ;
    UEC : TUEC ;

    procedure Compare( M : integer ) ;

    begin
        M := M div sizeof( word ) ;
        while( M > 0 ) do
        begin
            dec( M ) ;
            if( Buffer[ M ] <> Work[ M ] ) then
            begin
                Error ;
            end ;
        end ;
    end ;

begin
    for Loop := 0 to 1024 do
    begin
        Buffer[ Loop ] := Loop ;
    end ;

    Test.Init ;
    Test.Initialize ;
    if( Test.Size <> 0 ) then
    begin
        Error ;
    end ;
    Test.Create_New( 512, 512, 0, Application_Manager^.Heap ) ;
    if( Test.Read_Data( Buffer, 0, 1, UEC ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Write_Data( Buffer, 0, 1, UEC ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Min_Storage <> Application_Manager^.Heap.Min_Storage ) then
    begin
        Error ;
    end ;
    if( Test.Read_Only ) then
    begin
        Error ;
    end ;
    if( Test.Write_Only ) then
    begin
        Error ;
    end ;
    if( not Test.Contiguous_Store ) then
    begin
        Error ;
    end ;
    if( Test.Max_Storage <> Test.Size ) then
    begin
        Error ;
    end ;
    if( Test.Extend( 0 ) <> 0 ) then
    begin
        Error ;
    end ;
    Test.Extended_Size ;

    Test.Set_Max_Storage( 10 ) ;
    if( Test. Size <> 10 ) then
    begin
        Error ;
    end ;
    if( Test.Write_Data( Buffer, 0, 11, UEC ) <> 10 ) then
    begin
        Error ;
    end ;
    if( Test.Read_Data( Work, 0, 11, UEC ) <> 10 ) then
    begin
        Error ;
    end ;
    Compare( 10 ) ;
    Test.Extend( 510 ) ;
    if( Test.Size <> 520 ) then
    begin
        Error ;
    end ;
    if( Test.Write_Data( Buffer, 500, 10, UEC ) <> 10 ) then
    begin
        Error ;
    end ;
    if( Test.Read_Data( Work, 500, 10, UEC ) <> 10 ) then
    begin
        Error ;
    end ;
    Compare( 10 ) ;
    if( Test.Write_Data( Buffer, 0, 520, UEC ) <> 10 ) then
    begin
        Error ;
    end ;
    Test.Set_Contiguous( True ) ;
    if( Test.Read_Data( Buffer, 0, 520, UEC ) <> 10 ) then
    begin
        Error ;
    end ;
    Compare( 520 ) ;
    Test.Set_Contiguous( False ) ;
    if( Test.Read_Data( Buffer, 0, 520, UEC ) <> 10 ) then
    begin
        Error ;
    end ;
    Compare( 520 ) ;
    Test.Define( Test.Root ) ;
    if( Test.Read_Data( Buffer, 0, 520, UEC ) <> 10 ) then
    begin
        Error ;
    end ;
    Compare( 520 ) ;

    Test.Terminate ;
    Test.Done ;
end ;
{$ENDIF}


end.

