{
        File Name    : StoreLis
        Package Name : LB
        Purpose      : Store List
        Institution  : Conroy & Conroy Co.
        Date Written : 29-Oct-2005
        Written By   : Alan Conroy
        Version      : 1.0

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit contains a list class where the data is stored on a
        managed store.
}

unit StoreLis ;

interface

uses // C&C...
     BigFH, // TBig_FH
     Common ; // TBase_Delphi_Object

type TStore_List64 = class( TBase_Delphi_Object )
                         public // Constructors and destructors...
                             constructor Create ;

                         private // Instance data...
                             _Address : int64 ;
                             _Capacity : int64 ;
                             _Count : int64 ;
                             List : int64 ; // First chunk of list
                             _Store : PSBig_FH ;
                             _Delta : longint ;

                         private // Cached data...
                             Cached_Address : int64 ; // Address of last chunk (0 if nothing cached)
                             Cached_Size : int64 ; // Size of last chunk
                             Cached_Offset : int64 ; // First offset in last chunk

                         protected // Property handlers...
                             function Get_Item( Index : int64 ) : int64 ;
                             procedure Set_Item( Index, Value : int64 ) ;
                             procedure Set_Address( Value : int64 ) ;
                             procedure Set_Capacity( Value : int64 ) ;
                             procedure Set_Count( Value : int64 ) ;
                             procedure Set_Delta( Value : longint ) ;
                             procedure Set_Store( Value : PSBig_FH ) ;

                         protected // Internal utility routines...
                             procedure Update_Header ;
                             function Address_Of( Index : int64 ) : int64 ;

                         public // API...
                             function Add( Value : int64 ) : int64 ;
                             function Add_Insert( Value : int64 ) : int64 ;
                             function IndexOf( Value : int64 ) : int64 ;

                             property Address : int64
                                 read _Address
                                 write Set_Address ;
                             property Items[ Index : int64 ] : int64
                                 read Get_Item
                                 write Set_Item ; default ;
                             property Capacity : int64
                                 read _Capacity
                                 write Set_Capacity ;
                             property Count : int64
                                 read _Count
                                 write Set_Count ;
                             property Delta : longint
                                 read _Delta
                                 write Set_Delta ;
                             property Store : PSBig_FH
                                 read _Store
                                 write Set_Store ;
                     end ; // TStore_List64

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses // C&C...
     _App, // Application_Manager
     _UEHDefs ; // TUEC

{ TStore_List64 is a class that wraps a list in the database.  In the database, the
  list has a header (see List_Record for the header image).  The List's data is
  organized in a single-linked list of "chunks".  Each chunk contains a size,
  pointer, and the data for that chunk, like so:
    Size : int64                    Size of this chunk, in elements (bytes div 8)
    Next : int64                    Pointer to next chunk (0 if this is the last chunk)
    Delta : int64                   Minimum capacity extension size ( < 1 is = 1 )
    Data : array[ size ] of int64   Data for this chunk

  An example list:
      Header: Capacity = 10
              Count = 9
              Delta = 0
              List ---------> Size = 7
                              Next -----------> Size = 3
                              Item[ 0 ]         Item[ 7 ]
                              Item[ 1 ]         Item[ 8 ]
                              Item[ 2 ]         Item[ 9 ] (unused)
                              Item[ 3 ]
                              Item[ 4 ]
                              Item[ 5 ]
                              Item[ 6 ]
}
type List_Record = record
                        Capacity : int64 ;
                        Count : int64 ;
                        Delta : int64 ;
                        List : int64 ; // First chunk of list
                        Extension : int64 ;
                   end ;

// TStore_List64 methods...

// Constructors and destructors...

constructor TStore_List64.Create ;

begin
    inherited Create ;

    _Delta := 4 ;
end ;


// Property handlers...

function TStore_List64.Get_Item( Index : int64 ) : int64 ;

var Address : int64 ;

begin
    if( ( Index < 0 ) or ( Index >= _Count ) ) then
    begin
        Application_Manager^.Fatal_Exception( 'Invalid list index' ) ;
    end ;
    Address := Address_Of( Index ) ;
    Store.Read( Address, sizeof( Result ), Result ) ;
end ;


procedure TStore_List64.Set_Item( Index, Value : int64 ) ;

var Address : int64 ;

begin
    if( ( Index < 0 ) or ( Index >= _Count ) ) then
    begin
        Application_Manager^.Fatal_Exception( 'Invalid list index' ) ;
    end ;
    Address := Address_Of( Index ) ;
    Store.Write( Address, sizeof( Value ), Value ) ;
end ;


procedure TStore_List64.Set_Address( Value : int64 ) ;

var Rec : List_Record ;

begin
    if( Address <> Value ) then
    begin
        _Address := Value ;
        if( ( Value <> 0 ) and ( _Store <> nil ) ) then
        begin
            Store.Read( Address, sizeof( Rec ), Rec ) ;
            _Count := Rec.Count ;
            _Capacity := Rec.Capacity ;
            _Delta := Rec.Delta ;
            List := Rec.List ;
        end else
        begin
            _Count := 0 ;
	    _Capacity := 0 ;
            List := 0 ;
        end ;
        Cached_Address := 0 ;
    end ;
end ;


procedure TStore_List64.Set_Capacity( Value : int64 ) ;

var Current, Next : int64 ;
    Size : int64 ;

begin
    if( Value > _Capacity ) then // For now, we don't shrink the list size
    begin
        if( ( Delta > 1 ) and ( Value < Delta ) ) then
        begin
            Value := Delta ;
        end ;
        _Capacity := Value ;
        if( Cached_Address <> 0 ) then
        begin
            Current := Cached_Address ;
            Value := Value - Cached_Offset ;
        end else
        begin
            if( List = 0 ) then
            begin
                List := Store.Allocate( ( Value + 2 ) * sizeof( int64 ) ) ;
                Store.Fill( List, ( Value + 2 ) * sizeof( int64 ), 0 ) ;
                Store.Write( List, sizeof( Value ), Value ) ;
                Update_Header ;
                exit ;
            end ;
            Current := List ;
        end ;
        while( Value >= 0 ) do
        begin
            Store.Read( Current, sizeof( Size ), Size ) ;
            Value := Value - Size ;
            Store.Read( Current + sizeof( int64 ), sizeof( Next ), Next ) ;
            if( Next = 0 ) then // Found end of current list
            begin
                Next := Store.Allocate( ( Value + 2 ) * sizeof( int64 ) ) ;
                Store.Fill( Next, ( Value + 2 ) * sizeof( int64 ), 0 ) ;
                Size := Value ;
                Store.Write( Next, sizeof( Size ), Size ) ;
                Store.Write( Current + sizeof( int64 ), sizeof( Next ), Next ) ;
                break ;
            end ;
            Current := Next ; // Follow pointer to next chunk
        end ; // while( Value >= 0 )
    end ; // if( Value > _Capacity )
end ; // TStore_List64.Set_Capacity


procedure TStore_List64.Set_Count( Value : int64 ) ;

begin
    if( Value <> _Count ) then
    begin
        if( Value > _Capacity ) then
        begin
            Capacity := Value ;
        end ;
        _Count := Value ;
        Update_Header ;
    end ;
end ;


procedure TStore_List64.Set_Delta( Value : longint ) ;

begin
    if( Value <> _Delta ) then
    begin
        _Delta := Value ;
        Update_Header ;
    end ;
end ;


procedure TStore_List64.Set_Store( Value : PSBig_FH ) ;

begin
    if( Value <> _Store ) then
    begin
        if( Value <> nil ) then
        begin
            Value.Attach ;
        end ;
        if( _Store <> nil ) then
        begin
            _Store.Detach ;
        end ;
        _Store := Value ;
        Cached_Address := 0 ;
        _Address := 0 ;
    end ;
end ;


// Internal utility routines...

function TStore_List64.Address_Of( Index : int64 ) : int64 ;

var Current, Next : int64 ;
    Size : int64 ;
    Start_Index : integer ;

begin
    Result := 0 ;
    if( Cached_Address <> 0 ) then
    begin
        Current := Cached_Address ;
        Index := Index - Cached_Offset ;
        Start_Index := Cached_Offset ;
    end else
    begin
        Current := List ;
        Start_Index := 0 ;
    end ;
    while( Index >= 0 ) do
    begin
        Store.Read( Current, sizeof( Size ), Size ) ;
        if( Index < Size ) then
        begin
            Result := Current + Index * sizeof( int64 ) + ( 2 * sizeof( int64 ) ) ;
                 // Chunk root  ==== index offset =====   ===== size & next =====
            Cached_Address := Current ;
            Cached_Size := Size ;
            Cached_Offset := Start_Index ;
            exit ;
        end ;
        Index := Index - Size ;
        Start_Index := Start_Index + Size ;
        Store.Read( Current + sizeof( int64 ), sizeof( Next ), Next ) ;
        if( Next = 0 ) then // Should not happen
        begin
            exit ; { UNTESTED }
        end ;
        Current := Next ; // Follow pointer to next chunk
    end ;
end ; // TStore_List64.Address_Of


procedure TStore_List64.Update_Header ;

var Temp : List_Record ;

begin
    if( ( Address = 0 ) and ( Store <> nil ) ) then
    begin
        _Address := Store.Allocate( sizeof( Temp ) ) ;
    end ;
    if( Address <> 0 ) then { UNTESTED: else only happens when store is full }
    begin
        fillchar( Temp, sizeof( Temp ), 0 ) ;
        Temp.Capacity := _Capacity ;
        Temp.Count := _Count ;
        Temp.List := List ;
        Temp.Delta := Delta ;
        Store.Write( Address, sizeof( Temp ), Temp ) ;
    end ;
end ;


// API...

function TStore_List64.Add( Value : int64 ) : int64 ;

begin
    Result := _Count ;
    Count := _Count + 1 ;
    Set_Item( Result, Value ) ;
end ;


// Find first 0 entry and set to Value, else add to end.
function TStore_List64.Add_Insert( Value : int64 ) : int64 ;

begin
    Result := IndexOf( 0 ) ;
    if( Result = -1 ) then
    begin
        Result := Add( Value ) ;
    end else
    begin
        Items[ Result ] := Value ;
    end ;
end ;


function TStore_List64.IndexOf( Value : int64 ) : int64 ;

var Index : integer ;

begin
    for Index := 0 to Count - 1 do
    begin
        if( Items[ Index ] = Value ) then
        begin
            Result := Index ;
            exit ;
        end ;
    end ;
    Result := -1 ;
end ;



{$IFDEF Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


procedure Test_Unit ;

var Address : int64 ;
    Test : TStore_List64 ;
    Store : TSBig_FH ;
    _Result : TUEC ;

begin
    Store.Init( 16, 0, 'C:\LB\QC\StoreList.tst', _Result, nil ) ;
    Store.Attach ;
    Test := TStore_List64.Create ;
    Test.Store := @Store ;

    if( Test.Add( 1 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test[ 0 ] <> 1 ) then
    begin
        Error ;
    end ;
    Test.Capacity := 7 ;
    if( Test.Add( 5 ) <> 1 ) then
    begin
	Error ;
    end ;
    if( Test[ 0 ] <> 1 ) then
    begin
        Error ;
    end ;
    if( Test[ 1 ] <> 5 ) then
    begin
        Error ;
    end ;
    Test.Count := 7 ;
    if( Test[ 0 ] <> 1 ) then
    begin
        Error ;
    end ;
    if( Test[ 1 ] <> 5 ) then
    begin
        Error ;
    end ;
    if( Test[ 6 ] <> 0 ) then
    begin
        Error ;
    end ;
    Test.Delta := 8 ;
    Test.Delta := 8 ;
    Test.Count := Test.Count ;
    Test.Store := Test.Store ;
    Address := Test.Address ;
    Test.Count := Test.Count - 1 ;
    Test.Capacity := 16 ;
    Test.Free ;
    Test := TStore_List64.Create ;
    Test.Store := @Store ;
    Test.Address := Address ;
    if( Test[ 0 ] <> 1 ) then
    begin
        Error ;
    end ;
    if( Test[ 1 ] <> 5 ) then
    begin
        Error ;
    end ;
    if( Test[ 5 ] <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Delta <> 8 ) then
    begin
        Error ;
    end ;
    Test.Store := nil ;
    Test.Store := @Store ;
    Test.Address := Address ;
    Test.Address := 0 ;
    Test.Address := 0 ;
    Test.Capacity := 1 ;
    try
        Test[ -1 ] := 0 ;
        Error ;
    except
    end ;
    try
        Test[ -1 ] ;
        Error ;
    except
    end ;

    Test.Free ;
    Store.Done ;
end ;
{$ENDIF}



end.
