{$N+}
{$A-}
unit FSFH ;
{
        Program Name : FSFH
        Package Name : LB
        Purpose      : Store Heap and File System routines
        Institution  : Conroy & Conroy Co.
        Date Written : 1998
        Written By   : Alan Conroy
        Version      : 1.2F

        Copyright (C) 1998-2006 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *  1.0A                                                     *
        *  23-Jan-2002   EAC        Added Separator to TFind_Record.*
        *                                                           *
        *  1.1                                                      *
        *  4-Mar-2002    EAC        Reorganized all of the code.    *
        *                                                           *
        *  1.2                                                      *
        *  11-Jun-2002   EAC        Added data caching.             *
        *  19-Aug-2002              Removed extra writes during	    *
        *                           extensions.                     *
        *   2-Dec-2002   EAC        Handle strings in objects.      *
        *  23-Oct-2003   EAC        Reset error code in Read.       *
        *  15-Dec-2004   EAC        Catch certain data corruptions. *
        *   9-Aug-2005   EAC        Catch more data corruptions.    *
        *  11-Apr-2006   EAC        Don't allow "\" in filenames or *
        *                           "..." as the filename.  Catch   *
        *                           more data corruptions.  Also    *
        *                           rewrote Tree_Delete, which was  *
        *                           possibly causing corruption in  *
        *                           the first place.                *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit contains the Store Heap and File System routines.

        TXFile_System File structure:
                o Each device is a separate file system
                o A device is a Store object.
                o Each file system is implemented as a store heap, with
                  allocation resolution of 256 bytes.

                Structure:
                        Boot record (store heap "root")
                                |
                                |
                                v
                        Device header record
                                |
                                |
                                v
                        Root Directory record
                                |
                                |
                                v
                        Root Directory root entry
                                  /\
                                 /  \
                              null  file entry
                                          /\
                                         /  \
                                file entry  file entry

        Note: Each directory has an entry in it's parent directory.
        This master entry points to the root entry (a copy of the
        master entry except for pointers), which is the root of a binary
        tree of file entries.  The following illustrates this for a directory
        structure of \A\B, \A\C, and \A\C\D (created in that order):

                  Master entry (A)  [has directory flag set]
                        |
                        | (allocation_info)
                        |
                        v
                  Root entry        [has name of NUL]
                        |
                        | (right_pointer)
                        |
                        v
                  File entry (B)
                        /\
        (left_pointer) /  \ (right_pointer)
                      /    \
                    null   File (master) entry (C)  [has directory flag set]
                                |
                                | (allocation_info)
                                |
                                v
                           Root entry               [has name of NUL]
                                |
                                | (right_pointer)
                                |
                                v
                           File entry (D)

        Store Heap structure:
                Root (size = resolution, address = 0)
                Header (size = resolution, address = resolution)
                data (size is variable)
                Allocation Table (size is variable)

        Note that in old versions of the Store Heap, there was no header.
        Thus the current store heap component is incompatible with older
        versions.
}

interface

{$I EDEFINES.INC}

uses
{$IFDEF WINDOWS}
     WinDOS, Objects, Strings,
{$ELSE}
  {$IFDEF DELPHI}
     SysUtils,
  {$ELSE}
     DOS, WinDOS, Objects,
  {$ENDIF}
{$ENDIF}
     Allocatr, { TAT }
     Collect,
     _DebugIn, { PSDebug_Interface }
     _UEHDefs, { TEU }
     UEHDefs, { TStandard_Error_Interface }
     TypeDefs, { pByte_Array }
     Files, { Must be AFTER TypeDefs }
     Heaps, { PSHeap }
     MStores, { TSManaged_Store }
     _Stores ; { TStore_Address }

const FS_Maximum_File_Name = 127 ; { Maximum size of file names }

type FS_Boot_Record = record
                          Bootstrap : array[ 0..63 ] of byte ;
                          Header : TStore_Address ;
                          Physical_Block_Size : smallint ;
                          Filler : smallint ; { 2 unused bytes out of 8 }
                          Reserved : array[ 0..187 - sizeof( TStore_Address ) ] of byte ;
                      end ;

     FH_Header = record { 16-bytes }
                     ID : smallint ;
                     ID1 : byte ;
                     Version : byte ; { FH format version 10 = V1.0 }
                     AT_Offset : longint ; { Current AT location }
                     AT_Size : longint ; { Current AT size }
                     AT_Offset_High : longint ; { Upper 32 bits for 64-bit file systems }
                 end ;

      { Store Heap processing flags... }
const FHF_NoExtend = 1 ; { Don't extend file }
      FHF_NoClose = 2 ; { Don't close file when done }
      FHF_Fast_Allocate = 4 ; { Don't reallocate deallocated memory }
      FHF_Fast_Deallocate = 8 ; { Don't deallocate memory }
      FHF_Read_Only = 16 ; { Don't allow writes }
      { 32-32768 = reserved }

      { Store Heap error codes... }
      FHErr_Facility = 5 ;
      FHErr_Success = 0 ;
      FHErr_No_Room_On_Device = 1 ;
      FHErr_Corrupted_File_Structure = 2 ;
      FHErr_Invalid_Pointer_Operation = 3 ;
      FHErr_Deallocation_Error = 4 ;
      FHErr_Size_Mismatch = 5 ;
      FHErr_Read_Out_Of_Range = 6 ;
      FHErr_Read_Size_Error = 7 ;
      FHErr_Write_Out_Of_Range = 8 ;
      FHErr_Write_Size_Error = 9 ;
      FHErr_Bad_Minimum = 10 ;
      FHErr_Write_Failure = 11 ;

type TFH_Statistics = record
                          Index_Reads : longint ; { Cached index physical read count }
                          Index_Writes : longint ; { Cached index physical write count }
                          Index_Cache_Hits : longint ; { Number of index reads that were cached }
                      end ;

     pDirectory_Entry = ^TDirectory_Entry ;
     TDirectory_Entry = record
                            Name : string[ FS_Maximum_File_Name ] ;
                            Version : smallint ;
                            Version_Limit : smallint ;
                            Size : TStore_Address ;
                            Owner : longint ;
                            Created_By : longint ;
                            Creation_Time : longint ;
                            Creation_Date : longint ;
                            Revision_Time : longint ;
                            Revision_Date : longint ;
                            Access_Time : longint ;
                            Access_Date : longint ;
                            Backup_Time : longint ;
                            Backup_Date : longint ;
                            Expiration_Time : longint ;
                            Expiration_Date : longint ;
                            Protection : longint ;
                            Attributes : longint ; { FSFF flags }
                            Extension : longint ;
                            End_Of_File : longint ;
                            Record_Size : longint ;
                            Uncompressed_Size : longint ;
                            Caching_Info : TStore_Address ;
                            Access_Info : TStore_Address ;
                            Locks : longint ;
                            Auditing_Info : TStore_Address ;
                            Access_Stats : TStore_Address ;
                            Allocation_Info : TStore_Address ;
                            { Pointer to allocation list for file.  For
                              directories: pointer to L/R records for files
                              in the directory. }

                            Next_Alpha_Entry : TStore_Address ;
                            Organization : byte ;
                            Record_Format : byte ;
                            Left_Pointer : TStore_Address ; { < pointer }
                            Right_Pointer : TStore_Address ; { > pointer }
                            Back_Pointer : TStore_Address ; { Father node pointer }
                            Reserved : array[ 0..9 - sizeof( TStore_Address ) ] of byte ;
                        end ; { TDirectory_Entry }

const Max_Cluster_In_Block = ( sizeof( TDirectory_Entry ) - sizeof( TStore_Address ) ) div sizeof( TStore_Address ) - 1 ;

type PFH = ^TFH ;
     TFH = object( TSManaged_Store ) { Store Heap }
              private
{$IFDEF Debug}
                  Sanity : string[ 3 ] ;
{$ENDIF}
                  AT : TAT ;
{                  AT_Initialized : boolean ; }
{$IFDEF Verify}
                  Address_List : TInteger_List ;
                  Size_List : TInteger_List ;
{$ENDIF}
                  Byte_Mapping : longint ; { How much one AT byte maps }

              public
                  FH_Store : PSStore ;
                  Allow_Cache : boolean ; { True if AT is cached }
                  Device : integer ; { File device }
                  Flags : integer ; { Processing flags.  See FHF_* }
                  Memory : PSManaged_Store ; { Where to allocate cached information }
                  Own_Allocator : boolean ; { True if we own Memory }
                  Resolution : longint ; { How much one AT bit maps }
                  Track_FH : boolean ;
                  Serial_Number : longint ;
                  FH_Statistics : TFH_Statistics ;
                  Extension_Resolution : longint ; { Minimum amount to extend file by, in percentage of file size }
                  Last_Error : TUEC ;
                  Lowest_Dynamic : TStore_Address ; { Lowest modifiable address }

              public { Constructors and destructors... }
                  constructor Init( Minimum, F : integer ; const Name : string ;
                      var _Result : tUEC ; Mem : PSManaged_Store ) ;
                  constructor Init_On_Store( Minimum, F : integer ; S : PSStore ;
                      var _Result : tUEC ; Mem : PSManaged_Store ) ;
                  procedure Common_Init( Minimum, F : integer ; S : PSStore ;
                      var _Result : tUEC ; Mem : PSManaged_Store ) ;
                  destructor Done ; virtual ; { Finish up FH access }

              private { Internal utility routines }
                  procedure Log_Error( C, S : integer ) ;
                  procedure Update_Header ;

              public { Overrides... }
                  function Classname : string ; virtual ;
                  function Is_Class( const Name : string ) : boolean ; virtual ;

                  function Extended_Size : TStore_Address ; virtual ;

              public { API... }
                  procedure Determine_AT_Metrics( var UEC : TUEC ) ; virtual ;
                  procedure New( _Size : TStore_Address ) ; virtual ;
                  { Create new FH, if empty }

                  function Allocate( _Size : TStore_Address ) : TStore_Address ;
                      virtual ;
                  { Allocate space }

                  function Allocate_At( _Position, _Size : TStore_Address ) : boolean ;
                      virtual ;
                  { Allocate space at a specific location }

                  procedure Deallocate( PTR : TStore_Address ; _Size : TStore_Address ) ;
                      virtual ;
                  { Deallocate space in FH }

                  procedure Read( PTR : TStore_Address ; _Size : TStore_Address ;
                      var Buffer ) ; virtual ; { Read data from FH }
                  procedure Read_Heap( PTR : TStore_Address ; _Size : TStore_Address ;
                      var Buffer ) ; virtual ; { Read data from FH }
                  function Reallocate( PTR : TStore_Address ;
                      Old, _New : TStore_Address ) : TStore_Address ; virtual ;
                      { Reallocate existing data }

                  procedure Write( PTR : TStore_Address ;
                      _Size : TStore_Address ; var Buffer ) ; virtual ;
                  procedure Write_Heap( PTR : TStore_Address ; _Size : TStore_Address ;
                      var Buffer ) ; virtual ; { Write data to FH }
                  procedure Set_Root( var Buffer ) ; virtual ;
                  { Write data to root block of FH }

                  procedure Get_Root( var Buffer ) ; virtual ;
                  { Read data from root block of FH }

                  function Path : string ; virtual ;

                  function SpaceAvail : TStore_Address ; virtual ;
                  function XSpaceAvail : TStore_Address ; virtual ;
                  function MaxSpace : TStore_Address ; virtual ;
                  function Debugger : PSDebug_Interface ; virtual ;

                  procedure Set_Cache_Flag( Flag : boolean ) ;
                  function Valid_Data_Address( Address : TStore_Address ) : boolean ;
                  procedure Write_Cache ;

                  procedure Log_Allocation( Address : TStore_Address ; _Size : TStore_Address ) ;
                  procedure Log_Deallocation( Address : TStore_Address ; _Size : TStore_Address ) ;
                  procedure Log_Read( Address : TStore_Address ; _Size : TStore_Address ) ;
                  procedure Log_Write( Address : TStore_Address ; _Size : TStore_Address ) ;
           end ; { TFH }

     TNative_File_Record = record
                               Name : string ;
                               Directory_Entry : TStore_Address ;
                               File_Entry : TStore_Address ; { Location of file entry }
                           end ;
     PNative_File_Record = ^TNative_File_Record ;

     PXFile_System = ^TXFile_System ;

     FS_Header_Record = record
                            ID : string[ 127 ] ; { Label for this file system device }
                            Root : longint ; { Pointer to root of file system/directories }
                            Boot_Pointer : longint ; { Pointer to main bootstrap (0 = not bootable) }
                            Boot_Size : longint ; { Size of main bootstrap, in bytes }
                            Version : smallint ; { File system structure version }
                            Default_Cluster_Size : smallint ; { Default cluster size for allocations (0 is invalid) }
                            Initialized : array[ 0..1 ] of longint ; { Currently unused }
                            Flags : smallint ; { FSD_* }
                            Lowest_Dynamic : longint ; { Lowest address that can be modified }
                            Lock_PID : longint ; { PID who currently has File System locked (0=none) }
                            Lock_Time : longint ; { Time stamp of PID lock }
                            Header_Version : longint ; { Version of header }
                            Last_Lock_PID : longint ; { Last PID who locked this File System }
                            Lock_Instance : longint ; { Instance ID for Lock_PID }
                            Last_Lock_Instance : longint ; { Instance ID for Last_Lock_PID }
                            Lock_Context : longint ; { Context for Lock_Pid/Lock_Instance }
                            Last_Lock_Context : longint ; { Context for Last_Lock_PID/Last_Lock_Instance }
                            Reserved : array[ 0..65 ] of byte ; { Currently unused }
                        end ;

     TAllocation_Record = record
                                Cluster : array[ 0..Max_Cluster_In_Block ] of TStore_Address ;
                                Next : TStore_Address ;
                          end ;

     TXFile_System = object( TFile_System )
                        private
                            First_Lock : boolean ; { True if we haven't yet locked the file }
                            Locked : boolean ; { True if we are locked }
                            Cached_FS_Header : FS_Header_Record ;
                            Serial_Number : integer ;
                            Memory_Heap : pSHeap ; { Heap to use for volatile storage }
                            Boot : FS_Boot_Record ; { Cached boot block from }
                            Root : TStore_Address ;

                            { Cached information about last opened file... }
                            { Note that we have duplicated variables rather
                              than an array since using an array actually
                              makes the code run SLOWER.  A second entry gives
                              a small improvement over one entry. }
                            Cached_File_Name1 : string ;
                            Cached_Directory_Entry1 : TStore_Address ;
                            Cached_File_Entry1 : TStore_Address ;
                            Cached_Path1 : string ;
                            Cached_File_Name2 : string ;
                            Cached_Directory_Entry2 : TStore_Address ;
                            Cached_File_Entry2 : TStore_Address ;
                            Cached_Path2 : string ;
                            Cached_Entries : integer ; { Number of cached entries }
{$IFDEF MUTEX}
                            Lock_Mutex : integer ;
{$ENDIF}
{$IFDEF Debug}
                            Sanity : string[ 13 ] ;
{$ENDIF}

                        public { Public instance data... }
                            File_Heap : PFH ; { Store heap which controls file }
                            Flags : integer ; { File system flags }

                        public { Constructors and destructors... }
                            constructor Init( const Name : string ;
                                _Flags : integer ; var Res : TUEC ) ;
                            constructor Init_From_Store( Store : PSStore ;
                                _Flags : integer ; var Res : TUEC ) ;
                            destructor Done ; virtual ;
                            procedure Common_Init ;

                        public { API... }
                            procedure Attach ; virtual ;

                            procedure Detach ; virtual ;

                            procedure Close( Handle : integer ) ; virtual ;

                            function Create( const Name : string ;
                                Mode : integer ) : integer ; virtual ;

                            function Debugger : PSDebug_Interface ; virtual ;

                            procedure Erase( const Name : string ) ; virtual ;

                            function File_Name( Handle : integer ) : string ;
                                 virtual ;

                            function First_Directory( const Name : string ) : integer ;
                                virtual ;

                            procedure Flush( Handle : integer ) ; virtual ;

                            function Get_File_Attributes( const Name : string ) : longint ;
                                virtual ;

                            function Get_Size( Handle : integer ) : TStore_Address ;
                                 virtual ;

                            function Get_Time( Handle, Typ : integer ) : Longint ;
                                 virtual ;

                            function Get_File_Object( Mode : integer ) : PSFile ;
                                virtual ;

                            procedure Read( Handle : integer ;
                                Position : TStore_Address ; Len : TStore_Address ;
                                var Buffer ; var Count : TStore_Address ) ; virtual ;

                            procedure Rename( const Old, New : string ) ;
                                virtual ;

                            procedure Set_File_Attributes( const Name : string ;
                                Value : longint ) ; virtual ;

                            procedure Set_Time( Handle : integer ;
                                Typ, Time : Longint ) ; virtual ;

                            procedure Set_Size( Handle : integer ;
                                New_Size : TStore_Address ) ; virtual ;

                            procedure Write( Handle : integer ;
                                Position : TStore_Address ; Len : TStore_Address ;
                                var Buffer ; var Count : TStore_Address ) ;
                                virtual ;

                            function DeviceSize( Nam : string ) : TStore_Address ;
                                virtual ;

                            function DeviceFree( Nam : string ) : TStore_Address ;
                                virtual ;

                            function MKDir( Nam : string ) : longint ;
                                virtual ;

                            procedure Get_Header( _Path : string ;
                                var Header : FS_Header_Information ;
                                var Res : TUEC ) ; virtual ;

                            procedure Set_Header( _Path : string ;
                                var Header : FS_Header_Information ;
                                var Res : TUEC ) ; virtual ;

                            function Facility : longint ; virtual ;

                            function Version : longint ; virtual ;

                            function Lock( Nam : string ) : boolean ; virtual ;

                            procedure Unlock( Nam : string ) ; virtual ;

                            function FindFirst( Nam : string ; Attributes : longint ;
                                var Find_Record : TFind_Record ) : longint ;
                                virtual ;

                            function FindNext( var Find_Record : TFind_Record ) : longint ;
                                virtual ;

                            procedure Initialize( Nam, ID : string ;
                                _Flags : longint ; var UEC : tUEC ) ; virtual ;

                            procedure Initialize_Store( S : PSStore ;
                                ID : string ; _Flags : longint ;
                                var UEC : tUEC ) ; virtual ;

                            procedure Verify( Name : string ; var Res : TUEC ) ;
                                virtual ;

                            function Normalize_Path( Name : string ) : string ;
                                virtual ;

                            function Open( const Name : string ;
                                Mode : integer ) : integer ; virtual ;

                            function Valid_Filename( var S : string ) : boolean ;
                                virtual ;

                            function Path : string ; virtual ;

                            procedure Set_Device( Name : string ;
                                _Flags : integer ; var Res : TUEC ) ; virtual ;

                            procedure Connect_Store_Heap( var Res : TUEC ) ;
                                virtual ;

                            procedure Set_Cache( Limit : longint ) ; virtual ;

                        private { Internal utility routines... }
                            function Allocate_Entry : TStore_Address ;
                            procedure Clear_Cache ;
                            procedure Clear_File_Lookup_Cache( _Root : TStore_Address ) ;
                            function Correct_Entry( Current, Directory_Entry : TStore_Address ;
                                var Entry : TDirectory_Entry ;
                                const _Root : TStore_Address ) : boolean ;
                            procedure Delete_File( File_Entry : TStore_Address ) ;
                            procedure Extend_File( File_Entry : TStore_Address ;
                                New_Size : longint ;
                                var Entry : tDirectory_Entry ) ;

                            { Find specified file.  Returns the address of the
                              file entry for the matching file.  Directory_Entry
                              is ignored when passed and is the directory entry
                              of the file otherwise.  If file is not found, both
                              the result and Directory_Entry are 0.  If no file
                              is specified and the path is found, the result is
                              0 and Directory_Entry is the directory entry of
                              the path.  The returned Directory entry is the
                              address of the file entry for the directory.  Name
                              must be uppercase and with no leading/trailing
                              spaces.}
                            function Find_File( Name : string ;
                                var Directory_Entry : TStore_Address ) : longint ;

                            procedure Fix_Back_Pointer( Previous, Current : TStore_Address ;
                                var Entry : TDirectory_Entry ) ;
                            function Get_Absolute( Allocation, Offset : longint ) : longint ;

                            function Get_File_Rec( Handle : integer ) : PNative_File_Record ;

                            procedure Read_Header( _File_Heap : PFH ;
                                Address : TStore_Address ;
                                var Header : FS_Header_Record ) ;
                            procedure Write_Header( _File_Heap : PFH ;
                                Address : TStore_Address ;
                                var Header : FS_Header_Record ) ;
                            procedure Read_Entry( Address : TStore_Address ;
                                var Entry : TDirectory_Entry ) ;
                            procedure Write_Entry( Address : TStore_Address ;
                                var Entry : TDirectory_Entry ) ;
                            procedure Read_Allocation( Address : TStore_Address ;
                                var Entry : TAllocation_Record ) ;
                            procedure Write_Allocation( Address : TStore_Address ;
                                var Entry : TAllocation_Record ) ;
                            procedure Set_Back_Pointer( Address, New_Back : longint ) ;
                            function Tree_Find( PTR : longint ;
                                Key : string ) : longint ;
                            procedure Tree_Add( var PTR : TStore_Address ;
                                Entry : TDirectory_Entry ;
                                Existing_Ptr : TStore_Address ) ;
                            procedure Tree_Delete( Index : TStore_Address ) ;
                            function Tree_Next( Index : longint ) : longint ;
                            procedure Repair_Tree( Directory_Entry, Current : TStore_Address ) ;
                            procedure Set_Error( X : integer ) ;
                            procedure Verify_Lock ;
                    end ; { TXFile_System }

const Entry_Size = sizeof( TDirectory_Entry ) ;

const TNative_FileErr_Facility = 44 ;

type PNative_File = ^TNative_File ;
     TNative_File = object( TSFile )
                        public { API... }
                            function Facility : longint ; virtual ;

                            function Version : longint ; virtual ;
                      end ; { TNative_File }


      { File System Device flags... }
const FSDF_Dirty = 1 ; { File structure partially/entirely cached }
      { FS Device Flags 2+ are reserved for future use }

function FS_Wildcard( const Wildcard, Filename : string ) : boolean ;

{$IFDEF Test}
procedure Test_FS ;
{$ENDIF}

implementation

uses
{$IFDEF DELPHI}
     { Borland... }
     Windows, { SECURITY_ATTRIBUTES }
     Forms, { Application }
  {$IFDEF TEST}
     Dialogs, { ShowMessage }
  {$ENDIF}

     { Conroy & Conroy Co. }
     Compatib, { Reallocmem }

{$ELSE}
     { Borland... }
  {$IFNDEF WINDOWS}
     Strings, { strpas }
  {$ENDIF}
{$ENDIF}

     { Conroy & Conroy Co. }
     _App,
     ASCIIDef, { LF }
     _Cache, { PSCache }
     Common, { TObject }
     CommonUt, { Edit }
     Dates, { Time_Seconds }
     DebugInt, { TText_Debugger }
     ErrMastr, { Error_Code }
     FSError,
     Num1s, { Num1 }
     O_S, { OS }
     Standard, { Get_Serial_Number }
     _Types, { Small_String }
     UStrings ; { Boolean_To_String }

{$IFDEF Verify}
var Global_Address_List : TInteger_List ;
    Global_Size_List : TInteger_List ;
{$ENDIF}

function FS_To_FH_Flags( Flags : integer ) : integer ;

var _Result : integer ;

begin
    _Result := 0 ;
    if( ( Flags and FSF_Fixed_Size ) <> 0 ) then
    begin
        _Result := FHF_NoExtend ;
    end ;
    if( ( Flags and FSF_Read_Only ) <> 0 ) then
    begin
        _Result := _Result or FHF_Read_Only ;
    end ;
    FS_To_FH_Flags := _Result ;
end ;


{$IFDEF DELPHI}
type tFS_EI = class( TStandard_Error_Interface )
                            public
                                function Get_Facility : longint ;
                                    override ; stdcall ;
                                function Get_Name : PChar ; override ; stdcall ;
                                function Translate_Error( Code : longint ) : PChar ;
                                    override ; stdcall ;
                                procedure Done ; override ; stdcall ;
                                function Facility_Version : longint ; override ;
                                function Version : longint ; override ;
                  end ;
     pFS_EI = tFS_EI ;
     tFH_EI = class( TStandard_Error_Interface )
                            public
                                function Get_Facility : longint ;
                                    override ; stdcall ;
                                function Get_Name : PChar ; override ; stdcall ;
                                function Translate_Error( Code : longint ) : PChar ;
                                    override ; stdcall ;
                                procedure Done ; override ; stdcall ;
                                function Facility_Version : longint ; override ;
                                function Version : longint ; override ;
                  end ;
     pFH_EI = tFH_EI ;
{$ELSE}
type tFS_EI = object( TStandard_Error_Interface )
                            public
                                function Get_Facility : longint ; virtual ;
                                function Get_Name : PChar ; virtual ;
                                function Translate_Error( Code : longint ) : PChar ;
                                    virtual ;
                                function Facility_Version : longint ; virtual ;
                                function Version : longint ; virtual ;
                  end ;
     pFS_EI = ^tFS_EI ;
     tFH_EI = object( TStandard_Error_Interface )
                            public
                                function Get_Facility : longint ; virtual ;
                                function Get_Name : PChar ; virtual ;
                                function Translate_Error( Code : longint ) : PChar ;
                                    virtual ;
                                function Facility_Version : longint ; virtual ;
                                function Version : longint ; virtual ;
                  end ;
     pFH_EI = ^tFH_EI ;
{$ENDIF}

{ tFS_EI method... }

function tFS_EI.Get_Facility : longint ;

begin
    Get_Facility := FSErr_Facility ;
end ;


const FS_Facility_Name : PChar = 'FS' ;

function tFS_EI.Get_Name : PChar ;

begin
    Get_Name := FS_Facility_Name ;
end ;


var FS_Error_Text : array[ 0..127 ] of char ;

function tFS_EI.Translate_Error( Code : longint ) : PChar ;

var FS_Error : string ;
    UEC : tUEC ;
{$IFDEF DELPHI}
    R : tUEC ;
{$ELSE}
    R : double ;
{$ENDIF}

begin
    if( Code < 0 ) then { DOS Error text }
    begin
        R := Error_Code( 1, -Code ) ;
        UEC := tUEC( R ) ;
        FS_Error := strpas( Get_Error_Text( UEC, -1 ) ) + #0 ;
        move( FS_Error[ 1 ], FS_Error_Text[ 0 ], length( FS_Error ) ) ;
        Translate_Error := @FS_Error_Text ;
        exit ;
    end ;
{$IFDEF DELPHI}
    if( Code > 65536 ) then
    begin
        Code := Code - 65536 ;
        FS_Error := strpas( Get_Error_Text( Error_Code( 0, -Code ), -1 ) ) + #0 ;
        move( FS_Error[ 1 ], FS_Error_Text[ 0 ], length( FS_Error ) ) ;
        Translate_Error := @FS_Error_Text ;
        exit ;
    end ;
{$ENDIF}
    case Code of
        FSErr_Success : FS_Error := 'Success' ;
        FSErr_FileNotOpen : FS_Error := 'I/O operation on unopened channel' ;
        FSErr_EndOfFile : FS_Error := 'Attempt to read past end of file' ;
        FSErr_FileNotFound : FS_Error := 'File lookup failed' ;
        FSErr_FileNameLength : FS_Error := 'Error in size of file name' ;
        FSErr_FileName : FS_Error := 'Error in file name contents' ;
        FSErr_DirectoryNotFound : FS_Error := 'Directory lookup failed' ;
        FSErr_InvalidPath : FS_Error := 'Path contained reference to non-directory' ;
        FSErr_NoRoomOnDevice : FS_Error := 'Device has no room' ;
        FSErr_ProtectionViolation : FS_Error := 'Protection violation' ;
        FSErr_DirectoryExists : FS_Error := 'Directory already exists' ;
        FSErr_FileNotEmpty : FS_Error := 'File is not empty' ;
        FSErr_DirectoryNotEmpty : FS_Error := 'Directory is not empty' ;
        FSErr_CorruptedFileStructure : FS_Error := 'File structure corruption found' ;
        FSErr_FileExists : FS_Error := 'File already exists' ;
        FSErr_IllegalOperationType : FS_Error := 'Illegal operation for file type' ;
        FSErr_LineTooLong : FS_Error := 'Text line is too long' ;
        FSErr_InvalidDevice : FS_Error := 'Invalid device' ;
        FSErr_Internal_Duplicate : FS_Error := 'Internal Duplicate' ;
        FSErr_Invalid_Open_Mode : FS_Error := 'Invalid Open Mode' ;
        FSErr_Unsupported_Operation : FS_Error := 'Unsupported File System' ;
        FSErr_Device_In_Use : FS_Error := 'Device In Use' ;
        FSErr_Unknown_Device_Type : FS_Error := 'Unknown Device Type' ;
        FSErr_Internal_Error_On_Add : FS_Error := 'Internal error on add' ;
        FSErr_Invalid_Record_Size : FS_Error := 'Invalid record size' ;
        FSErr_Internal_Corruption : FS_Error := 'Internal corruption' ;
        FSErr_Timeout : FS_Error := 'Timeout' ;
        FSErr_Invalid_Handle: FS_Error := 'Invalid handle' ;
        FSErr_Data_Format_Error: FS_Error := 'Data format error' ;
        FSErr_Read_Only: FS_Error := 'File system is read-only' ;
        else FS_Error := 'Unknown error number ' + num1( Code ) ;
    end ;
    FS_Error := FS_Error + #0 ;
    move( FS_Error[ 1 ], FS_Error_Text[ 0 ], length( FS_Error ) ) ;
    Translate_Error := @FS_Error_Text ;
end ; { tFS_EI.Translate_Error }


{$IFDEF DELPHI}
procedure tFS_EI.Done ;

begin
    { Done by unit }
end ;
{$ENDIF}


function tFS_EI.Facility_Version : longint ;

begin
    Facility_Version := 0 ;
end ;


function tFS_EI.Version : longint ;

begin
    Version := 10 ; { Compliant with UEH V1.0 specification }
end ;



{ tFH_EI methods... }

function tFH_EI.Get_Facility : longint ;

begin
    Get_Facility := FHErr_Facility ;
end ;


const FH_Facility_Name : PChar = 'FH' ;

function tFH_EI.Get_Name : PChar ;

begin
    Get_Name := FH_Facility_Name ;
end ;


var FH_Error_Text : array[ 0..127 ] of char ;

function tFH_EI.Translate_Error( Code : longint ) : PChar ;

var Err : string ;

begin
    case Code of
        FHErr_Success : Err := 'Success' ;
        FHErr_No_Room_On_Device : Err := 'No Room On Device' ;
        FHErr_Corrupted_File_Structure : Err := 'Corrupted File Structure' ;
        FHErr_Invalid_Pointer_Operation : Err := 'Invalid Pointer Operation' ;
        FHErr_Deallocation_Error : Err := 'Deallocation Error' ;
        FHErr_Size_Mismatch : Err := 'Size Mismatch' ;
        FHErr_Read_Out_Of_Range : Err := 'Read Out Of Range' ;
        FHErr_Read_Size_Error : Err := 'Read Size Error' ;
        FHErr_Write_Out_Of_Range : Err := 'Write Out Of Range' ;
        FHErr_Write_Size_Error : Err := 'Write Size Error' ;
        FHErr_Bad_Minimum : Err := 'Bad minimum' ;
        else Err := 'Unknown error code ' + num1( Code ) ;
    end ;
    Err := Err + #0 ;
    move( Err[ 1 ], FH_Error_Text[ 0 ], length( Err ) ) ;
    Translate_Error := @FH_Error_Text ;
end ; { tFH_EI.Translate_Error }


{$IFDEF DELPHI}
procedure tFH_EI.Done ;

begin
    { Done by unit }
end ;
{$ENDIF}


function tFH_EI.Facility_Version : longint ;

begin
    Facility_Version := 0 ;
end ;


function tFH_EI.Version : longint ;

begin
    Version := 10 ; { Compliant with UEH V1.0 specification }
end ;



{ TByte_Array_Debugger... }

{$IFDEF DELPHI}
type TByte_Array_Debugger = class( TText_Debugger )
                                private
                                    _Instance : PByte_Array ;
                                    _Size : integer ;
                                    S : string ;

                                public
                                    constructor Create( I : PByte_Array ; S : integer ) ;

                                    function Count : longint ; override ;

                                    function Child( Ordinal : longint ) : PDebug_Interface ;
                                        override ;
                            end ;
     PByte_Array_Debugger = TByte_Array_Debugger ;

constructor TByte_Array_Debugger.Create( I : PByte_Array ; S : integer ) ;

begin
    inherited Create ;

    _Instance := I ;
    _Size := S ;
end ;


{$ELSE}
type PByte_Array_Debugger = ^TByte_Array_Debugger ;
     TByte_Array_Debugger = object( TText_Debugger )
                                private
                                    _Instance : PByte_Array ;
                                    _Size : integer ;
                                    S : string[ 255 ] ;

                                public
                                    constructor Init( I : PByte_Array ;
                                        Size : integer ) ;

                                    function Count : longint ; virtual ;

                                    function Child( Ordinal : longint ) : PDebug_Interface ;
                                        virtual ;
                            end ;

constructor TByte_Array_Debugger.Init( I : PByte_Array ; Size : integer ) ;

begin
    inherited Init ;

    _Instance := I ;
    _Size := Size ;
end ;
{$ENDIF}


function TByte_Array_Debugger.Count : longint ;

begin
    Count := _Size ;
end ;


function TByte_Array_Debugger.Child( Ordinal : longint ) : PDebug_Interface ;

var R : PText_Debugger ;

begin
{$IFDEF DELPHI}
    R := TText_Debugger.Create ;
{$ELSE}
    new( R, Init ) ;
{$ENDIF}
    S := Num1( Ordinal ) + ' : ' + Num1( _Instance^[ Ordinal ] ) ;
{$IFDEF DELPHI}
    R.Title := PChar( S ) ;
{$ELSE}
    S := S + #0 ;
    R^.Set_Title( @S[ 1 ] ) ;
{$ENDIF}
    Child := PDebug_Interface( R ) ;
end ;



{ THeap_Statistics_Debugger... }

{$IFDEF DELPHI}
type THeap_Statistics_Debugger = class( TText_Debugger )
                                     private
                                         _Instance : PFH ;
                                         S : string ;

                                     public
                                         constructor Create( I : PFH ) ;

                                         function Count : longint ; override ;

                                         function Child( Ordinal : longint ) : PDebug_Interface ;
                                             override ;
                                 end ;
     PHeap_Statistics_Debugger = THeap_Statistics_Debugger ;

constructor THeap_Statistics_Debugger.Create( I : PFH ) ;

begin
    inherited Create ;

    _Instance := I ;
end ;


{$ELSE}
type PHeap_Statistics_Debugger = ^THeap_Statistics_Debugger ;
     THeap_Statistics_Debugger = object( TText_Debugger )
                          private
                              _Instance : PFH ;
                              S : string[ 255 ] ;

                          public
                              constructor Init( I : PFH ) ;

                              function Count : longint ; virtual ;

                              function Child( Ordinal : longint ) : PDebug_Interface ;
                                  virtual ;
                      end ;

constructor THeap_Statistics_Debugger.Init( I : PFH ) ;

begin
    inherited Init ;

    _Instance := I ;
end ;
{$ENDIF}


function THeap_Statistics_Debugger.Count : longint ;

begin
    Count := 21 ;
end ;


function THeap_Statistics_Debugger.Child( Ordinal : longint ) : PDebug_Interface ;

var R : PText_Debugger ;

begin
{$IFDEF DELPHI}
    R := TText_Debugger.Create ;
    case Ordinal of
        0 : S := 'Allocations = ' + Num1( _Instance.Statistics.Allocations ) ;
        1 : S := 'Bytes_Allocated = ' + Num1( _Instance.Statistics. Bytes_Allocated) ;
        2 : S := 'Bytes_Copied = ' + Num1( _Instance.Statistics.Bytes_Copied ) ;
        3 : S := 'Bytes_Deallocated = ' + Num1( _Instance.Statistics.Bytes_Deallocated ) ;
        4 : S := 'Bytes_Filled = ' + Num1( _Instance.Statistics.Bytes_Filled ) ;
        5 : S := 'Bytes_Read = ' + Num1( _Instance.Store_Statistics.Bytes_Read ) ;
        6 : S := 'Bytes_Written = ' + Num1( _Instance.Store_Statistics.Bytes_Written ) ;
        7 : S := 'Copies = ' + Num1( _Instance.Statistics.Copies ) ;
        8 : S := 'Deallocations = ' + Num1( _Instance.Statistics.Deallocations ) ;
        9 : S := 'Error_Count = ' + Num1( _Instance.Store_Statistics.Error_Count ) ;
        10 : S := 'Extensions = ' + Num1( _Instance.Statistics.Extensions ) ;
        11 : S := 'Fills = ' + Num1( _Instance.Statistics.Fills ) ;
        12 : S := 'Index_Cache_Hits = ' + Num1( _Instance.FH_Statistics.Index_Cache_Hits ) ;
        13 : S := 'Index_Reads = ' + Num1( _Instance.FH_Statistics.Index_Reads ) ;
        14 : S := 'Index_Writes = ' + Num1( _Instance.FH_Statistics.Index_Writes ) ;
        15 : S := 'Reads = ' + Num1( _Instance.Store_Statistics.Reads ) ;
        16 : S := 'Reallocations = ' + Num1( _Instance.Statistics.Reallocations ) ;
        17 : S := 'Reallocations_In_Place = ' + Num1( _Instance.Statistics.Reallocations_In_Place ) ;
        18 : S := 'Root_Reads = ' + Num1( _Instance.Statistics.Root_Reads ) ;
        19 : S := 'Root_Writes = ' + Num1( _Instance.Statistics.Root_Writes ) ;
        20 : S := 'Writes = ' + Num1( _Instance.Store_Statistics.Writes ) ;
    end ;
    R.Title := PChar( S ) ;
{$ELSE}
    new( R, Init ) ;
    S := S + #0 ;
    R^.Set_Title( @S[ 1 ] ) ;
{$ENDIF}
    Child := PDebug_Interface( R ) ;
end ;



{ Internal utility functions... }
procedure Add_To_FH_List( X : pointer ) ; forward ;
procedure Remove_From_FH_List( X : pointer ) ; forward ;
procedure Flush_FH( T : pointer ; const X : string ) ; forward ;


{ Public API functions... }

function FS_Wildcard( const Wildcard, Filename : string ) : boolean ;
{ This function is used by the File_System to do wildcard comparisons on
  filenames.  if Filename matches Wildcard, the function returns True,
  otherwise, it returns False.  Note: this function is case-sensitive. }

    function Match_Wildcard( Ref_Spec, In_Spec : string ) : Boolean ;
    { Return True if value in In_Spec matches wildcard in Ref_Spec. }

    var Dummy, Dummy1 : Integer ;
        Ref, Inp : string ; { Work variables }

    begin
        if( ( Ref_Spec = '\...' ) or ( Ref_Spec = '\...\' ) ) then
        begin
            Ref_Spec := '\*' ; { Handle special directory wildcard }
        end ;

        if( pos( '.', Ref_Spec ) = 0 ) then { Implied extension }
        begin
            Ref_Spec := Ref_Spec + '.' ;
        end ;
        if( pos( '.', In_Spec ) = 0 ) then { Implied extension }
        begin
            In_Spec := In_Spec + '.' ;
        end ;

        Dummy := Pos( '*', Ref_Spec + '*' ) ;
        Ref := Copy( Ref_Spec, 1, Dummy - 1 ) ;
        Ref_Spec := Copy( Ref_Spec, Dummy, 254 ) ;

        Dummy := Pos( '*', In_Spec + '*' ) ;
        Inp := Copy( In_Spec, 1, Dummy - 1 ) ;
        In_Spec := Copy( In_Spec, Dummy, 254 ) ;

        { Search forward from start for matches... }
        if( Length( Inp ) < Length( Ref ) ) then { Quick check and exit }
        begin
            Match_Wildcard := False ;
            Exit ;
        end ;
        Dummy := 1 ;
        while( Dummy <= Length( Ref ) ) do
        begin
            if(
                ( Ref[ Dummy ] <> '?' )
                and
                ( Ref[ Dummy ] <> Inp[ Dummy ] )
              ) then
            begin
                Match_Wildcard := False ;
                Exit ;
            end ;
            Dummy := Dummy + 1 ; { Try next character }
        end ;
        Inp := Copy( Inp, Length( Ref ) + 1, 254 ) ; { Get remaining characters }

        if( ( Length( Inp ) > 0 ) and ( copy( Ref_Spec, 1, 1 ) <> '*' ) ) then
        begin
            Match_Wildcard := False ;
            Exit ;
        end ;

        { Search backward from end for matches... }
        Dummy := length( Ref_Spec ) ;
        while( ( Dummy > 0 ) and ( Ref_Spec[ Dummy ] <> '*' ) ) do
        begin
            dec( Dummy ) ;
        end ;
        Ref := Copy( Ref_Spec, Dummy + 1 , 254 ) ;
        Ref_Spec := Copy( Ref_Spec, 1, Dummy ) ;

        if( length( In_Spec ) > 0 ) then
        begin
            Dummy := length( In_Spec ) ;
            while( ( Dummy > 0 ) and ( In_Spec[ Dummy ] <> '*' ) ) do
            begin
                dec( Dummy ) ;
            end ;
            Inp := Copy( In_Spec, Dummy + 1, 254 ) ;
            In_Spec := Copy( In_Spec, 1, Dummy ) ;
        end ;

        if( ( Length( Inp ) > 0 ) and ( copy( Ref_Spec, 1, 1 ) <> '*' ) ) then
        begin
            Match_Wildcard := False ;
            Exit ;
        end ;

        Dummy := length( Ref ) ;
        Dummy1 := length( Inp ) ;
        while( ( Dummy > 0 ) and ( Dummy1 > 0 ) ) do
        begin
            if(
                ( Ref[ Dummy ] <> '?' )
                and
                ( Ref[ Dummy ] <> Inp[ Dummy1 ] )
              ) then
            begin
                Match_Wildcard := False ;
                Exit ;
            end ;
            Dummy := Dummy - 1 ; { Try next character }
            Dummy1 := Dummy1 - 1 ;
        end ;
        Inp := Copy( Inp, 1, Dummy1 ) ; { Get remaining characters }

        if( ( Length( Inp ) > 0 ) and ( Ref_Spec <> '*' ) ) then
        begin
            Match_Wildcard := False ;
            Exit ;
        end ;

        Match_Wildcard := True ; { Everything matched }
    end ; { FS_Wildcard.Match_Wildcard }


    function Match_Dir( Ref_Spec, In_Spec : string ) : Boolean ;

    label Loop ;

    var Dummy : Integer ;
        Ref, Inp : string ; {Work variables}

    begin
Loop:
        if(
            ( ( Length( Ref_Spec ) = 0 )
            or
            ( Ref_Spec = '\...\' ) ) and ( Length( In_Spec ) = 0 )
          ) then
        begin
            Match_Dir := True ;
            Exit ;
        end ;
        if( Ref_Spec = '\...\' ) then { Match everything else }
        begin
            Match_Dir := True ;
            Exit ;
        end ;
        Dummy := Pos( '\', Copy( Ref_Spec, 2, 254 ) + '\' ) + 1 ;
        Ref := Copy( Ref_Spec, 1, Dummy - 1 ) ; { Current reference directory to use }
        Ref_Spec := Copy( Ref_Spec, Dummy, 254 ) ;

        Dummy := Pos( '\', Copy( In_Spec, 2, 254 ) + '\' ) + 1 ;
        Inp := Copy( In_Spec, 1, Dummy - 1 ) ; { Current input directory to use }
        In_Spec := Copy( In_Spec, Dummy, 254 ) ;

        if not Match_Wildcard( Ref, Inp ) then
        begin
            Match_Dir := False ;
            Exit ;
        end ;
        goto Loop ; { Check next directory in path }
    end ;


var Dummy : Integer ; { Dummy variable }
    Ref_Dev, In_Dev : string[ 2 ] ; { Input/output device }
    Ref_Dir, In_Dir : string ; { Input/output directory path }
    Ref_Name, In_Name : string ; { Input/output filename }

begin { FS_Wildcard }
    { Break filename and wildcard into component pieces... }
    if( ( Wildcard = '*' ) or ( Wildcard = '*.*' ) ) then
    begin
        FS_Wildcard := True ;
        exit ;
    end ;
    Dummy := length( Filename ) ;
    while( ( Dummy > 0 ) and ( Filename[ Dummy ] <> '\' ) ) do
    begin
        Dummy := Dummy - 1 ;
    end ;
    In_Dir := copy( Filename, 1, Dummy ) ;
    In_Name := copy( Filename, Dummy + 1, 254 ) ;
    Dummy := length( Wildcard ) ;
    while( ( Dummy > 0 ) and ( Wildcard[ Dummy ] <> '\' ) ) do
    begin
        Dummy := Dummy - 1 ;
    end ;
    Ref_Dir := copy( Wildcard, 1, Dummy ) ;
    Ref_Name := copy( Wildcard, Dummy + 1, 254 ) ;
    Dummy := Pos( ':', In_Dir ) ; { Locate device in filename }
    In_Dev := Copy( In_Dir, 1, Dummy ) ;
    In_Dir := Copy( In_Dir, Dummy + 1, 254 ) ;
    Dummy := Pos( ':', Ref_Dir ) ; { Locate device in wildcard }
    Ref_Dev := Copy( Ref_Dir, 1, Dummy ) ;
    Ref_Dir := Copy( Ref_Dir, Dummy + 1, 254 ) ;

    if( ( Ref_Dev <> '*:' ) and ( Ref_Dev <> In_Dev ) ) then { Device mismatch }
    begin
        FS_Wildcard := False ;
        Exit ;
    end ;

    if( not Match_Dir( Ref_Dir, In_Dir ) ) then
    begin
        FS_Wildcard := False ;
        Exit ;
    end ;
    if( not Match_Wildcard( Ref_Name, In_Name ) ) then
    begin
        FS_Wildcard := False ;
        Exit ;
    end ;

    FS_Wildcard := True ; { Everything matched }
end ; { FS_Wildcard }



var FS_Buffer : array[ 0..Entry_Size - 1 ] of byte ;

{ TXFile_System methods... }

constructor TXFile_System.Init( const Name : string ; _Flags : integer ;
    var Res : TUEC ) ;

begin
    inherited Init ;

    fillchar( Res, sizeof( Res ), 0 ) ;
    Flags := _Flags ;
    Common_Init ;
    if( length( Name ) <> 0 ) then
    begin
        Set_Device( Name, Flags, Res ) ;
    end ;
end ; { TXFile_System.Init }


constructor TXFile_System.Init_From_Store( Store : PSStore ; _Flags : integer ;
    var Res : TUEC ) ;

var FH_Flags : integer ;

begin
    inherited Init ;

    Flags := _Flags ;
    Common_Init ;

    if( File_Heap <> nil ) then
    begin
        dispose( File_Heap, Done ) ;
    end ;

    { Convert from FSF flags to FH flags... }
    if( ( _Flags and FSF_Fixed_Size ) <> 0 ) then
    begin
        FH_Flags := FHF_NoExtend ;
    end else
    begin
        FH_Flags := 0 ;
    end ;
    if( ( _Flags and FSF_Read_Only ) <> 0 ) then
    begin
        FH_Flags := _Flags or FHF_Read_Only ;
    end ;
    if( ( _Flags and FSF_NoOwnStore ) <> 0 ) then
    begin
        FH_Flags := _Flags or FHF_NoClose ;
    end ;

    File_Heap := new( PFH, Init_On_Store( sizeof( Boot ), FH_Flags, Store, Res, Memory_Heap ) ) ;
    File_Heap^.Initialize ;
    File_Heap^.Own_Allocator := False ;
    if( Res.Code <> 0 ) then
    begin
        dispose( Store, Done ) ;
        exit ;
    end ;

    Connect_Store_Heap( Res ) ;

    { Reopen store heap with appropriate blocking size... }
    if( Boot.Physical_Block_Size <> sizeof( Boot ) ) then
    begin
        dispose( File_Heap, Done ) ;
        File_Heap := new( PFH, Init_On_Store( Boot.Physical_Block_Size, FH_Flags, Store, Res, Memory_Heap ) ) ;
        File_Heap^.Initialize ;
        File_Heap^.Own_Allocator := False ;
    end ;
end ;


destructor TXFile_System.Done ;

var Loop : integer ;

begin
    { Close all files... }
    for Loop := File_List.Count downto 1 do
    begin
        Close( Loop ) ;
    end ;

    if( File_Heap <> nil ) then
    begin
        dispose( File_Heap, Done ) ;
        File_heap := nil ;
    end ;

    dispose( Memory_Heap, Done ) ;
    Memory_Heap := nil ;

    inherited Done ;
end ;


procedure TXFile_System.Common_Init ;

begin
    Memory_Heap := new( pSHeap, Init ) ;
    Serial_Number := Get_Serial_Number ;
    First_Lock := True ;
    Locked := False ;
    File_List.Init( 0, 4 ) ;
    File_Heap := nil ;
    Cached_Entries := 0 ;
{$IFDEF DELPHI}
    System.Initialize( Cached_File_Name1 ) ;
    System.Initialize( Cached_Path1 ) ;
    System.Initialize( Cached_File_Name2 ) ;
    System.Initialize( Cached_Path2 ) ;
{$ENDIF}
{$IFDEF Debug}
    Sanity := 'TXFile_System' ;
{$ENDIF}
end ;


const Lock_Wait_Period = 3000 ; { 30 seconds }

{ Locks the file system for synchronous I/O.  Returns True if this process was
  the previously locking process. }
function TXFile_System.Lock( Nam : string ) : boolean ;

    function _Lock : boolean ;

    var Boot : FS_Boot_Record ;
        Res : TUEC ;

        procedure Lock_Header ;

        begin
            { Set lock... }
            Cached_FS_Header.Lock_PID := PID ;
            Cached_FS_Header.Lock_Instance := HInstance ;
            Cached_FS_Header.Lock_Context := Serial_Number ;
            Cached_FS_Header.Lock_Time := Time_Seconds ;
            File_Heap^.Write_Heap( Boot.Header, sizeof( Cached_FS_Header ), Cached_FS_Header ) ;
        end ;

        { Wait for the lock to be released (return false) or until we time-out
          (return true). }
        function Wait : boolean ;

        begin
            if( Time_Seconds < Cached_FS_Header.Lock_Time ) then { Went past midnight }
            begin
                Cached_FS_Header.Lock_Time := Cached_FS_Header.Lock_Time - 24 * 60 * 6000 ;
            end ;
            while( Time_Seconds < Cached_FS_Header.Lock_Time + Lock_Wait_Period ) do
            begin
{$IFDEF DELPHI}
                Application.ProcessMessages ; { Allow message processing to happen }
{$ENDIF}
                File_Heap^.Read_Heap( Boot.Header, sizeof( Cached_FS_Header ), Cached_FS_Header ) ;
                Lock_Header ;

                { Now check for success... }
                File_Heap^.Read_Heap( Boot.Header, sizeof( Cached_FS_Header ), Cached_FS_Header ) ;
                if(
                    ( Cached_FS_Header.Lock_PID = PID )
                    and
{$WARNINGS OFF}
                    ( Cached_FS_Header.Lock_Instance = HInstance )
{$WARNINGS ON}
                    and
                    ( Cached_FS_Header.Lock_Context = Serial_Number )
                  ) then
                begin
                    Wait := False ; { No problem }
                    exit ;
                end ;
            end ;

            { Lock timed-out }

            { We timed-out and couldn't force the lock... }
            Res.Facility := FSErr_Facility ;
            Res.Code := FSErr_Timeout ;
            IO_Error := Res ;
            Wait := True ;
        end ;

    label Try_Again ;

    begin { TXFile_System.Lock._Lock }
        _Lock := False ;
        Nam := Edit( Nam, 2 or 4 ) ;
        if( length( Nam ) = 0 ) then { Null filename is invalid }
        begin
            Res.Code := FSErr_FileName ; { Invalid file name }
            Res.Facility := FSErr_Facility ;
            IO_Error := Res ;
            exit ;
        end ;

        { Find DOS file in list... }
        File_Heap^.Get_Root( Boot ) ;
        if(
            ( Boot.Physical_Block_Size = 0 )
            or
            ( Boot.Header < Boot.Physical_Block_Size )
          ) then
        begin
            Res.Facility := FSErr_Facility ; { TEST: Untested }
            Res.Code := FSErr_CorruptedFileStructure ;
            IO_Error := Res ;
        end else
        begin
            File_Heap^.Read_Heap( Boot.Header, sizeof( Cached_FS_Header ), Cached_FS_Header ) ;
            if( Cached_FS_Header.Default_Cluster_Size = 0 ) then
            begin
                Res.Facility := FSErr_Facility ; { TEST: Untested }
                Res.Code := FSErr_CorruptedFileStructure ;
                IO_Error := Res ;
            end else
            begin
                Res := File_Heap^.Last_Error ;
            end ;
        end ;

        if( Res.Code <> 0 ) then
        begin
            IO_Error := Res ;
            exit ;
        end ;

{$IFDEF MUTEX}
        S := Replace( Path, '\', '*', 0 ) ;
        Lock_Mutex := CreateMutex( nil, True, PChar( S ) ) ;
        while( GetLastError = ERROR_ALREADY_EXISTS ) do
        begin
            CloseHandle( Lock_Mutex ) ;
            Sleep( 0 ) ;
            Application.ProcessMessages ;
            Lock_Mutex := CreateMutex( nil, True, PChar( S ) ) ;
        end ;
        Lock_Header ;
        Lock := True ;
{$ELSE}
Try_Again:
        File_Heap^.Get_Root( Boot ) ;
        File_Heap^.Read_Heap( Boot.Header, sizeof( Cached_FS_Header ), Cached_FS_Header ) ;
        if(
            ( Cached_FS_Header.Lock_PID = 0 )
            or
            (
              ( Cached_FS_Header.Lock_PID = PID )
              and
{$WARNINGS OFF}
              ( Cached_FS_Header.Lock_Instance = HInstance )
{$WARNINGS ON}
              and
              ( Cached_FS_Header.Lock_Context = Serial_Number )
            )
          ) then
        begin { We can lock it }
            Lock_Header ;

            { Now re-read the header to make sure we succeeded... }
            File_Heap^.Read_Heap( Boot.Header, sizeof( Cached_FS_Header ), Cached_FS_Header ) ;
            if(
                ( Cached_FS_Header.Lock_PID <> PID )
                or
{$WARNINGS OFF}
                ( Cached_FS_Header.Lock_Instance <> HInstance )
{$WARNINGS ON}
                or
                ( Cached_FS_Header.Lock_Context <> Serial_Number )
              ) then { Failed to lock the file }
            begin
                if( Wait ) then
                begin
                    exit ;
                end ;
                goto Try_Again ;
            end ;

            { Lock successful... }
            if(
                ( Cached_FS_Header.Last_Lock_PID <> PID )
                or
{$WARNINGS OFF}
                ( Cached_FS_Header.Last_Lock_Instance <> HInstance )
{$WARNINGS ON}
                or
                ( Cached_FS_Header.Last_Lock_Context <> Serial_Number )
                or
                First_Lock
              ) then
            begin
                Cached_FS_Header.Last_Lock_PID := PID ;
                Cached_FS_Header.Last_Lock_Instance := HInstance ;
                Cached_FS_Header.Last_Lock_Context := Serial_Number ;
                File_Heap^.Write_Heap( Boot.Header, sizeof( Cached_FS_Header ), Cached_FS_Header ) ;
            end else
            begin
                _Lock := True ;
            end ;
            exit ;
        end ;
        if( not Wait ) then
        begin
            goto Try_Again ;
        end ;
{$ENDIF}
    end ; { TXFile_System.Lock._Lock }

begin { TXFile_System.Lock }
    Lock := _Lock ;
    if( ( Flags and FSF_Read_Only ) <> 0 ) then
    begin
        Locked := False ;
        exit ;
    end ;
    Locked := ( IO_Error.Code = 0 ) ;
    First_Lock := False ;
end ; { TXFile_System.Lock }


procedure TXFile_System.Unlock( Nam : string ) ;

var _Boot : FS_Boot_Record ;
    Res : TUEC ;
    _Time : longint ;

begin
    if( not Locked ) then
    begin
        exit ;
    end ;
    Nam := Edit( Nam, 2 or 4 ) ;
    if( length( Nam ) = 0 ) then { Null filename is invalid }
    begin
        Res.Code := FSErr_FileName ; { Invalid file name }
        Res.Facility := FSErr_Facility ;
        IO_Error := Res ;
        exit ;
    end ;

    { Find DOS file in list... }
    File_Heap^.Get_Root( _Boot ) ;
    if(
        ( _Boot.Physical_Block_Size = 0 )
        or
        ( _Boot.Header < _Boot.Physical_Block_Size )
      ) then
    begin
        Res.Facility := FSErr_Facility ; { TEST: Untested }
        Res.Code := FSErr_CorruptedFileStructure ;
        IO_Error := Res ;
    end else
    begin
        File_Heap^.Read_Heap( _Boot.Header, sizeof( Cached_FS_Header ), Cached_FS_Header ) ;
        if( Cached_FS_Header.Default_Cluster_Size = 0 ) then
        begin
            Res.Facility := FSErr_Facility ; { TEST: Untested }
            Res.Code := FSErr_CorruptedFileStructure ;
            IO_Error := Res ;
        end ;
    end ;

    { Unlock the file... }
{$IFDEF MUTEX}
    if( Lock_Mutex <> 0 ) then
    begin
        CloseHandle( Lock_Mutex ) ;
        Lock_Mutex := 0 ;
    end ;
{$ELSE}
    _Time := Cached_FS_Header.Lock_Time ;
    if( Time_Seconds < _Time ) then { Went past midnight }
    begin
        _Time := _Time - 24 * 60 * 6000 ;
    end ;
    if( Time_Seconds >= _Time + Lock_Wait_Period ) then { Lock expired }
    begin
        File_Heap^.Read_Heap( _Boot.Header, sizeof( Cached_FS_Header ), Cached_FS_Header ) ;
    end ;
    if( Cached_FS_Header.Lock_PID = PID ) then
    begin
        Cached_FS_Header.Lock_PID := 0 ;
        File_Heap^.Write_Heap( _Boot.Header, sizeof( Cached_FS_Header ), Cached_FS_Header ) ;
    end ;
{$ENDIF}
    Locked := False ;
end ; { TXFile_System.Unlock }


procedure TXFile_System.Erase( const Name : string ) ;

var Entry, Dir_Entry : TDirectory_Entry ;
    File_Entry, Directory_Entry : TStore_Address ;
    _Name : string ;

begin
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    if( ( Flags and FSF_Read_Only ) <> 0 ) then
    begin
        Set_Error( FSErr_Read_Only ) ;
        exit ;
    end ;
    _Name := Edit( Name, 4 or 8 or 32 or 128 ) ;
    if( length( _Name ) = 0 ) then { Not erasing an open file }
    begin
        Set_Error( FSErr_FileNameLength ) ;
        exit ;
    end ;
    Directory_Entry := 0 ;
    File_Entry := Find_File( _Name, Directory_Entry ) ;
    if( IO_Error.Code <> 0 ) then { Some error }
    begin
        exit ;
    end ;
    if( File_Entry = 0 ) then { Directory exists, but not file }
    begin
        Set_Error( FSErr_FileNameLength ) ;
        exit ;
    end ;
    Read_Entry( File_Entry, Entry ) ;
    if( ( Entry.Attributes and FSFF_Directory ) <> 0 ) then { Deleting a directory }
    begin
        Read_Entry( Entry.Allocation_Info, Dir_Entry ) ;
        if( Dir_Entry.Right_Pointer <> 0 ) then { Directory has something in it }
        begin
            Set_Error( FSErr_DirectoryNotEmpty ) ;
            exit ;
        end ;
        Clear_Cache ;
        {TODO: close any instances of this file, if opened - or else implement deletion flag}
        Tree_Delete( File_Entry ) ; { Remove file from directory }
    end else { Normal file deletion }
    begin
        Delete_File( File_Entry ) ;
    end ;
end ; { TXFile_System.Erase }


function TXFile_System.File_Name( Handle : integer ) : string ;

var File_Rec : PNative_File_Record ;

begin
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    if( Handle = 0 ) then
    begin
        Set_Error( FSErr_FileNotOpen ) ;
        exit ;
    end ;

    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    File_Name := File_Rec^.Name ;
end ; { TXFile_System.File_Name }


function TXFile_System.First_Directory( const Name : string ) : integer ;

begin
    First_Directory := pos( '\', Name + '\' ) ;
end ;


procedure TXFile_System.Flush( Handle : integer ) ;

var File_Rec : PNative_File_Record ;

begin
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    if( Handle = 0 ) then
    begin
        Set_Error( FSErr_FileNotOpen ) ;
        exit ;
    end ;

    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    Set_Error( 0 ) ;
    { This routine does nothing since native files don't need flushing. }
end ;


function TXFile_System.Valid_Filename( var S : string ) : boolean ;
{ Return true if a valid filename or false otherwise. }

begin
    if( ( length( S ) = 0 ) or ( length( S ) > FS_Maximum_File_Name ) ) then
    begin
        Valid_Filename := false ; { Bad filename size }
        Set_Error( FSErr_FileNameLength ) ;
        exit ;
    end ;
    if(
        ( S <> Edit( S, 4 or 8 or 16 or 128 ) )
        or
        ( pos( '*', S ) > 0 )
        or
        ( pos( '?', S ) > 0 )
        or
        ( pos( '\', S ) > 0 )
        or
        ( S = '...' )
      ) then
    begin
        Valid_Filename := false ; { Found a bad character }
        Set_Error( FSErr_FileName ) ;
        exit ;
    end ;
    Valid_Filename := true ; { File name is valid }
end ;


function TXFile_System.Path : string ;

begin
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    Path := File_Heap^.Path ;
end ;


procedure TXFile_System.Set_Device( Name : string ; _Flags : integer ;
    var Res : TUEC ) ;

var FH_Flags : integer ;

begin
    if( File_Heap <> nil ) then
    begin
        dispose( File_Heap, Done ) ;
    end ;

    if( ( _Flags and FSF_Fixed_Size ) <> 0 ) then
    begin
        FH_Flags := FHF_NoExtend ;
    end else
    begin
        FH_Flags := 0 ;
    end ;
    if( ( _Flags and FSF_Read_Only ) <> 0 ) then
    begin
        FH_Flags := FH_Flags or FHF_Read_Only ;
    end ;
    File_Heap := new( PFH, Init( sizeof( Boot ), FH_Flags, Name, Res, Memory_Heap ) ) ;
    File_Heap^.Initialize ;
    File_Heap^.Own_Allocator := False ;
    if( Res.Code <> 0 ) then
    begin
        dispose( File_Heap, Done ) ;
        File_Heap := nil ;
        exit ;
    end ;

    Connect_Store_Heap( Res ) ;

    { Reopen store heap with appropriate blocking size... }
    if( Boot.Physical_Block_Size <> sizeof( Boot ) ) then
    begin
        dispose( File_Heap, Done ) ;
        File_Heap := new( PFH, Init( Boot.Physical_Block_Size, FH_Flags, Name, Res, Memory_Heap ) ) ;
        File_Heap^.Initialize ;
        File_Heap^.Own_Allocator := False ;
    end ;
end ; { TXFile_System.Set_Device }


procedure TXFile_System.Connect_Store_Heap( var Res : TUEC ) ;

var _FS_Header : FS_Header_Record ;

begin
    File_Heap^.Get_Root( Boot ) ;
    if(
        ( Boot.Physical_Block_Size = 0 )
        or
        ( Boot.Header < Boot.Physical_Block_Size )
      ) then
    begin
        Res.Facility := FSErr_Facility ; { TEST: Untested }
        Res.Code := FSErr_CorruptedFileStructure ;
        exit ;
    end else
    begin
        Read_Header( File_Heap, Boot.Header, _FS_Header ) ;
        if( _FS_Header.Default_Cluster_Size = 0 ) then
        begin
            Res.Facility := FSErr_Facility ; { TEST: Untested }
            Res.Code := FSErr_CorruptedFileStructure ;
            exit ;
        end ;
    end ;

    File_Heap^.Read_Heap( Boot.Header, sizeof( Cached_FS_Header ), Cached_FS_Header ) ;
    Root := Cached_FS_Header.Root ;
    Clear_Cache ;
end ; { TXFile_System.Connect_Store_Heap }


procedure TXFile_System.Set_Cache( Limit : longint ) ;

var Cache : PSCache ;

begin
    if( Limit = 0 ) then { Disable caching }
    begin
        File_Heap^.Set_Cache( nil ) ;
        exit ;
    end ;
    if( File_Heap^.Get_Cache = nil ) then
    begin
        new( Cache, Init ) ;
        Cache^.Initialize ;
        File_Heap^.Set_Cache( Cache ) ;
    end ;
    if( Limit > 0 ) then
    begin
        Cache := File_Heap^.Get_Cache ;
        Cache^.Set_Max_Size( Limit ) ;
    end ;
end ;


function TXFile_System.Allocate_Entry : TStore_Address ;

var _Result : TStore_Address ;

begin
    _Result := File_Heap^.Allocate( Entry_Size ) ;
    if( _Result <> 0 ) then
    begin
        File_Heap^.Write_Heap( _Result, Entry_Size, FS_Buffer ) ;
    end ;
    Allocate_Entry := _Result ;
end ;


procedure TXFile_System.Clear_Cache ;

begin
    Clear_File_Lookup_Cache( Root ) ;
end ;


procedure TXFile_System.Clear_File_Lookup_Cache( _Root : TStore_Address ) ;
{ Clear cached file lookup information. }

begin
    Cached_Entries := 0 ;
end ;


function TXFile_System.FindFirst( Nam : string ; Attributes : longint ;
    var Find_Record : TFind_Record ) : longint ;

label Backup, Loop ;

var Current, Previous : longint ;
    Directory_Entry : TStore_Address ;
    Dummy : integer ;
    Entry : TDirectory_Entry ;
    _FS_Header : FS_Header_Record ;
    _Path : string ;
    Work : string ;

begin { TXFile_System.FindFirst }
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        FindFirst := FSErr_FileNotFound ; { Lookup failure }
        exit ;
    end ;
    Set_Error( 0 ) ;
    Nam := Edit( Nam, 4 or 8 or 32 or 128 ) ;

    { Get directory portion (won't be a wildcard) }
    Dummy := length( Nam ) ;
    while( ( Dummy > 0 ) and ( Nam[ Dummy ] <> '\' ) ) do
    begin
        Dummy := Dummy - 1 ;
    end ;

    { Get root node of directory tree... }
    Work := copy( Nam, 1, Dummy ) ;
    if( Work = '\' ) then
    begin
        Read_Header( File_Heap, Boot.Header, _FS_Header ) ;
        Directory_Entry := _FS_Header.Root ;
    end else
    begin
        Find_File( Work, Directory_Entry ) ;
        if( Directory_Entry = 0 ) then
        begin
            FindFirst := FSErr_FileNotFound ; { Lookup failure }
            exit ;
        end ;
    end ;
    Read_Entry( Directory_Entry, Entry ) ; { Directory file entry }
    Read_Entry( Entry.Allocation_Info, Entry ) ; { Read directory root }
    _Path := copy( Nam, 1, Dummy - 1 ) ;
    Nam := copy( Nam, Dummy + 1, 254 ) ;
    Find_Record.Wildcard_Context := Nam ;
    Find_Record.Wildcard_Attributes := Attributes ;

Loop:
    Current := Entry.Right_Pointer ; { Get first file entry for directory }
    if( Current <> 0 ) then { Not an empty directory }
    begin
        Read_Entry( Current, Entry ) ;
        while( Entry.Left_Pointer <> 0 ) do { Start at first file }
        begin
            Current := Entry.Left_Pointer ;
            Read_Entry( Current, Entry ) ;
        end ;
    end ;

    { Now look for a matching file }
    while( Current <> 0 ) do { Until done }
    begin
        if( FS_Wildcard( Nam, Edit( Entry.Name, 4 or 8 or 16 or 32 or 128 ) ) ) then { Wildcard match }
        begin
            if(
                ( ( Entry.Attributes and FSFF_Directory ) = 0 )
                or
                ( ( Attributes and FSFF_Directory ) <> 0 )
              ) then { Only match directories if requested to }
            begin
                FindFirst := 0 ;
                move( Entry, Find_Record.Data, sizeof( Entry ) ) ;
                Find_Record.Found_Name := Entry.Name ;
                Find_Record.Found_Attributes := Entry.Attributes ;
                Find_Record.Last_Name := Uppercase( _Path + '\' + Entry.Name ) ;
                exit ;
            end ;
        end ;
        if( Entry.Right_Pointer = 0 ) then { Need to backup }
        begin
Backup:
            Previous := Current ;
            Current := Entry.Back_Pointer ;
            if( Current <> 0 ) then { Still files to process }
            begin
                Read_Entry( Current, Entry ) ;
                if( Previous = Entry.Right_Pointer ) then
                { Coming back from right branch }
                begin
                    goto Backup ;
                end ; { Otherwise check this node }
            end ;
        end else
        begin
            goto Loop ; { Go right }
        end ;
    end ;

    { No matching file was found if we get here }
    FindFirst := FSErr_FileNotFound ; { Lookup failure }
end ; { TXFile_System.FindFirst }


function TXFile_System.FindNext( var Find_Record : TFind_Record ) : longint ;

    function _Valid_Filename( X : string ) : boolean ;

    begin
        Result := Valid_Filename( X ) ;
    end ;

label Backup, Loop, Restart ;

var Current, Previous : TStore_Address ;
    Directory_Entry : TStore_Address ; { Address of directory that we are searching }
    Dummy : integer ;
    Entry : TDirectory_Entry ;
    Previous_Name : string ;
    Previous_Current : TStore_Address ;
    S : string ;

begin { TXFile_System.FindNext }
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        FindNext := FSErr_FileNotFound ; { Lookup failure }
        exit ;
    end ;
    Set_Error( 0 ) ;
    S := uppercase( File_Heap^.Path ) ;
    if( S = copy( Find_Record.Last_Name, 1, length( S ) ) ) then
    begin
        S := '' ;
    end ;
    if( length( S ) > 0 ) then
    begin
        S := S + '\' ;
    end ;

Restart:
    { Get root node of directory tree... }
    Current := Find_File( Find_Record.Last_Name, Directory_Entry ) ;
    if( Current = 0 ) then { File has been deleted or renamed since last lookup }
    begin
        FindNext := FSErr_FileNotFound ; { Lookup failure }
        exit ;
    end ;
    Read_Entry( Current, Entry ) ;

Loop:
    Previous_Current := Current ;
    Previous_Name := Uppercase( Entry.Name ) ;
    if( Entry.Right_Pointer <> 0 ) then { At least one more file }
    begin
        Current := Entry.Right_Pointer ; { Get next file entry }
        Read_Entry( Current, Entry ) ;
        if( not File_Heap^.Read_Only ) then
        begin
            if( Entry.Back_Pointer <> Previous_Current ) then { Bad back pointer }
            begin
                { Fix the directory entry... }
                Fix_Back_Pointer( Previous_Current, Current, Entry ) ;
                if( Entry.Back_Pointer <> Previous_Current ) then { Must have been cross-linked }
                begin
                    goto Restart ;
                end ;
            end ;
            if( not _Valid_Filename( Entry.Name ) ) then
            begin
                Delete_File( Current ) ;
                goto Restart ;
            end ;
            if( Previous_Name > Uppercase( Entry.Name ) ) then { Right branch is BEFORE node }
            begin
                { Correct the directory structure... }
                Repair_Tree( Directory_Entry, Current ) ;
                goto Restart ; { ...and try again }
            end ;
        end ; { if( not File_Heap^.Read_Only ) }

        while( Entry.Left_Pointer <> 0 ) do { Start at first file on this branch }
        begin
            Previous_Current := Current ;
            Previous_Name := Uppercase( Entry.Name ) ;
            Current := Entry.Left_Pointer ;
            Read_Entry( Current, Entry ) ;
            if( not File_Heap^.Read_Only ) then
            begin
                if( Entry.Back_Pointer <> Previous_Current ) then { Bad back pointer }
                begin
                    { Fix the directory entry... }
                    Fix_Back_Pointer( Previous_Current, Current, Entry ) ;
                    if( Entry.Back_Pointer <> Previous_Current ) then { Must have been cross-linked }
                    begin
                        goto Restart ;
                    end ;
                end ;
                if( not _Valid_Filename( Entry.Name ) ) then
                begin
                    Delete_File( Current ) ;
                    goto Restart ;
                end ;
                if( Previous_Name < Uppercase( Entry.Name ) ) then { Left branch is AFTER node }
                begin
                    { Correct the directory structure... }
                    Repair_Tree( Directory_Entry, Current ) ;
                    goto Restart ; { ...and try again }
                end ;
            end ; { if( not File_Heap^.Read_Only ) }
        end ; { while( Entry.Left_Pointer <> 0 ) }
    end else
    begin
        if( Entry.Back_Pointer = 0 ) then { Cannot backup further - no more files }
        begin
            FindNext := FSErr_FileNotFound ; { Lookup failure  TEST: untested }
            exit ;
        end ;
        while( Entry.Back_Pointer <> 0 ) do
        begin
            Previous := Current ;
            Current := Entry.Back_Pointer ;
            Read_Entry( Current, Entry ) ;
            if( Entry.Left_Pointer = Previous ) then { Came from left branch }
            begin
                break ; { We are at the next file }
            end ;
            if( Entry.Back_Pointer = 0 ) then { Cannot backup further }
            begin
                FindNext := FSErr_FileNotFound ; { Lookup failure }
                exit ;
            end ;
        end ; { while( Entry.Back_Pointer <> 0 ) }
    end ; { if( Entry.Right_Pointer <> 0 ) }

    { Now look for a matching file }
    while( Current <> 0 ) do { Until done }
    begin
        if( FS_Wildcard( Find_Record.Wildcard_Context, Edit( Entry.Name, 4 or 8 or 16 or 32 or 128 ) ) ) then
        { Wildcard match }
        begin
            if(
                ( ( Entry.Attributes and FSFF_Directory ) = 0 ) { Not a directory }
                or
                ( ( Find_Record.Wildcard_Attributes and FSFF_Directory ) <> 0 ) { Requested directories }
              ) then { Only match directories if requested to }
            begin
                FindNext := 0 ;
                move( Entry, Find_Record.Data, sizeof( Entry ) ) ;
                Find_Record.Found_Name := Entry.Name ;
                Find_Record.Found_Attributes := Entry.Attributes ;
                S := Find_Record.Last_Name ;
                Dummy := length( S ) ;
                while( ( Dummy > 0 ) and ( S[ Dummy ] <> '\' ) ) do
                begin
                    dec( Dummy ) ;
                end ;
                S := copy( S, 1, Dummy ) ;
                Find_Record.Last_Name := Uppercase( S + Entry.Name ) ;
                exit ;
            end ;
        end ;
        if( Entry.Right_Pointer = 0 ) then { Need to backup }
        begin
Backup:
            Previous := Current ;
            Current := Entry.Back_Pointer ;
            if( Current <> 0 ) then { Still more files to process }
            begin
                Read_Entry( Current, Entry ) ;
                if( Previous = Entry.Right_Pointer ) then
                { Coming back from right branch }
                begin
                    goto Backup ;
                end ; { Otherwise check this node }
            end ;
        end else
        begin
            goto Loop ; { Go right } { TEST: untested }
        end ;
    end ; { while( Current <> 0 ) }

    { No matching file was found if we get here }
    FindNext := FSErr_FileNotFound ; { Lookup failure }
end ; { TXFile_System.FindNext }


function TXFile_System.Normalize_Path( Name : string ) : string ;

begin
    Normalize_Path := Name ;
end ;


function TXFile_System.Open( const Name : string ; Mode : integer ) : integer ;
{ Open a file and return a pointer to the file object.  Parameters:
        Nam = Name of file to open
        Mode = Open mode (see FSOM_* constants) }

var File_Rec : PNative_File_Record ;
    _Name : string ;

begin
{$IFDEF Development}
    CC_Log( 'TXFile_System.Open "' + Name + '"' ) ;
{$ENDIF}
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        Open := 0 ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    _Name := Edit( Name, 4 or 8 or 32 or 128 ) ;
    if( length( _Name ) = 0 ) then { Null filename is invalid }
    begin
        IO_Error.Code := FSErr_FileName ; { Invalid file name }
        IO_Error.Facility := FSErr_Facility ;
        Open := 0 ;
        exit ;
    end ;

    new( File_Rec ) ;
    File_Rec^.Name := _Name ;
    File_Rec^.File_Entry := Find_File( _Name, File_Rec^.Directory_Entry ) ;

    if( IO_Error.Code <> 0 ) then
    begin
        Open := 0 ; { Open failure }
        dispose( File_Rec ) ;
    end else
    begin
        File_List.Insert( File_Rec ) ;
        Open := File_List.Count ;
    end ;
end ; { TXFile_System.Open }


procedure TXFile_System.Attach ;

var UEC : TUEC ;

begin
    Open_File_System( Path, 0, UEC ) ;
end ;


procedure TXFile_System.Detach ;

begin
    Close_File_System( Path ) ;
end ;


procedure TXFile_System.Close( Handle : integer ) ;

var File_Rec : PNative_File_Record ;

begin
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    dispose( PNative_File_Record( File_List.At( Handle - 1 ) ) ) ;
    File_List.AtPut( Handle - 1, nil ) ;

    { Reclaim unused handles at the end of the file list }
    while(
           ( File_List.Count > 0 )
           and
           ( File_List.At( File_List.Count - 1 ) = nil )
         ) do
    begin
        File_List.AtDelete( File_List.Count - 1 ) ;
    end ;
end ; { TXFile_System.Close }


function TXFile_System.Create( const Name : string ;
    Mode : integer ) : integer ;


var _Name : string ;
    New_File_Entry : TDirectory_Entry ;

    procedure Create_File_Entry ;

    begin
        fillchar( New_File_Entry, sizeof( New_File_Entry ), 0 ) ;
        New_File_Entry.Name := _Name ;
        New_File_Entry.Version := 1 ;
        New_File_Entry.Version_Limit := 1 ;
        New_File_Entry.Extension := Entry_Size ;
        New_File_Entry.Back_Pointer := 0 ; { Root node }
    end ;

var Directory_Entry, File_Entry : TStore_Address ;
    Dummy : integer ;
    Entry : TDirectory_Entry ;
    Up_Name : string ;

begin { TXFile_System.Create }
    Create := 0 ; { Assume failure }
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    if( ( Flags and FSF_Read_Only ) <> 0 ) then
    begin
        Set_Error( FSErr_Read_Only ) ;
        exit ;
    end ;
    Erase( Name ) ; { Delete existing file, if any }
    if( ( IO_Error.Code <> 0 ) and ( IO_Error.Code <> FSErr_FileNotFound ) ) then
    begin { If not success or file-not-found error, don't continue }
        Create := 0 ;
        exit ;
    end ;

    Set_Error( 0 ) ; { No error }

    { Find file's directory... }
    _Name := Edit( Name, 4 or 8 or 128 ) ;
    Up_Name := Uppercase( _Name ) ;
    Dummy := length( _Name ) ;
    while( ( Dummy > 0 ) and ( Name[ Dummy ] <> '\' ) ) do
    { Find beginning of file name }
    begin
        Dummy := Dummy - 1 ;
    end ;
    if( Dummy = 1 ) then { Root directory }
    begin
        Dummy := 0 ;
        _Name := copy( _Name, 2, 254 ) ;
    end ;
    if( Dummy = 0 ) then { No directory = root directory }
    begin
        Directory_Entry := Root ;
    end else
    begin
        File_Entry := Find_File( copy( Up_Name, 1, Dummy - 1 ), Directory_Entry ) ; { Find directory }
        if( IO_Error.Code <> 0 ) then
        begin
            Create := 0 ; { Should not get here due to erase above } { TEST: Untested }
            exit ;
        end ;
        Directory_Entry := File_Entry ;
        _Name := copy( _Name, Dummy + 1, 254 ) ;
    end ;

    { Check disk space }
    if( File_Heap^.XSpaceAvail < Entry_Size ) then
    begin
        Set_Error( FSErr_NoRoomOnDevice ) ; { TEST: Untested }
        Create := 0 ;
        exit ;
    end ;

    { Create new file... }
    Read_Entry( Directory_Entry, Entry ) ;
    Directory_Entry := Entry.Allocation_Info ;
    Create_File_Entry ;
    File_Entry := Directory_Entry ;
    Tree_Add( File_Entry, New_File_Entry, 0 ) ; { Add file to directory }

    Create := Open( Name, Mode or FSOM_Output ) ;
end ; { TXFile_System.Create }


function TXFile_System.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ; { TODO: Not yet implemented }
end ;


procedure TXFile_System.Initialize( Nam, ID : string ; _Flags : longint ;
    var UEC : tUEC ) ;
{ Initialize a file system on the device specified with Nam.  Give it the
  specified ID.  Returns condition code in UEC. }

var Dummy : TUEC ;
    F : file ;
    PF : PSFile ;

begin
    Set_Error( 0 ) ;
    Flags := _Flags ;

    { Create the FH file }
    assign( F, Nam ) ;
    {$I-}
    rewrite( F ) ;
    {$I+}
    Dummy.Code := IOResult ;
    if( Dummy.Code <> 0 ) then
    begin
        UEC.Facility := 1 ;
        UEC.Code := Dummy.Code ;
        IO_Error := UEC ;
        exit ;
    end ;
    System.close( F ) ;
    PF := Files.Open( Nam, FSOM_Output, UEC ) ;
    if( UEC.Code <> 0 ) then
    begin
        IO_Error := UEC ;
        exit ;
    end ;

    Initialize_Store( PF, ID, Flags, UEC ) ;
    PF^.Close ;
end ; { TXFile_System.Initialize }


procedure TXFile_System.Initialize_Store( S : PSStore ; ID : string ;
    _Flags : longint ; var UEC : tUEC ) ;

var _Boot : FS_Boot_Record ;
    Dummy : tUEC ;
    Entry : TDirectory_Entry ;
    FH_Flags : longint ;
    _File_Heap : PFH ;
    Header : FS_Header_Record ;
    Temp : integer ;

begin
    { Initialize for store heap access... }
    Flags := _Flags ;
    if( ( _Flags and FSF_Fixed_Size ) <> 0 ) then
    begin
        FH_Flags := FHF_NoExtend ;
    end else
    begin
        FH_Flags := 0 ;
    end ;
    FH_Flags := FH_Flags or FHF_NoClose ; { Don't close - if store is in memory, closing it might make it go away }
    _File_Heap := new( PFH, Init_On_Store( Entry_Size, FH_Flags, S, Dummy, Memory_Heap ) ) ;
    _File_Heap^.Initialize ;
    _File_Heap^.Own_Allocator := False ;
    if( Dummy.Code <> 0 ) then { Should not happen }
    begin
        UEC := Dummy ; { TEST: untested }
        IO_Error := UEC ;
        dispose( _File_Heap, Done ) ;
        exit ;
    end ;

    { Create device boot record... }
    fillchar( _Boot, sizeof( _Boot ), 0 ) ;
    _Boot.Header := _File_Heap^.Allocate( sizeof( Header ) ) ; { Allocate room for header }
    _Boot.Physical_Block_Size := Entry_Size ;
    _File_Heap^.Set_Root( _Boot ) ;

    { Create device header record... }
    fillchar( Header, sizeof( Header ), 0 ) ;
    Header.ID := ID ;
    Header.Root := _File_Heap^.Allocate( sizeof( Entry ) ) ; { Root directory }
    Header.Boot_Pointer := 0 ; { Pointer to bootstrap area (0 = no bootstrap) }
    Header.Boot_Size := 0 ; { Size of bootstrap, in bytes }
    Header.Version := 0 ; { File system structure version }
    Header.Default_Cluster_Size := 1 ;
    Header.Initialized[ 0 ] := 0 ;
    Header.Initialized[ 1 ] := 0 ;
    Header.Flags := 0 ;
    fillchar( Header.Reserved, 102, 0 ) ;
    _File_Heap^.Write_Heap( _Boot.Header, Entry_Size, Header ) ;

    { Create root directory master entry record... }
    fillchar( Entry, sizeof( Entry ), 0 ) ;
    Entry.Name := '' ;
    Entry.Version := 1 ;
    Entry.Version_Limit := 1 ;
    Entry.Size := Entry_Size ;
    Entry.Attributes := FSFF_Directory ; { FSFF flags }
    Entry.Extension := Entry_Size ;
    Entry.Record_Size := Entry_Size ;
    Entry.Uncompressed_Size := Entry_Size ;
    Entry.Allocation_Info := _File_Heap^.Allocate( Entry_Size ) ;
    if( Entry.Allocation_Info = 0 ) then
    begin
        UEC.Facility := FSErr_Facility ;
        UEC.Code := FSErr_NoRoomOnDevice ; { TEST: untested }
        IO_Error := UEC ;
        dispose( _File_Heap, Done ) ;
        exit ;
    end ;
    _File_Heap^.Write_Heap( Header.Root, Entry_Size, Entry ) ;

    { Now write root directory root entry... }
    Temp := Entry.Allocation_Info ;
    Entry.Allocation_Info := 0 ;
    _File_Heap^.Write_Heap( Temp, Entry_Size, Entry ) ;

    { Done with initialization... }
    dispose( _File_Heap, Done ) ;
    UEC.Facility := FSErr_Facility ;
    UEC.Code := 0 ;
end ; { TXFile_System.Initialize_Store }


function TXFile_System.DeviceSize( Nam : string ) : TStore_Address ;

var Dummy : TStore_Address ;
    F : file ;

begin
    DeviceSize := -1 ; { Assume error }
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    if( ( length( Nam ) = 0 ) or ( ( Flags and FSF_Fixed_Size ) <> 0 ) ) then
    begin
        DeviceSize := File_Heap^.Extended_Size ;
        exit ;
    end ;
    if( ( length( Nam ) > 1 ) and ( Nam[ 2 ] <> ':' ) ) then
    begin
        Dummy := DiskFree( 0 ) ;
    end else
    begin
        Dummy := DiskFree( ord( Nam[ 1 ] ) - 64 ) ;
    end ;

    assign( F, Nam ) ;
    {$I-}
    reset( F ) ;
    {$I+}
    IO_Error.Code := IOResult ;
    if( IO_Error.Code <> 0 ) then
    begin
        DeviceSize := -1 ;
        IO_Error.Facility := 1 ;
    end else
    begin
        DeviceSize := filesize( F ) + Dummy ; { Total max size }
        System.close( F ) ;
    end ;
end ; { TXFile_System.DeviceSize }


function TXFile_System.DeviceFree( Nam : string ) : TStore_Address ;

var Dummy : tUEC ;
    _File_Heap : PFH ; { Store heap which controls file }

begin
    DeviceFree := -1 ; { Assume error }
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    IO_Error.Facility := FSErr_Facility ;
    IO_Error.Code := 0 ;
    if( ( length( Nam ) = 0 ) or ( ( Flags and FSF_Fixed_Size ) <> 0 ) ) then
    begin
        DeviceFree := File_Heap^.XSpaceAvail ;
        exit ;
    end ;
    _File_Heap := new( PFH, Init( Entry_Size, FS_To_FH_Flags( Flags ), Nam, Dummy, Memory_Heap ) ) ;
    _File_Heap^.Initialize ;
    if( Dummy.Code <> 0 ) then
    begin
        IO_Error := _File_Heap^.Last_Error ;
    end else
    begin
        DeviceFree := _File_Heap^.XSpaceAvail ;
    end ;
    _File_Heap^.Own_Allocator := False ;
    dispose( _File_Heap, Done ) ;
end ; { TXFile_System.DeviceFree }


function TXFile_System.MKDir( Nam : string ) : longint ;
{ Create a subdirectory named Nam }

    function FNF : boolean ;

    begin
        FNF := (
                 ( IO_Error.Code = FSErr_FileNotFound )
                 and
                 ( IO_Error.Facility = FSErr_Facility )
               ) ;
    end ;

var Handle : integer ;
{$IFNDEF DELPHI}
    File_Entry : TStore_Address ;
{$ENDIF}
    Directory_Entry : TStore_Address ;
    Name : string ;

begin { TXFile_System.MkDir }
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        MkDir := FSErr_InvalidDevice ;
        exit ;
    end ;
    Set_Error( 0 ) ;
    Nam := Edit( Nam, 4 or 8 or 128 ) ;
    Name := Uppercase( Nam ) ;
{$IFNDEF DELPHI}
    File_Entry :=
{$ENDIF}
        Find_File( Name, Directory_Entry ) ;
    if( not FNF ) then
    begin
        if( IO_Error.Code = 0 ) then { File was found }
        begin
            MKDir := FSErr_FileExists ;
            Set_Error( FSErr_FileExists ) ;
        end else
        begin
            MKDir := IO_Error.Code ;
        end ;
        exit ;
    end ;
    Handle := Create( Nam, FSOM_OUTPUT ) ;
    if( Handle <> 0 ) then
    begin
        Close( Handle ) ;
    end else
    begin
        MKDir := IO_Error.Code ;
        exit ;
    end ;
    Set_File_Attributes( Nam, FSFF_Directory ) ;
    MKDir := 0 ;
    Set_Error( FSErr_Success ) ;
end ; { TXFile_System.MKDir }


procedure TXFile_System.Read( Handle : integer ; Position : TStore_Address ;
    Len : TStore_Address ; var Buffer ; var Count : TStore_Address ) ;

var _Buffer : PByte_Array ;
    Cluster : longint ; { Current cluster being allocated }
    Dummy, Dummy1 : longint ;
    Entry : TDirectory_Entry ;
    Offset : integer ; { Source buffer offset }
    This_Count : word ; { Count of bytes to move for current cluster }
    File_Rec : PNative_File_Record ;
    _Position : TStore_Address ;

begin
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    { Setup... }
    Set_Error( 0 ) ;
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;
    if( File_Rec^.File_Entry = 0 ) then { Must be a directory }
    begin
        Set_Error( FSErr_IllegalOperationType ) ;
        exit ;
    end ;
{$IFDEF DELPHI}
    _Position := Position ;
{$ELSE}
    _Position := trunc( Position ) ;
{$ENDIF}

    Count := 0 ;
    Read_Entry( File_Rec^.File_Entry, Entry ) ;
    if( ( Entry.Attributes and FSFF_Directory ) <> 0 ) then { A directory }
    begin
        Set_Error( FSErr_ProtectionViolation ) ;
        exit ;
    end ;
    if( _Position + Len > Entry.End_Of_File ) then { Need to truncate }
    begin
        Len := Entry.End_Of_File - _Position ; { Only read what is there }
        if( Len = -1 ) then { At EOF }
        begin
            exit ;
        end ;
        if( Len < 0 ) then
        begin
            Set_Error( FSErr_Internal_Corruption ) ;
            exit ;
        end ;
    end ;

    { Now read from file... }
    getmem( _Buffer, Entry.Extension ) ; { Allocate a cluster-sized buffer }
    Cluster := _Position div Entry.Extension ; { Cluster to start reading from }
    Dummy1 := _Position - ( Cluster * Entry.Extension ) ; { Offset }
    Offset := 0 ; { Offset in source buffer }
    while( Len > 0 ) do { Until done reading from file }
    begin
        Dummy := Get_Absolute( Entry.Allocation_Info, Cluster ) ;
        File_Heap^.Read_Heap( Dummy, Entry.Extension, _Buffer^ ) ;
        This_Count := Entry.Extension - Dummy1 ; { Amount remaining in cluster }
        if( This_Count > Len ) then
        begin
            This_Count := Len ; { Move no more than asked to }
        end ;
        move( _Buffer^[ Dummy1 ], TByte_Array( Buffer )[ Offset ], This_Count ) ;
        Offset := Offset + This_Count ; { New source position }
        Len := Len - This_Count ; { Bytes left to read }
        Count := Count + This_Count ;
        Cluster := Cluster + 1 ; { Next cluster next time }
        Dummy1 := 0 ; { Next source starts at offset 0 }
    end ;
    freemem( _Buffer, Entry.Extension ) ; { Deallocate it }
end ; { TXFile_System.Read }


procedure TXFile_System.Rename( const Old, New : string ) ;

var Dummy : longint ;
    Entry, Other : TDirectory_Entry ;
    File_Entry, Directory_Entry : TStore_Address ;
    _New : string ;
    Old_File_Entry : longint ;

begin
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    if( ( Flags and FSF_Read_Only ) <> 0 ) then
    begin
        Set_Error( FSErr_Read_Only ) ;
        exit ;
    end ;
    Old_File_Entry := Find_File( Edit( Old, 4 or 8 or 32 or 128 ), Directory_Entry ) ;
    if( IO_Error.Code <> 0 ) then { Some error }
    begin
        exit ;
    end ;
    Read_Entry( Old_File_Entry, Entry ) ;

{$IFNDEF DELPHI}
    File_Entry :=
{$ENDIF}
        Find_File( Edit( New, 4 or 8 or 32 or 128 ), Directory_Entry ) ;
    if( IO_Error.Code = 0 ) then { Found it }
    begin
        Set_Error( FSErr_FileExists ) ; { File exists with new name }
        exit ;
    end ;
    if(
        ( IO_Error.Facility <> FSErr_Facility )
        or
        ( IO_Error.Code <> FSErr_FileNotFound )
      ) then { Some other error }
    begin
        exit ;
    end ;
    Set_Error( 0 ) ;

    { Find file's new directory... }
    _New := Edit( New, 4 or 8 or 32 or 128 ) ;
    Dummy := length( _New ) ;
    while( ( Dummy > 0 ) and ( _New[ Dummy ] <> '\' ) ) do
    { Find beginning of file name }
    begin
        Dummy := Dummy - 1 ;
    end ;
    if( Dummy = 1 ) then { Root directory }
    begin
        Dummy := 0 ;
        _New := copy( _New, 2, 254 ) ;
    end ;
    if( Dummy = 0 ) then { No directory = root directory }
    begin
        Directory_Entry := Root ;
    end else
    begin
        File_Entry := Find_File( copy( _New, 1, Dummy - 1 ), Directory_Entry ) ; { Find directory }
        if( IO_Error.Code <> 0 ) then { Shouldn't happen since above Find_File on New worked }
        begin
            exit ;
        end ;
        Directory_Entry := File_Entry ;
    end ;
    Entry.Name := copy( _New, Dummy + 1, 254 ) ;
    Read_Entry( Directory_Entry, Other ) ;
    Directory_Entry := Other.Allocation_Info ; { Root of new directory }

    { Now move file to new name/location... }
    Set_Error( 0 ) ;
    Tree_Delete( Old_File_Entry ) ; { Remove file from old location }
    Tree_Add( Directory_Entry, Entry, 0 ) ; { Add file in new location }
end ; { TXFile_System.Rename }


procedure TXFile_System.Get_Header( _Path : string ;
    var Header : FS_Header_Information ; var Res : TUEC ) ;

var _Boot : FS_Boot_Record ;
    _File_Heap : TFH ;
    Original : FS_Header_Record ;

begin
    _File_Heap.Init( Entry_Size, FS_To_FH_Flags( Flags ), _Path, Res, Memory_Heap ) ;
    if( Res.Code <> 0 ) then
    begin
        exit ;
    end ;
    _File_Heap.Own_Allocator := False ;
    _File_Heap.Get_Root( _Boot ) ;
    Read_Header( @_File_Heap, _Boot.Header, Original ) ;
    Set_Error( 0 ) ;
    Header.ID := Original.ID ;
    Header.Root := Original.Root ;
    Header.Version := Original.Version ;
    Header.Default_Cluster_Size := Original.Default_Cluster_Size ;
    Header.Flags := Original.Flags ;
    _File_Heap.Done ;
    Res := IO_Error ;
end ; { TXFile_System.Get_Header }


procedure TXFile_System.Set_Header( _Path : string ;
    var Header : FS_Header_Information ; var Res : TUEC  ) ;

var _Boot : FS_Boot_Record ;
    _File_Heap : TFH ;
    Original : FS_Header_Record ;

begin
    if( ( Flags and FSF_Read_Only ) <> 0 ) then
    begin
        Set_Error( FSErr_Read_Only ) ;
        exit ;
    end ;
    _File_Heap.Init( Entry_Size, FS_to_FH_Flags( Flags ), _Path, Res, Memory_Heap ) ;
    if( Res.Code <> 0 ) then
    begin
        exit ;
    end ;
    _File_Heap.Own_Allocator := False ;
    _File_Heap.Get_Root( _Boot ) ;
    Read_Header( @_File_Heap, _Boot.Header, Original ) ;
    Original.ID := Header.ID ;
    Original.Default_Cluster_Size := Header.Default_Cluster_Size ;
    Original.Flags := Header.Flags ;
    Write_Header( @_File_Heap, _Boot.Header, Original ) ;
    _File_Heap.Done ;
    Set_Error( 0 ) ;
    Res := IO_Error ;
end ; { TXFile_System.Set_Header }



function TXFile_System.Facility : longint ;

begin
    Facility := FSErr_Facility ;
end ;


function TXFile_System.Version : longint ;

begin
    Version := 10 ; { Version 1.0 }
end ;


procedure TXFile_System.Verify( Name : string ; var Res : TUEC ) ; { TODO: Untested }

    function Check_Pointer( P : TStore_Address ) : boolean ;

    begin
        Check_Pointer := ( P = 0 ) or ( P <= File_Heap^.AT.Get_Offset - sizeof( TDirectory_Entry ) ) ;
    end ;


    procedure Verify_Directory( Current : TStore_Address ) ;

    var Entry : TDirectory_Entry ;
        Name : string ;
        Previous : TStore_Address ;

    label Backup, Loop ;

    begin
        Read_Entry( Current, Entry ) ;
        { Get root of directory tree }
Loop:
        Current := Entry.Right_Pointer ; { Get first file entry }
        if( Current <> 0 ) then { Not an empty directory }
        begin
            Read_Entry( Current, Entry ) ;
            while( Entry.Left_Pointer <> 0 ) do { Start at first file }
            begin
                Current := Entry.Left_Pointer ;
                Read_Entry( Current, Entry ) ;
            end ;
        end ;

        { Now look at all files }
        while( Current <> 0 ) do { Until done }
        begin
            if( length( Name ) > 0 ) then
            begin
                if( Name > uppercase( Entry.Name ) ) then
                begin
                    Set_Error( FSErr_CorruptedFileStructure ) ;
                    exit ;
                end ;
            end ;
            Name := uppercase( Entry.Name ) ;
            if( ( Entry.Attributes and FSFF_Directory ) <> 0 ) then
            begin
                Verify_Directory( Current ) ;
                if( IO_Error.Code <> 0 ) then
                begin
                    exit ;
                end ;
            end else
            begin
                { Verify file pointers... }
                if( Check_Pointer( Entry.Caching_Info ) ) then
                begin
                    Set_Error( FSErr_CorruptedFileStructure ) ;
                    exit ;
                end ;
                if( Check_Pointer( Entry.Access_Info ) ) then
                begin
                    Set_Error( FSErr_CorruptedFileStructure ) ;
                    exit ;
                end ;
                if( Check_Pointer( Entry.Auditing_Info ) ) then
                begin
                    Set_Error( FSErr_CorruptedFileStructure ) ;
                    exit ;
                end ;
                if( Check_Pointer( Entry.Access_Stats ) ) then
                begin
                    Set_Error( FSErr_CorruptedFileStructure ) ;
                    exit ;
                end ;
                if( Check_Pointer( Entry.Allocation_Info ) ) then
                begin
                    Set_Error( FSErr_CorruptedFileStructure ) ;
                    exit ;
                end ;
                if( Check_Pointer( Entry.Next_Alpha_Entry ) ) then
                begin
                    Set_Error( FSErr_CorruptedFileStructure ) ;
                    exit ;
                end ;
                if( Check_Pointer( Entry.Left_Pointer ) ) then
                begin
                    Set_Error( FSErr_CorruptedFileStructure ) ;
                    exit ;
                end ;
                if( Check_Pointer( Entry.Right_Pointer ) ) then
                begin
                    Set_Error( FSErr_CorruptedFileStructure ) ;
                    exit ;
                end ;
                if( Check_Pointer( Entry.Back_Pointer ) ) then
                begin
                    Set_Error( FSErr_CorruptedFileStructure ) ;
                    exit ;
                end ;
            end ;
            if( Entry.Right_Pointer = 0 ) then { Need to backup }
            begin
Backup:
                Previous := Current ;
                Current := Entry.Back_Pointer ;
                if( Current <> 0 ) then { Still files to process }
                begin
                    Read_Entry( Current, Entry ) ;
                    if( Previous = Entry.Right_Pointer ) then
                    { Coming back from right branch }
                    begin
                        goto Backup ;
                    end ; { Otherwise check this node }
                end ;
            end else
            begin
                goto Loop ; { Go right }
            end ;
        end ;
    end ; { TXFile_System.Verify.Verify_Directory }


var _Root : TStore_Address ;

begin { TXFile_System.Verify }
    Set_Error( 0 ) ;

    File_Heap^.Read_Heap( Boot.Header + 128, 4, _Root ) ;
    Verify_Directory( _Root ) ;

    Res := IO_Error ;
end ; { TXFile_System.Verify }


procedure TXFile_System.Set_File_Attributes( const Name : string ;
    Value : longint ) ;

var Dummy : longint ;
    Convert_To_Directory : boolean ; { True if converting to directory }
    Directory_Entry, File_Entry : TStore_Address ;
    Entry : TDirectory_Entry ; { File entry }
    Root_Entry : TDirectory_Entry ; { Directory root entry }
    _Name : string ;

begin
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    if( ( Flags and FSF_Read_Only ) <> 0 ) then
    begin
        Set_Error( FSErr_Read_Only ) ;
        exit ;
    end ;
    _Name := Edit( Name, 4 or 8 or 32 or 128 ) ;
    if( length( _Name ) = 0 ) then
    begin
        Set_Error( FSErr_FileNameLength ) ;
        exit ;
    end ;
    File_Entry := Find_File( _Name, Directory_Entry ) ;
    if( IO_Error.Code <> 0 ) then
    begin
        exit ;
    end ;
    Read_Entry( File_Entry, Entry ) ;
    if(
        ( ( Entry.Attributes and FSFF_Directory ) = 0 )
        and
        ( ( Value and FSFF_Directory ) = FSFF_Directory )
      ) then { Change a file to a directory }
    begin
        if( Entry.Allocation_Info <> 0 ) then { File is not empty }
        begin
            Set_Error( FSErr_FileNotEmpty ) ;
            exit ;
        end ;
        Entry.Allocation_Info := Allocate_Entry ;
        if( Entry.Allocation_Info = 0 ) then
        begin
            Set_Error( FSErr_NoRoomOnDevice ) ; { TEST: Untested }
            exit ;
        end ;
        Convert_To_Directory := True ;
    end else
    begin
        Convert_To_Directory := False ;
    end ;
    if(
        ( ( Entry.Attributes and FSFF_Directory ) = FSFF_Directory )
        and
        ( ( Value and FSFF_Directory ) = 0 )
      ) then { Change a directory to a file }
    begin
        if( Entry.Allocation_Info = 0 ) then
        begin
            Set_Error( FSErr_CorruptedFileStructure ) ; { TEST: Untested }
            exit ;
        end ;
        Read_Entry( Entry.Allocation_Info, Root_Entry ) ;
        if( Root_Entry.Right_Pointer <> 0 ) then
        begin
            Set_Error( FSErr_DirectoryNotEmpty ) ;
            exit ;
        end ;
        File_Heap^.Deallocate( Entry.Allocation_Info, Entry_Size ) ;
        Entry.Allocation_Info := 0 ;
    end ;
    Entry.Attributes := Value ;
    Write_Entry( File_Entry, Entry ) ;
    if( Convert_To_Directory ) then
    begin
        Dummy := Entry.Allocation_Info ;
        fillchar( Entry, Entry_Size, 0 ) ;
        Write_Entry( Dummy, Entry ) ;
        Directory_Entry := 0 ;
    end ;
end ; { TXFile_System.Set_File_Attributes }


procedure TXFile_System.Set_Time( Handle : integer ; Typ, Time : Longint ) ;

var Dummy : TStore_Address ;
    Entry : TDirectory_Entry ;
    File_Rec : PNative_File_Record ;

begin
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    if( ( Flags and FSF_Read_Only ) <> 0 ) then
    begin
        Set_Error( FSErr_Read_Only ) ;
        exit ;
    end ;

    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    if( File_Rec^.File_Entry = 0 ) then { Must be a directory }
    begin
        Dummy := File_Rec^.Directory_Entry ; ;
    end else
    begin
        Dummy := File_Rec^.File_Entry ;
    end ;
    Read_Entry( Dummy, Entry ) ;
    Entry.Revision_Time := Time ;
    Write_Entry( Dummy, Entry ) ;
end ;


procedure TXFile_System.Set_Size( Handle : integer ; New_Size : TStore_Address ) ;

var Allocation : tAllocation_Record ;
    Allocation_Next : longint ;
    Current, Dummy : longint ;
    Entry : tDirectory_Entry ;
    Modified_Allocation : boolean ;
    File_Rec : PNative_File_Record ;

begin
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    if( ( Flags and FSF_Read_Only ) <> 0 ) then
    begin
        Set_Error( FSErr_Read_Only ) ;
        exit ;
    end ;

    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;
    if( File_Rec^.File_Entry = 0 ) then { Must be a directory }
    begin
        Set_Error( FSErr_IllegalOperationType ) ;
        exit ;
    end ;

    { Setup... }
    if( New_Size < -1 ) then
    begin
        New_Size := 0 ;
    end ;
    Set_Error( 0 ) ;

    { Get file entry... }
    Read_Entry( File_Rec^.File_Entry, Entry ) ;
    if( ( Entry.Attributes and FSFF_Directory ) <> 0 ) then { A directory }
    begin
        Set_Error( FSErr_ProtectionViolation ) ; { TEST: Untested }
        exit ;
    end ;

    if( New_Size > Entry.End_Of_File ) then { Expand }
    begin
        Extend_File( File_Rec^.File_Entry, New_Size, Entry ) ;
    end ;
    if( New_Size < Entry.End_Of_File ) then { Truncate }
    begin
        Entry.End_Of_File := New_Size ;
        Entry.Uncompressed_Size := ( Entry.End_Of_File + Entry.Extension - 1 )
            and ( not ( Entry.Extension - 1 ) ) ; { New size }
        Entry.Size := Entry.Uncompressed_Size ;
        Write_Entry( File_Rec^.File_Entry, Entry ) ;
        Current := Entry.Allocation_Info ;
        File_Heap^.Read_Heap( Entry.Allocation_Info, Entry_Size, Allocation ) ;
        Dummy := New_Size div Entry.Extension ; { New last cluster }
        while( Dummy > Max_Cluster_In_Block ) do { Find last allocation record }
        begin
            Dummy := Dummy - Max_Cluster_In_Block - 1 ;
            Current := Allocation.Next ;
            File_Heap^.Read_Heap( Allocation.Next, Entry_Size, Allocation ) ;
        end ;
        if( New_Size > 0 ) then
        begin
            Dummy := Dummy + 1 ; { Start with cluster just after last used one }
        end ;
        while( Current <> 0 ) do
        { Until end of allocation records for file }
        begin
            Modified_Allocation := False ;
            while( Dummy <= Max_Cluster_In_Block ) do { Until end of clusters in record }
            begin
                if( Allocation.Cluster[ Dummy ] = 0 ) then { End of used clusters }
                begin
                    break ;
                end ;
                File_Heap^.Deallocate( Allocation.Cluster[ Dummy ],
                    Entry.Extension ) ;
                Allocation.Cluster[ Dummy ] := 0 ;
                Modified_Allocation := True ;
                Dummy := Dummy + 1 ;
            end ;
            Allocation_Next := Allocation.Next ; { Save link to next block }
            if( Allocation.Cluster[ 0 ] <> 0 ) then { Something still in this block }
            begin
                if( Allocation.Next <> 0 ) then
                begin
                    Allocation.Next := 0 ;
                    Modified_Allocation := True ;
                end ;
                if( Modified_Allocation ) then
                begin
                    File_Heap^.Write_Heap( Current, Entry_Size, Allocation ) ;
                end ;
            end else
            begin { Nothing in this block }
                File_Heap^.Deallocate( Current, Entry_Size ) ;
                if( Current = Entry.Allocation_Info ) then { File was reduced to size 0 }
                begin
                    Entry.Allocation_Info := 0 ;
                    Write_Entry( File_Rec^.File_Entry, Entry ) ;
                end ;
            end ; { if( Allocation.Cluster[ 0 ] <> 0 ) }
            Current := Allocation_Next ; { Move to next allocation block in chain }
            Dummy := 0 ;
            if( Current <> 0 ) then
            begin
                File_Heap^.Read_Heap( Current, Entry_Size, Allocation ) ;
            end ;
        end ; { while( Current <> 0 ) }
    end ;
end ; { TXFile_System.Set_Size }


procedure TXFile_System.Write( Handle : integer ; Position : TStore_Address ;
    Len : TStore_Address ; var Buffer ; var Count : TStore_Address ) ;

var _Buffer : PByte_Array ;
    Cluster : longint ; { Current cluster being allocated }
    Dummy, Dummy1 : longint ;
    Entry : TDirectory_Entry ;
    Offset : integer ; { Source buffer offset }
    This_Count : word ; { Count of bytes to move for current cluster }
    File_Rec : PNative_File_Record ;
    _Position : TStore_Address ;

begin
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    if( ( Flags and FSF_Read_Only ) <> 0 ) then
    begin
        Set_Error( FSErr_Read_Only ) ;
        exit ;
    end ;

    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    if( File_Rec^.File_Entry = 0 ) then { Must be trying to write to a directory }
    begin
        Set_Error( FSErr_IllegalOperationType ) ;
        exit ;
    end ;
{$IFDEF DELPHI}
    _Position := Position ;
{$ELSE}
    _Position := trunc( Position ) ;
{$ENDIF}
    Set_Error( 0 ) ;
    Count := 0 ;
    Read_Entry( File_Rec^.File_Entry, Entry ) ;
    if( ( Entry.Attributes and FSFF_Directory ) <> 0 ) then { A directory }
    begin
        Set_Error( FSErr_ProtectionViolation ) ;
        exit ;
    end ;

    if( _Position + Len > Entry.End_Of_File ) then { Need to extend }
    begin
        Extend_File( File_Rec^.File_Entry, _Position + Len, Entry ) ;
        if( IO_Error.Code <> 0 ) then { Could not extend the file }
        begin
            exit ;
        end ;
    end ;

    { Now write data to file... }
    getmem( _Buffer, Entry.Extension ) ; { Allocate a cluster-sized buffer }
    Cluster := _Position div Entry.Extension ; { Cluster to start writing to }
    Dummy1 := _Position - ( Cluster * Entry.Extension ) ; { Offset }
    Offset := 0 ; { Offset in source buffer }
    while( Len > 0 ) do { Until done writing to file }
    begin
        Dummy := Get_Absolute( Entry.Allocation_Info, Cluster ) ;
        if( Dummy <= 0 ) then { This should not happen }
        begin
            Set_Error( FHErr_Corrupted_File_Structure ) ;
            exit ;
        end ;
        File_Heap^.Read_Heap( Dummy, Entry.Extension, _Buffer^ ) ;
        This_Count := Entry.Extension - Dummy1 ;
        if( This_Count > Len ) then
        begin
            This_Count := Len ; { Move no more than asked to }
        end ;
        move( TByte_Array( Buffer )[ Offset ], _Buffer^[ Dummy1 ], This_Count ) ;
        File_Heap^.Write_Heap( Dummy, Entry.Extension, _Buffer^ ) ;
        Offset := Offset + This_Count ; { New source position }
        Len := Len - This_Count ; { Bytes left to write }
        Count := Count + This_Count ;
        Cluster := Cluster + 1 ; { Next cluster next time }
        Dummy1 := 0 ; { Next destination starts at offset 0 }
    end ;
    Freemem( _Buffer, Entry.Extension ) ;
end ; { TXFile_System.Write }


function TXFile_System.Get_File_Attributes( const Name : string ) : longint ;

var Entry : TDirectory_Entry ;
    Directory_Entry, File_Entry : TStore_Address ;
    _Name : string ;

begin
    Get_File_Attributes := 0 ;
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    { Setup... }
    _Name := Edit( Name, 4 or 8 or 32 or 128 ) ;
    if( length( _Name ) = 0 ) then
    begin
        Set_Error( FSErr_FileNameLength ) ;
        exit ;
    end ;

    Set_Error( 0 ) ;
    File_Entry := Find_File( _Name, Directory_Entry ) ;
    if( IO_Error.Code <> 0 ) then
    begin
        Get_File_Attributes := 0 ;
        exit ;
    end ;
    if( File_Entry = 0 ) then { Directory lookup }
    begin
        Read_Entry( Directory_Entry, Entry ) ;
    end else
    begin
        Read_Entry( File_Entry, Entry ) ;
    end ;
    Get_File_Attributes := Entry.Attributes ;
end ;


function TXFile_System.Get_File_Object( Mode : integer ) : PSFile ;

begin
    Get_File_Object := new( PNative_File, Init( @self, ( ( Mode and FSOM_Text ) <> 0 ), 1 ) ) ;
end ;


function TXFile_System.Get_Size( Handle : integer ) : TStore_Address ;

var Entry : TDirectory_Entry ;
    File_Rec : PNative_File_Record ;

begin
    Get_Size := -1 ;
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    if( File_Rec^.File_Entry = 0 ) then { Must be a directory }
    begin
        Get_Size := Entry_Size ;
    end else
    begin
        Read_Entry( File_Rec^.File_Entry, Entry ) ;
        Get_Size := Entry.End_Of_File ;
    end ;
end ;


function TXFile_System.Get_Time( Handle, Typ : integer ) : Longint ;

var Entry : TDirectory_Entry ;
    File_Rec : PNative_File_Record ;

begin
    Get_Time := 0 ;
    if( File_Heap = nil ) then
    begin
        Set_Error( FSErr_InvalidDevice ) ;
        exit ;
    end ;
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    if( File_Rec^.File_Entry = 0 ) then { Must be a directory }
    begin
        Read_Entry( File_Rec^.Directory_Entry, Entry ) ;
    end else
    begin
        Read_Entry( File_Rec^.File_Entry, Entry ) ;
    end ;
    case Typ of
        0 : Get_Time := Entry.Revision_Time ;
        1 : Get_Time := Entry.Creation_Time ;
        2 : Get_Time := Entry.Access_Time ;
        3 : Get_Time := Entry.Backup_Time ;
        4 : Get_Time := Entry.Expiration_Time ;
        else Get_Time := 0 ;
    end ;
end ; { TXFile_System.Get_Time }


{ Internal utility routines... }

procedure TXFile_System.Delete_File( File_Entry : TStore_Address ) ;

var Allocation : TAllocation_Record ;
    Dummy : integer ;
    Entry : TDirectory_Entry ;

begin
    { Setup... }
    Clear_Cache ;
    Read_Entry( File_Entry, Entry ) ;

    {TODO: close any instances of this file, if opened - or else implement deletion flag}

    { Free allocation for file }
    while( Entry.Allocation_Info <> 0 ) do
    begin
        File_Heap^.Read_Heap( Entry.Allocation_Info, Entry_Size, Allocation ) ;
        for Dummy := 0 to Max_Cluster_In_Block do
        begin
            if( Allocation.Cluster[ Dummy ] <> 0 ) then
            begin
                File_Heap^.Deallocate( Allocation.Cluster[ Dummy ],
                    Entry.Extension ) ;
            end else
            begin
                break ; { Early out (allocation list isn't sparse - 0 is end) }
            end ;
        end ;
        File_Heap^.Deallocate( Entry.Allocation_Info, Entry_Size ) ;
        Entry.Allocation_Info := Allocation.Next ;
    end ;
    Tree_Delete( File_Entry ) ; { Remove file from directory }
end ;


procedure TXFile_System.Extend_File( File_Entry : TStore_Address ;
    New_Size : longint ; var Entry : tDirectory_Entry ) ;

var Allocation : TAllocation_Record ;
    Cluster : longint ; { Current cluster being allocated }
    Current : longint ; { Current allocation record pointer }
    Dummy, Dummy1 : longint ;
    New_Entry_Size : TStore_Address ;

begin
    { Modify file's directory entry... }
    Dummy1 := Entry.Size ; { Original size }
    Entry.End_Of_File := New_Size ; { New end of file }
    New_Entry_Size := ( Entry.End_Of_File + Entry.Extension - 1 )
        and ( not ( Entry.Extension - 1 ) ) ; { New size }
    if( File_Heap^.XSpaceAvail < New_Entry_Size - Dummy1 ) then { No room }
    begin
        Set_Error( FSErr_NoRoomOnDevice ) ; { TEST: Untested }
        exit ;
    end ;

    { Now allocate additional space for file... }
    if( Entry.Allocation_Info = 0 ) then { File is empty }
    begin
        Entry.Allocation_Info := Allocate_Entry ;
        if( Entry.Allocation_Info = 0 ) then
        begin
            Set_Error( FSErr_NoRoomOnDevice ) ; { TEST: Untested }
            exit ;
        end ;
        fillchar( Allocation, Entry_Size, 0 ) ;
        File_Heap^.Write_Heap( Entry.Allocation_Info, Entry_Size, Allocation ) ;
        Write_Entry( File_Entry, Entry ) ;
    end else
    begin
        File_Heap^.Read_Heap( Entry.Allocation_Info, Entry_Size, Allocation ) ;
    end ;

    { Update size entries in directory entry record... }
    Entry.Uncompressed_Size := New_Entry_Size ;
    Entry.Size := Entry.Uncompressed_Size ;

    Current := Entry.Allocation_Info ;
    Dummy := Entry.Size div Entry.Extension ; { Maximum used clusters }
    Dummy1 := 1 ;
    while( Dummy1 <= Dummy ) do
    begin
        for Cluster := 0 to Max_Cluster_In_Block do
        begin
            if( Dummy1 <= Dummy ) then { Still more clusters to allocate }
            begin
                if( Allocation.Cluster[ Cluster ] = 0 ) then { CLuster not yet allocated }
                begin
                    Allocation.Cluster[ Cluster ] :=
                        File_Heap^.Allocate( Entry.Extension ) ;
                    if( Allocation.Cluster[ Cluster ] = 0 ) then
                    begin
                        Set_Error( FSErr_NoRoomOnDevice ) ; { TEST: Untested }
                        exit ;
                    end ;
                end ; { if( Allocation.Cluster[ Cluster ] = 0 ) }
                Dummy1 := Dummy1 + 1 ;
            end else
            begin
                break ; { Once we are done, skip of out loop }
            end ; { if( Dummy1 <= Dummy ) }
        end ; { for Cluster := 0 to Max_Cluster_In_Block }
        File_Heap^.Write_Heap( Current, Entry_Size, Allocation ) ;
        if( Dummy1 <= Dummy ) then { More allocation needed }
        begin
            if( Allocation.Next = 0 ) then
            begin
                Allocation.Next := Allocate_Entry ;
                if( Allocation.Next = 0 ) then
                begin
                    Set_Error( FSErr_NoRoomOnDevice ) ; { TEST: Untested }
                    exit ;
                end ;
                File_Heap^.Write_Heap( Current, Entry_Size, Allocation ) ;
                Current := Allocation.Next ;
                fillchar( Allocation, Entry_Size, 0 ) ;
            end else
            begin
                File_Heap^.Write_Heap( Current, Entry_Size, Allocation ) ;
                Current := Allocation.Next ;
                File_Heap^.Read_Heap( Current, Entry_Size, Allocation ) ;
            end ;
        end ;
    end ;
    File_Heap^.Write_Heap( Current, Entry_Size, Allocation ) ;
    Write_Entry( File_Entry, Entry ) ; { Update on-disk entry }
end ; { TXFile_System.Extend_File }


function Check_Pointer( const P : TStore_Address ; File_Heap : PFH ) : boolean ;

begin
    Check_Pointer := False ; { Assume failure }
{$WARNINGS OFF}
    if( { These checks should be determined at compile-time since Entry_Size is a constant... }
        ( Entry_Size = 16 )
        or
        ( Entry_Size = 32 )
        or
        ( Entry_Size = 64 )
        or
        ( Entry_Size = 128 )
        or
        ( Entry_Size = 256 )
        or
        ( Entry_Size = 512 )
        or
        ( Entry_Size = 1024 )
        or
        ( Entry_Size = 2048 )
        or
        ( Entry_Size = 4096 )
        or
        ( Entry_Size = 8192 )
        or
        ( Entry_Size = 16384 )
        or
        ( Entry_Size = 32768 )
      ) then { A power of 2 }
{$WARNINGS ON}
    begin
        if( ( P and ( Entry_Size - 1 ) ) <> 0 ) then { Not an multiple of Entry size }
        begin
            exit ;
        end ;
    end else
    begin
        if( P <> ( ( P div Entry_Size ) * Entry_Size ) ) then
        begin
            exit ;
        end ;
    end ;
    if( ( P <> 0 ) and ( not File_Heap^.Valid_Data_Address( P ) ) ) then
    begin
        exit ;
    end ;
    Check_Pointer := True ;
end ; { Check_Pointer }


function Valid_Entry( Position : TStore_Address ;
    const Entry : TDirectory_Entry ; const File_Heap : PFH ;
    const Root : TStore_Address ) : boolean ;

    function _Check_Pointer( const P : TStore_Address ; File_Heap : PFH ) : boolean ;

    begin
        _Check_Pointer := False ; { Assume failure }
        if( ( P > 0 ) and ( P <= Root ) ) then
        begin
            exit ;
        end ;
        if( ( Position <> 0 ) and ( P = Position ) ) then { Circular pointer }
        begin
            exit ;
        end ;
        _Check_Pointer := Check_Pointer( P, File_Heap ) ;
    end ;

begin { Valid_Entry }
    Valid_Entry := False ; { Assume failure }
    if( Entry.Reserved[ 0 ] <> 0 ) then { Corruption }
    begin
        exit ;
    end ;

    { Make sure all pointers are theoretically valid... }
    if( not _Check_Pointer( Entry.Left_Pointer, File_Heap ) ) then
    begin
        exit ;
    end ;
    if( not _Check_Pointer( Entry.Right_Pointer, File_Heap ) ) then
    begin
        exit ;
    end ;
    if( not _Check_Pointer( Entry.Back_Pointer, File_Heap ) ) then
    begin
        exit ;
    end ;
    if( not _Check_Pointer( Entry.Caching_Info, File_Heap ) ) then
    begin
        exit ;
    end ;
    if( not _Check_Pointer( Entry.Access_Info, File_Heap ) ) then
    begin
        exit ;
    end ;
    if( not _Check_Pointer( Entry.Auditing_Info, File_Heap ) ) then
    begin
        exit ;
    end ;
    if( not _Check_Pointer( Entry.Access_Stats, File_Heap ) ) then
    begin
        exit ;
    end ;
    if( not _Check_Pointer( Entry.Allocation_Info, File_Heap ) ) then
    begin
        exit ;
    end ;

    { Make sure we don't have invalid pointer combinations... }
    if( ( Entry.Back_Pointer <> 0 ) and ( Entry.Left_Pointer = Entry.Back_Pointer ) ) then
    begin
        exit ;
    end ;
    if( ( Entry.Back_Pointer <> 0 ) and ( Entry.Right_Pointer = Entry.Back_Pointer ) ) then
    begin
        exit ;
    end ;

    Valid_Entry := True ;
end ; { Valid_Entry }


function TXFile_System.Correct_Entry( Current, Directory_Entry : TStore_Address ;
    var Entry : TDirectory_Entry ; const _Root : TStore_Address ) : boolean ;
{ Validate an entry to make sure it is valid and correct it if it isn't (if
  possible).  Returns True if the entry was invalid. }

    function _Check_Pointer( const P : TStore_Address ; File_Heap : PFH ) : boolean ;

    begin
        _Check_Pointer := False ; { Assume failure }
        if( ( P > 0 ) and ( P <= _Root ) ) then
        begin
            exit ;
        end ;
        if( ( Current <> 0 ) and ( P = Current ) ) then { Circular pointer }
        begin
            exit ;
        end ;
        _Check_Pointer := Check_Pointer( P, File_Heap ) ;
    end ;


    procedure Recalculate_Size ;

    var _Entry : tAllocation_Record ;
        Loop : integer ;
        S : int64 ;

    begin
        { Make sure that the entry size doesn't exceed the actual allocated
          size. }
        S := 0 ;
        if( Entry.Allocation_Info <> Null ) then
        begin
            { UNTESTED: This block }
            Read_Allocation( Entry.Allocation_Info, _Entry ) ; { Read allocation data }
            while( _Entry.Next <> Null ) do
            begin
                S := S + ( Max_Cluster_In_Block + 1 ) * Entry.Extension ;
                Read_Allocation( _Entry.Next, _Entry ) ;
            end ;
            for Loop := 0 to Max_Cluster_In_Block do
            begin
                if( _Entry.Cluster[ Loop ] <> Null ) then
                begin
                    S := S + Entry.Extension ;
                end ;
            end ;
        end ;
        if( Entry.Size > S ) then
        begin
            { UNTESTED: This block }
            Entry.Size := S ;
            if( not File_Heap^.Read_Only ) then { Fix corruption }
            begin
                Write_Entry( Current, Entry ) ;
            end ;
        end ;
    end ;


    procedure Validate_Allocation ;

    var _Entry : tAllocation_Record ;
        Loop : integer ;
        Last_Allocation, This_Allocation : TStore_Address ;

        procedure Throw_Out_Allocation_Record ;

        begin
            if( Last_Allocation = Null ) then { Something wrong in first allocation record }
            begin
                if( ( Entry.Attributes and FSFF_Directory ) <> 0 ) then
                begin
                    exit ; { Leave directories alone, otherwise we can lose too much data }
                end ;
                Entry.Allocation_Info := Null ;
                if( not File_Heap^.Read_Only ) then { Fix corruption } { ELSE UNTESTED }
                begin
                    Write_Entry( Current, Entry ) ;
                end ;
            end else
            begin
                { UNTESTED: This block }
                if( not File_Heap^.Read_Only ) then { Fix corruption }
                begin
                    Read_Allocation( Last_Allocation, _Entry ) ; { Read allocation data }
                    _Entry.Next := Null ;
                    Write_Allocation( Last_Allocation, _Entry ) ;
                end ;
            end ;
            Recalculate_Size ;
        end ;

    var Have_Zero : boolean ;
        I : integer ;
        Valid : boolean ;

    begin
        { Validate allocation information... }
        Last_Allocation := Null ;
        This_Allocation := Entry.Allocation_Info ;
        while( This_Allocation <> Null ) do
        begin
            Read_Allocation( This_Allocation, _Entry ) ; { Read allocation data }
            { Validate non-sparse pointers... }
            Have_Zero := False ;
            Valid := True ;
            for Loop := 0 to Max_Cluster_In_Block do
            begin
                if( _Entry.Cluster[ Loop ] = Null ) then
                begin
                    Have_Zero := True ;
                end else
                begin
                    if( Have_Zero ) then { Had a previous zero }
                    begin
                        Correct_Entry := True ;
{$IFDEF Test}
                        Set_Error( FSErr_CorruptedFileStructure ) ;
                        exit ;
{$ENDIF}
                        Valid := False ;
                        break ; { Sparse data (probably a corruption, actually) }
                    end ;
                end ;
            end ;
            if( not Valid ) then { The whole allocation record is suspect }
            begin
                Throw_Out_Allocation_Record ;
                exit ;
            end ;

            if( Have_Zero ) then { Allocation record is not full } { ELSE UNTESTED }
            begin
                if( _Entry.Next <> Null ) then { Should not have another allocation record }
                begin
                    { UNTESTED: This block }
                    Correct_Entry := True ;
{$IFDEF Test}
                    Set_Error( FSErr_CorruptedFileStructure ) ;
                    exit ;
{$ENDIF}
                    _Entry.Next := Null ;
                    if( not File_Heap^.Read_Only ) then
                    begin
                        Write_Allocation( This_Allocation, _Entry ) ;
                    end ;
                end ;
            end ;
            for Loop := 0 to Max_Cluster_In_Block do
            begin
                if( not _Check_Pointer( _Entry.Cluster[ Loop ], File_Heap ) ) then
                begin
                    { ELSE UNTESTED: This block }
                    Correct_Entry := True ;
{$IFDEF Test}
                    Set_Error( FSErr_CorruptedFileStructure ) ;
                    exit ;
{$ENDIF}
                    if( Loop = 0 ) then { First pointer is bad }
                    begin
                        Throw_Out_Allocation_Record ;
                        exit ;
                    end ;
                    if( not File_Heap^.Read_Only ) then { ELSE UNTESTED }
                    begin
                        for I := Loop to Max_Cluster_In_Block do
                        begin
                            _Entry.Cluster[ I ] := Null ;
                        end ;
                        _Entry.Next := Null ;
                        Write_Allocation( This_Allocation, _Entry ) ;
                    end ;
                    Recalculate_Size ;
                    break ;
                end ; { if( not _Check_Pointer( _Entry[ Loop ], File_Heap ) ) }
            end ; { for Loop := 0 to Max_Cluster_In_Block }
            Last_Allocation := This_Allocation ;
            This_Allocation := _Entry.Next ;
        end ; { while( This_Allocation <> Null ) }
    end ; { .Validate_Allocation }

begin { TXFile_System.Correct_Entry }
    Correct_Entry := False ; { Assume no correction needed }

    if( Entry.Reserved[ 0 ] <> 0 ) then { Corruption }
    begin
        Correct_Entry := True ;
{$IFDEF Test}
        Directory_Entry := Null ;
        Set_Error( FSErr_CorruptedFileStructure ) ;
        exit ;
{$ENDIF}
        if( not File_Heap^.Read_Only ) then
        begin
            { Fix the directory entry... }
            fillchar( Entry.Reserved, sizeof( Entry.Reserved ), 0 ) ;
            Entry.Back_Pointer := Directory_Entry ;
            Entry.Right_Pointer := Null ;
            Entry.Left_Pointer := Null ;
            Entry.Caching_Info := Null ;
            Entry.Access_Info := Null ;
            Entry.Auditing_Info := Null ;
            Entry.Access_Stats := Null ;
            Entry.Allocation_Info := Null ;
            Write_Entry( Current, Entry ) ;
        end ;
        exit ;
    end ;
    if( Entry.Right_Pointer = Current ) then
    begin
        Correct_Entry := True ;
{$IFDEF Test}
        Directory_Entry := Null ;
        Set_Error( FSErr_CorruptedFileStructure ) ;
        exit ;
{$ENDIF}
        Entry.Right_Pointer := Null ; { Fix/ignore corruption }
        if( not File_Heap^.Read_Only ) then { Fix corruption }
        begin
            Write_Entry( Current, Entry ) ;
        end ;
    end ;
    if( Entry.Left_Pointer = Current ) then
    begin
        Correct_Entry := True ;
{$IFDEF Test}
        Set_Error( FSErr_CorruptedFileStructure ) ;
        exit ;
{$ENDIF}
        Entry.Left_Pointer := 0 ; { Fix/ignore corruption }
        if( not File_Heap^.Read_Only ) then { Fix corruption }
        begin
            Write_Entry( Current, Entry ) ;
        end ;
    end ;
    if( not Valid_Entry( Current, Entry, File_Heap, _Root ) ) then
    begin
        Correct_Entry := True ;
{$IFDEF Test}
        Set_Error( FSErr_CorruptedFileStructure ) ;
        exit ;
{$ENDIF}
        if( not File_Heap^.Read_Only ) then { Fix corruption }
        begin
            Read_Entry( Directory_Entry, Entry ) ;
            Entry.Allocation_Info := Allocate_Entry ;
            Write_Entry( Directory_Entry, Entry ) ;
            Current := Entry.Allocation_Info ;
            Read_Entry( Current, Entry ) ;
        end else
        begin
            Entry.Right_Pointer := Null ; { Ignore corruption }
        end ;
    end ;
    if( not _Check_Pointer( Entry.Right_Pointer, File_Heap ) ) then
    begin
        Correct_Entry := True ;
{$IFDEF Test}
        Set_Error( FSErr_CorruptedFileStructure ) ;
        exit ;
{$ENDIF}
        Entry.Right_Pointer := Null ; { Fix/ignore corruption }
        if( not File_Heap^.Read_Only ) then
        begin
            Write_Entry( Current, Entry ) ;
        end ;
    end ;

    if( ( Entry.Attributes and FSFF_Directory ) = 0 ) then
    begin
        Validate_Allocation ;
    end ;
end ; { TXFile_System.Correct_Entry }


function TXFile_System.Find_File( Name : string ;
    var Directory_Entry : TStore_Address ) : longint ;
{ Find specified file.  Returns the address of the file entry for the matching
  file.  Directory_Entry is ignored when passed and is the directory entry of
  the file otherwise.  If file is not found, both the result and Directory_Entry
  are 0.  If no file is specified and the path is found, the result is 0 and
  Directory_Entry is the directory entry of the path.  The returned Directory
  entry is the address of the file entry for the directory.  Name must be
  uppercase and with no leading/trailing spaces. }

var Current, Previous : longint ;
    _Directory_Entry : TStore_Address ;
    Dummy : integer ;
    Entry : TDirectory_Entry ;
    Entry_Name : string ;
    Found : boolean ;
    FS_Header : FS_Header_Record ;
    Full_Name : string ;
    _Path : string ;
    S : string ;
    Went_Right : boolean ;

begin { TXFile_System.Find_File }
    Set_Error( 0 ) ;
    if( copy( Name, 1, 1 ) = '\' ) then { Strip beginning backslash }
    begin
        Name := copy( Name, 2, 254 ) ;
    end ;

    if( ( Cached_Entries > 0 ) and ( Name = Cached_File_Name1 ) ) then { Same file we looked up last time }
    begin
        Directory_Entry := Cached_Directory_Entry1 ;
        Find_File := Cached_File_Entry1 ;
        exit ;
    end ;
    if( ( Cached_Entries > 1 ) and ( Name = Cached_File_Name2 ) ) then
    begin
        Directory_Entry := Cached_Directory_Entry2 ;
        Find_File := Cached_File_Entry2 ;
        exit ;
    end ;
    Full_Name := Name ;
    Find_File := 0 ;

    { Find directory portion... }
    Dummy := length( Name ) ;
    while( ( Dummy > 0 ) and ( Name[ Dummy ] <> '\' ) ) do
    begin
        dec( Dummy ) ;
    end ;
    _Path := copy( Name, 1, Dummy - 1 ) ;
    Found := False ;
    if( ( Cached_Entries > 0 ) and ( _Path = Cached_Path1 ) ) then
    begin
        Directory_Entry := Cached_Directory_Entry1 ;
        Name := copy( Name, Dummy + 1, length( Name ) ) ;
        Found := True ;
    end ;
    if( ( not Found ) and ( Cached_Entries > 1 ) and ( _Path = Cached_Path2 ) ) then
    begin
        Directory_Entry := Cached_Directory_Entry2 ;
        Name := copy( Name, Dummy + 1, length( Name ) ) ;
        Found := True ;
    end ;
    Read_Header( File_Heap, Boot.Header, FS_Header ) ;
    if( not Found ) then
    begin
        Directory_Entry := FS_Header.Root ; { Start at root directory }
        Read_Entry( Directory_Entry, Entry ) ;

        { Run down the directory tree until we get to the directory for this file... }
        Dummy := pos( '\', Name ) ;
        while( Dummy <> 0 ) do { Until no more directories }
        begin
            S := copy( Name, 1, Dummy - 1 ) ; { Next directory to find }
            if( not Valid_Filename( S ) ) then
            begin
                Directory_Entry := 0 ;
                exit ;
            end ;
            Name := copy( Name, Dummy + 1, 254 ) ; { Remaining name }
            Current := Entry.Allocation_Info ; { Get pointer to directory tree }
            if( Current = 0 ) then { Missing directory tree }
            begin
                if( not File_Heap^.Read_Only ) then
                begin
                    Entry.Allocation_Info := Allocate_Entry ; { Fix problem }
                end ;
                Current := Entry.Allocation_Info ;
                if( Current = 0 ) then { Cannot allocate a new entry }
                begin
                    Directory_Entry := 0 ;
                    Set_Error( FSErr_DirectoryNotFound ) ;
                    exit ;
                end ;
                Write_Entry( Directory_Entry, Entry ) ;
            end ;
            Read_Entry( Current, Entry ) ;
            Correct_Entry( Current, Directory_Entry, Entry,
                Cached_FS_Header.Root ) ;
            Current := Entry.Right_Pointer ; { Get pointer to first file }
            if( Current = 0 ) then { Nothing in this directory }
            begin
                Directory_Entry := 0 ;
                Set_Error( FSErr_DirectoryNotFound ) ;
                exit ;
            end ;
            Read_Entry( Current, Entry ) ; { Get root node }
            Entry_Name := uppercase( Entry.Name ) ;
            while( Entry_Name <> S ) do
            begin
                if( Entry_Name < S ) then { Need to follow right link }
                begin
                    if(
                        ( Entry.Right_Pointer <> 0 )
                        and
                        ( not File_Heap^.Valid_Data_Address( Entry.Right_Pointer ) )
                      ) then
                    begin
                        Entry.Right_Pointer := 0 ; { Fix corruption }
                        if( not File_Heap^.Read_Only ) then
                        begin
                            Write_Entry( Current, Entry ) ;
                        end ;
                    end ;
                    Current := Entry.Right_Pointer ;
                end else { Need to follow left link }
                begin
                    if(
                        ( Entry.Left_Pointer <> 0 )
                        and
                        ( not File_Heap^.Valid_Data_Address( Entry.Left_Pointer ) )
                      ) then
                    begin
                        Entry.Left_Pointer := 0 ; { Fix corruption }
                        if( not File_Heap^.Read_Only ) then
                        begin
                            Write_Entry( Current, Entry ) ;
                        end ;
                    end ;
                    Current := Entry.Left_Pointer ;
                end ;
                if( Current = 0 ) then { No link to follow }
                begin
                    Directory_Entry := 0 ;
                    Set_Error( FSErr_DirectoryNotFound ) ;
                    exit ;
                end ;
                Read_Entry( Current, Entry ) ; { Get next node }
                Entry_Name := uppercase( Entry.Name ) ;
            end ;
            { At this point, Entry contains the file entry for this directory }
            if( ( Entry.Attributes and FSFF_Directory ) = 0 ) then { Not a directory }
            begin
                Directory_Entry := 0 ;
                Set_Error( FSErr_InvalidPath ) ;
                exit ;
            end ;
            Directory_Entry := Current ;
            Dummy := pos( '\', Name ) ;
        end ; { while( Dummy <> 0 ) }

        { At this point, Directory_Entry is the pointer to the directory's file
          entry in its parent directory. }

        { Now find file within directory... }
        if( ( Directory_Entry = 0 ) or ( not Valid_Filename( Name ) ) ) then
        begin
            exit ;
        end ;
    end ; { if( not Found ) }

    if( length( Name ) = 0 ) then { No file passed - only a directory }
    begin
        Find_File := 0 ;
        exit ;
    end ;

    { Get directory name entry... }
    Read_Entry( Directory_Entry, Entry ) ;

    { Follow link to subdirectory tree... }
    if( Entry.Allocation_Info = 0 ) then { This shouldn't happen - it indicates corruption }
    begin
        if( ( not File_Heap^.Read_Only ) and ( Entry.Allocation_Info <> 0 ) ) then
        begin
            { Fix the situation... }
            Entry.Allocation_Info := File_Heap^.Allocate( Entry_Size ) ;
            Write_Entry( Directory_Entry, Entry ) ; {~~~}
        end ;
        Set_Error( FSErr_FileNotFound ) ;
        exit ;
    end ;
    _Directory_Entry := Directory_Entry ;
    Directory_Entry := Entry.Allocation_Info ; { Get pointer to directory tree }

    { Get root node of subdirectory tree... }
    Read_Entry( Directory_Entry, Entry ) ;
    if( Correct_Entry( Directory_Entry, _Directory_Entry, Entry, FS_Header.Root ) ) then { Corrupted entry }
    begin
        Directory_Entry := 0 ;
        Set_Error( FSErr_FileNotFound ) ;
        exit ;
    end ;

    { Follow pointer to first file in subdirectory tree... }
    Current := Entry.Right_Pointer ; { Get pointer to first file }
    if( Current = 0 ) then { Nothing in this directory }
    begin
        Set_Error( FSErr_FileNotFound ) ;
        exit ;
    end ;

    { Follow pointer to first node in tree... }
    Read_Entry( Current, Entry ) ; { Get first node }
    if( Entry.Reserved[ 0 ] <> 0 ) then { Corrupted entry }
    begin
        if( not File_Heap^.Read_Only ) then
        begin
            { Fix the directory entry... }
            Read_Entry( Directory_Entry, Entry ) ;
            Entry.Right_Pointer := 0 ;
            Write_Entry( Directory_Entry, Entry ) ;
        end ;
        Set_Error( FSErr_FileNotFound ) ;
        exit ;
    end ;

    { Traverse subtree, looking for file... }
    Entry_Name := uppercase( Entry.Name ) ;
    while( Entry_Name <> Name ) do
    begin
        Previous := Current ;
        if( Entry_Name < Name ) then { Need to follow right link }
        begin
            Current := Entry.Right_Pointer ;
            Went_Right := True ;
        end else { Need to follow left link }
        begin
            Current := Entry.Left_Pointer ;
            Went_Right := False ;
        end ;
        if( Current = 0 ) then { No link to follow }
        begin
            Directory_Entry := 0 ;
            Set_Error( FSErr_FileNotFound ) ;
            exit ;
        end ;
        Read_Entry( Current, Entry ) ; { Get next node }
        if(
            ( Entry_Name = uppercase( Entry.Name ) ) { Duplicate name }
            or
            ( Correct_Entry( Current, Directory_Entry, Entry, Cached_FS_Header.Root ) ) { Entry needs correction }
          ) then
        begin
            if( not File_Heap^.Read_Only ) then
            begin
                Current := Previous ;
                Read_Entry( Current, Entry ) ; { Get previous node }
                if( Went_Right ) then
                begin
                    Entry.Right_Pointer := 0 ;
                end else
                begin
                    Entry.Left_Pointer := 0 ;
                end ;
                Write_Entry( Current, Entry ) ; { Correct the node }
            end ;
        end ;
        Entry_Name := uppercase( Entry.Name ) ;
    end ; { while( Entry_Name <> Name )  }
    Find_File := Current ; { Pointer to this file's directory entry }
    Dummy := Cached_Entries ;
    if( Dummy > 2 ) then
    begin
        Dummy := 0 ;
    end else
    begin
        inc( Cached_Entries ) ;
    end ;
    if( Dummy = 0 ) then
    begin
        Cached_File_Name1 := Full_Name ;
        Cached_Directory_Entry1 := _Directory_Entry ;
        Cached_File_Entry1 := Current ;
        Cached_Path1 := _Path ;
    end else
    begin
        Cached_File_Name2 := Full_Name ;
        Cached_Directory_Entry2 := _Directory_Entry ;
        Cached_File_Entry2 := Current ;
        Cached_Path2 := _Path ;
    end ;
    Directory_Entry := _Directory_Entry ;
end ; { TXFile_System.Find_File }


procedure TXFile_System.Fix_Back_Pointer( Previous, Current : TStore_Address ;
    var Entry : TDirectory_Entry ) ;

    function _Valid_Filename( X : string ) : boolean ;

    begin
        Result := Valid_Filename( X ) ;
    end ;

    function Invalid_Directory_Entry : boolean ;

    begin
        Result := False ; { Assume no correction needed }

        if( Entry.Reserved[ 0 ] <> 0 ) then { Corruption }
        begin
            Result := True ;
            exit ;
        end ;
        if( Entry.Right_Pointer = Current ) then
        begin
            Result := True ;
            exit ;
        end ;
        if( Entry.Left_Pointer = Current ) then
        begin
            Result := True ;
            exit ;
        end ;
        if( not Valid_Entry( Current, Entry, File_Heap, Cached_FS_Header.Root ) ) then
        begin
            Result := True ;
            exit ;
        end ;
        if(
            ( Entry.Right_Pointer <> 0 )
            and
            ( not File_Heap^.Valid_Data_Address( Entry.Right_Pointer ) )
          ) then
        begin
            Result := True ;
        end ;
    end ;

var S : string ;
    Temp : TDirectory_Entry ;

begin
    { First check for a cross-link... }
    if(
        File_Heap^.Valid_Data_Address( Entry.Back_Pointer ) { Valid pointer }
        and
        ( ( Entry.Back_Pointer and 255 ) = 0 ) { Valid (potential) directory entry }
      ) then
    begin
        Read_Entry( Entry.Back_Pointer, Temp ) ;
        if( not Invalid_Directory_Entry ) then
        begin
            S := Uppercase( Temp.Name ) ;
            if(
                (
                  _Valid_Filename( Temp.Name )
                  and
                  (
                    ( ( Temp.Right_Pointer = Current ) and ( S < Uppercase( Entry.Name ) ) )
                    or
                    ( ( Temp.Left_Pointer = Current ) and ( S > Uppercase( Entry.Name ) ) )
                  )
                )
                or
                (
                  ( Temp.Allocation_Info = Current )
                  and
                  ( Temp.Name[ 0 ] = #0 )
                )
              ) then { Looks like a valid parent node }
            begin
                {TODO: Follow chain up to root directory to be sure }

                { Remove the second link to this entry }
                Read_Entry( Previous, Temp ) ;
                if( Temp.Left_Pointer = Current ) then
                begin
                    Temp.Left_Pointer := 0 ;
                end else
                if( Temp.Right_Pointer = Current ) then
                begin
                    Temp.Right_Pointer := 0 ;
                end ;
                Write_Entry( Previous, Temp ) ;
                exit ;
            end ; { if }
        end ; { if( not Invalid_Directory_Entry ) }
    end ; { if }

    { Correct the entry... }
    Entry.Back_Pointer := Previous ;
    Write_Entry( Current, Entry ) ;
end ; { TXFile_System.Fix_Back_Pointer }


function TXFile_System.Get_Absolute( Allocation, Offset : longint ) : longint ;
{ Translate a file-relative offset into a file-system address.
  Allocation = offset of first allocation record for the file.
  Offset = Cluster offset within file. }

var Allocation_Record : TAllocation_Record ;

begin
    File_Heap^.Read_Heap( Allocation, Entry_Size, Allocation_Record ) ;
    while( Offset > Max_Cluster_In_Block ) do
    begin
        Offset := Offset - Max_Cluster_In_Block - 1 ;
        Allocation := Allocation_Record.Next ;
        File_Heap^.Read_Heap( Allocation, Entry_Size, Allocation_Record ) ;
        if( Allocation = 0 ) then
        begin
            Get_Absolute := 0 ; { Beyond end of file } { TEST: Untested }
            exit ;
        end ;
    end ;
    Get_Absolute := Allocation_Record.Cluster[ Offset ] ;
    if(
        ( Allocation_Record.Cluster[ Offset ] <> 0 )
        and
        ( not File_Heap^.Valid_Data_Address( Allocation_Record.Cluster[ Offset ] ) )
      ) then
    begin
        Get_Absolute := 0 ; { Beyond end of file }
    end ;
end ;


function TXFile_System.Get_File_Rec( Handle : integer ) : PNative_File_Record ;

var File_Rec : PNative_File_Record ;

begin
    { Check for errors... }
    dec( Handle ) ; { Handles are zero-based }
    if( ( Handle < 0 ) or ( Handle >= File_List.Count ) ) then
    begin
        Get_File_Rec := nil ;
        Set_Error( FSErr_Invalid_Handle ) ;
        exit ;
    end ;
    File_Rec := PNative_File_Record( File_List.At( Handle ) ) ;
    Get_File_Rec := File_Rec ;
    if( File_Rec = nil ) then
    begin
        Set_Error( FSErr_FileNotOpen ) ;
        exit ;
    end ;
end ;


procedure TXFile_System.Read_Header( _File_Heap : PFH ;
    Address : TStore_Address ; var Header : FS_Header_Record ) ;

begin
    Verify_Lock ;
    if( Locked ) then
    begin
        Header := Cached_FS_Header ;
    end else
    begin
        _File_Heap^.Read_Heap( Address, sizeof( Header ), Header ) ;
    end ;
end ;


procedure TXFile_System.Write_Header( _File_Heap : PFH ;
    Address : TStore_Address ; var Header : FS_Header_Record ) ;

begin
    if( ( Flags and FSF_Read_Only ) <> 0 ) then
    begin
        Set_Error( FSErr_Read_Only ) ;
        exit ;
    end ;
    Verify_Lock ;
    if( Locked ) then
    begin
        Cached_FS_Header := Header ; { Copy to cached header }
    end ;
    _File_Heap^.Write_Heap( Address, sizeof( Header ), Header ) ;
end ;


procedure TXFile_System.Read_Entry( Address : TStore_Address ;
    var Entry : TDirectory_Entry ) ;

begin
    File_Heap^.Read_Heap( Address, Entry_Size, Entry ) ;
end ;


procedure TXFile_System.Write_Entry( Address : TStore_Address ;
    var Entry : TDirectory_Entry ) ;

begin
    File_Heap^.Write_Heap( Address, Entry_Size, Entry ) ;
end ;


procedure TXFile_System.Read_Allocation( Address : TStore_Address ;
    var Entry : TAllocation_Record ) ;

begin
    File_Heap^.Read_Heap( Address, Entry_Size, Entry ) ;
end ;


procedure TXFile_System.Write_Allocation( Address : TStore_Address ;
    var Entry : TAllocation_Record ) ;

begin
    File_Heap^.Write_Heap( Address, Entry_Size, Entry ) ;
end ;


procedure TXFile_System.Set_Back_Pointer( Address, New_Back : longint ) ;
{ Set back pointer of record at Address to New_Back. }

var Temp : TDirectory_Entry ;

begin
    Read_Entry( Address, Temp ) ;
    Temp.Back_Pointer := New_Back ;
    Write_Entry( Address, Temp ) ;
end ;


function TXFile_System.Tree_Find( PTR : longint ; Key : string ) : longint ; { TEST: untested }

label Loop ;

var Current : longint ;
    Entry : TDirectory_Entry ;
    Entry_Name : string ;
    Uppercase_Key : string ;

begin
    Current := Ptr ; { Start at root }
    Uppercase_Key := uppercase( Key ) ; { Non-case sensitive }
Loop:
    if( Current <> 0 ) then { not yet at end }
    begin
        Read_Entry( Current, Entry ) ;
        Entry_Name := Uppercase( Entry.Name ) ;
        if( Entry_Name = Uppercase_Key ) then { Found item in tree }
        begin
            Tree_Find := Current ;
            Exit ;
        end ;
        if( Entry_Name > Uppercase_Key ) then
        begin
            Current := Entry.Left_Pointer
        end else
        begin
            Current := Entry.Right_Pointer ; { Follow link }
        end ;
        goto Loop ;
    end ;
    Tree_Find := 0 ; { Item not found }
end ; { TXFile_System.Tree_Find }


procedure TXFile_System.Tree_Add( var PTR : TStore_Address ;
    Entry : TDirectory_Entry ; Existing_Ptr : TStore_Address ) ;
{ Add a file to the directory tree, whose root is at PTR.  The file is described
  in the Entry record.  The address of the file entry is returned in PTR. }

var Current, Previous : longint ;
    Entry_Name, This_Entry_Name, Previous_Entry_Name : string ;
    Previous_Entry, This_Entry : TDirectory_Entry ;

begin
    { Sanity check }
    if( PTR = 0 ) then
    begin
        Set_Error( FSErr_Internal_Error_On_Add ) ;
        exit ;
    end ;

    Previous := 0 ; { Get rid of compiler warning }
    Current := Ptr ; { Start at root }
    Entry_Name := Uppercase( Entry.Name ) ;
    while( Current <> 0 ) do { Until at end }
    begin
        Read_Entry( Current, This_Entry ) ;
        This_Entry_Name := Uppercase( This_Entry.Name ) ;
        if( Entry_Name = This_Entry_Name ) then { Found item in tree }
        begin
            Set_Error( FSErr_Internal_Duplicate ) ;
            exit ;
        end ;
        Previous := Current ; { Save current pointer as previous }
        Previous_Entry := This_Entry ;
        Previous_Entry_Name := This_Entry_Name ;
        if( Entry_Name < This_Entry_Name ) then
        begin
            Current := This_Entry.Left_Pointer ;
        end else
        begin
            Current := This_Entry.Right_Pointer ; { Follow link }
        end ;
    end ;
    if( Existing_Ptr = 0 ) then
    begin
        Current := Allocate_Entry ; { Create node }
    end else
    begin
        Current := Existing_Ptr ; { Node already exists on store }
    end ;
    if( Previous_Entry_Name > Entry_Name ) then
    begin
        Previous_Entry.Left_Pointer := Current
    end else
    begin
        Previous_Entry.Right_Pointer := Current ;
    end ;
    Write_Entry( Previous, Previous_Entry ) ;
    Entry.Left_Pointer := 0 ;
    Entry.Right_Pointer := 0 ;

    { In a pristine file system, the root directory "\" points to the first
      unallocated address of memory.  The first file created will be assigned
      this address.  So, if the newly allocated address (Current) is equal to
      the current tree pointer (Previous), we do nothing, otherwise the new
      entry would point back to itself, which is an infinite loop waiting to
      happen. }
    if( Previous <> Current ) then { TEST: untested (else condition) }
    begin
        Entry.Back_Pointer := Previous ;
    end ;
    Write_Entry( Current, Entry ) ;
    PTR := Current ; { Return pointer to file entry }
end ; { TXFile_System.Tree_Add }


procedure TXFile_System.Tree_Delete( Index : TStore_Address ) ;

var Entry, Other : TDirectory_Entry ;
    Tmp : longint ;

begin
    Clear_Cache ;
    Read_Entry( Index, Entry ) ;

    if( ( Entry.Right_Pointer <> 0 ) and ( Entry.Left_Pointer = 0 ) ) then
    { Right branch but no left branch }
    begin
        { Deleting node C (case 1):
          before     after
            A           A
           / \         / \
          x   C       x   D
             / \
            0   D

          Deleting node B (case 2):
          before     after
              D        D
             / \      / \
            B   x    C   x
           / \
          0   C
        }
        Set_Back_Pointer( Entry.Right_Pointer, Entry.Back_Pointer ) ;
        Read_Entry( Entry.Back_Pointer, Other ) ;
        if( Other.Left_Pointer = Index ) then
        begin
            Other.Left_Pointer := Entry.Right_Pointer ; { Case 2 }
        end else
        begin
            Other.Right_Pointer := Entry.Right_Pointer ; { Case 1 }
        end ;
        Write_Entry( Entry.Back_Pointer, Other ) ;
        File_Heap^.Deallocate( Index, Entry_Size ) ; { Delete node }
        exit ;
    end ; { if }

    if( ( Entry.Left_Pointer = 0 ) and ( Entry.Right_Pointer = 0 ) ) then
    { Terminal node }
    begin
        { Deleting node C (case 3):
          before     after
            A          A
           / \        / \
          x   C      x   0
             / \
            0   0

          Deleting node B (case 4):
          before     after
              D        D
             / \      / \
            B   x    0   x
           / \
          0   0
        }
        { If we get here when Entry.Back_Pointer is 0, it means that Entry is
          the only entry in its directory. }
        if( Entry.Back_Pointer = 0 ) then
        begin
            { TODO: Fix parent directory so it isn't pointing to a deleted entry
              anymore. }
        end else
        begin
            Read_Entry( Entry.Back_Pointer, Other ) ;
            if( Other.Left_Pointer = Index ) then
            begin
                Other.Left_Pointer := 0 ; { Case 4 }
            end else
            begin
                Other.Right_Pointer := 0 ; { Case 3 }
            end ;
            Write_Entry( Entry.Back_Pointer, Other ) ;
        end ;
        File_Heap^.Deallocate( Index, Entry_Size ) ; { Delete node }
        exit ;
    end ; { if }

    if( ( Entry.Right_Pointer = 0 ) and ( Entry.Left_Pointer <> 0 ) ) then
    { Left branch but no right branch }
    begin
        { Deleting node C (case 5):
          before     after
            A           A
           / \         / \
          x   C       x   B
             / \
            B   0

          Deleting node C (case 6):
          before     after
              D        D
             / \      / \
            C   x    B   x
           / \
          B   0
        }
        Set_Back_Pointer( Entry.Left_Pointer, Entry.Back_Pointer ) ;
        Read_Entry( Entry.Back_Pointer, Other ) ;
        if( Other.Left_Pointer = Index ) then
        begin
            Other.Left_Pointer := Entry.Left_Pointer ; { Case 6 } { TEST: Untested }
        end else
        begin
            Other.Right_Pointer := Entry.Left_Pointer ; { Case 5 }
        end ;
        Write_Entry( Entry.Back_Pointer, Other ) ;
        File_Heap^.Deallocate( Index, Entry_Size ) ; { Delete node }
        exit ;
    end ; { if }

    { Bring up left branch... }
    { Deleting node C (case 7):
      before       after
        A            A
       / \          / \
      x   C        x   E
         / \          / \
        B   E        D   F
           / \      /
          D   F    B

      Deleting node B (case 8):
       before        after
          E            E
         / \          / \
        B   x        D   x
       / \          / \
      A   D        C   y
         / \      /
        C   y    A

     (move left branch of deleted node to far left of deleted node's right node)
    }
    { First, link right node to parent of deleted node (thus deleting it from
      the tree)... }
    Set_Back_Pointer( Entry.Right_Pointer, Entry.Back_Pointer ) ;
    Read_Entry( Entry.Back_Pointer, Other ) ;
    if( Other.Left_Pointer = Index ) then
    begin
        Other.Left_Pointer := Entry.Right_Pointer ; { case 8 }
    end else
    begin
        Other.Right_Pointer := Entry.Right_Pointer ; { case 7 } { TEST : untested }
    end ;
    Write_Entry( Entry.Back_Pointer, Other ) ;

    { Now move deleted node's left branch to the far left of new branch... }
    Tmp := Entry.Right_Pointer ; { Start at replacement node }
    Read_Entry( Tmp, Other ) ;
    while( Other.Left_Pointer <> 0 ) do
    begin
        Tmp := Other.Left_Pointer ; { Go all the way left } { TEST: Untested }
        Read_Entry( Tmp, Other ) ;
    end ;
    Other.Left_Pointer := Entry.Left_Pointer ;
    Write_Entry( Tmp, Other ) ;
    Set_Back_Pointer( Entry.Left_Pointer, Tmp ) ;

    { Now physically delete the node from the store }
    File_Heap^.Deallocate( Index, Entry_Size ) ; { Actually delete node }
end ; { TXFile_System.Tree_Delete }


function TXFile_System.Tree_Next( Index : longint ) : longint ; { TEST: Untested }

label Loop ;

var Entry : TDirectory_Entry ;
    Previous : longint ;

begin
    Read_Entry( Index, Entry ) ;
    if( Entry.Right_Pointer = 0 ) then
    begin
Loop:
        if( Entry.Back_Pointer = 0 ) then { Cannot go further up tree }
        begin
            Tree_Next := 0 ; { No more items in tree }
            Exit ;
        end ;
        Previous := Index ; { Save previous position }
        Index := Entry.Back_Pointer ; { Backup one level }
        Read_Entry( Index, Entry ) ;
        if( length( Entry.Name ) = 0 ) then
        begin
            Tree_Next := 0 ;
            exit ;
        end ;
        if( Entry.Left_Pointer = Previous ) then { Came from left branch }
        begin
            Tree_Next := Index ;
            Exit ;
        end else
        begin
            goto Loop ; { Backup again }
        end ;
    end ;
    Index := Entry.Right_Pointer ; { Go down right branch }
    while( Entry.Left_Pointer <> 0 ) do
    begin
        Index := Entry.Left_Pointer ; { Go left }
        Read_Entry( Index, Entry ) ;
    end ;
    Tree_Next := Index ; { Return result }
end ; { TXFile_System.Tree_Next }


procedure TXFile_System.Repair_Tree( Directory_Entry, Current : TStore_Address ) ;

var Entry, Back_Entry, Dir_Entry : TDirectory_Entry ;
    Left, Right, Temp : TStore_Address ;

begin
    Clear_Cache ;
    Read_Entry( Current, Entry ) ;

    if( Entry.Back_Pointer <> 0 ) then { Disconnect from parent node }
    begin
        Read_Entry( Entry.Back_Pointer, Back_Entry ) ;
        if( Back_Entry.Left_Pointer = Current ) then
        begin
            Back_Entry.Left_Pointer := 0 ;
        end else
        if( Back_Entry.Right_Pointer = Current ) then
        begin
            Back_Entry.Right_Pointer := 0 ;
        end ;
        Write_Entry( Entry.Back_Pointer, Back_Entry ) ; { Snip }
    end ;

    { Save and clear our pointers... }
    Left := Entry.Left_Pointer ;
    Right := Entry.Right_Pointer ;
    Entry.Left_Pointer := 0 ;
    Entry.Right_Pointer := 0 ;
    Entry.Back_Pointer := 0 ;
    Write_Entry( Current, Entry ) ; { Snip, snip }

    { The directory entry passed in is the directory file entry, so move to it's
      root before adding back to the tree. }
    Read_Entry( Directory_Entry, Back_Entry ) ;
    Directory_Entry := Back_Entry.Allocation_Info ;

    { Add back to directory tree... }
    Set_Error( 0 ) ;
    Temp := Directory_Entry ; { Tree_Add changes the first parameter }
    Tree_Add( Temp, Entry, Current ) ;
    if( IO_Error.Code = FSErr_Internal_Duplicate ) then
    begin
        { This is an extra copy of the file, so throw it away }
        if( ( Entry.Attributes and FSFF_Directory ) <> 0 ) then { Deleting a directory }
        begin
            Read_Entry( Entry.Allocation_Info, Dir_Entry ) ;
            if( Dir_Entry.Right_Pointer <> 0 ) then { Directory has something in it }
            begin
                { TODO: We need to handle the files in the directory }
                exit ;
            end ;
            File_Heap^.Deallocate( Current, Entry_Size ) ; { Delete node }
        end else { Normal file }
        begin
            Delete_File( Current ) ;
        end ; { if }
    end ; { if( IO_Error.Code = FSErr_Internal_Duplicate ) }
    if( Left <> 0 ) then
    begin
        Repair_Tree( Directory_Entry, Left ) ;
    end ;
    if( Right <> 0 ) then
    begin
        Repair_Tree( Directory_Entry, Right ) ;
    end ;

    { Make sure no bad cache is left around }
    Clear_Cache ;
end ; { TXFile_System.Repair_Tree }


procedure TXFile_System.Set_Error( X : integer ) ;

begin
    IO_Error.Facility := FSErr_Facility ;
    IO_Error.Code := X ;
end ;


procedure TXFile_System.Verify_Lock ;

var _Time : longint ;

begin
    if( not Locked ) then
    begin
        exit ;
    end ;
    _Time := Cached_FS_Header.Lock_Time ;
    if( Time_Seconds < _Time ) then { Went past midnight }
    begin
        _Time := _Time - 24 * 60 * 6000 ;
    end ;
    if( Time_Seconds >= _Time + Lock_Wait_Period - 1 ) then { Lock expired }
    begin
        Locked := False ;
    end ;
end ;


{ TNative_File methods... }

{ API... }

function TNative_File.Facility : longint ;

begin
    Facility := TNative_FileErr_Facility ;
end ;


function TNative_File.Version : longint ;

begin
    Version := 10 ; { Version 1.0 }
end ;


{ Store heap variables.  These variables simply make life easier. }

var SH_Buffer : array[ 0..127 ] of byte ; { 128 bytes of zero (0). }


{$IFDEF Debug}
procedure Check_FH( O : TFH ) ;

begin
    if( O.Sanity <> 'TFH' ) then
    begin
        raise Exception.Create( 'Invalid type-cast' ) ;
        halt ;
    end ;
end ;
{$ENDIF}

{ TSFH_Debugger... }

type PSFH_Debugger = ^TSFH_Debugger ;
     TSFH_Debugger = object( TSText_Debugger )
                          private
                              _Instance : PFH ;
                              S : Small_String ;

                          public
                              constructor Init( I : PFH ) ;

                              function Count : longint ; virtual ;

                              function Child( Ordinal : longint ) : PSDebug_Interface ;
                                  virtual ;
                      end ;

constructor TSFH_Debugger.Init( I : PFH ) ;

begin
    inherited Init ;

    _Instance := I ;
end ;


function TSFH_Debugger.Count : longint ;

begin
    Count := 14 ;
end ;


function TSFH_Debugger.Child( Ordinal : longint ) : PSDebug_Interface ;

var R : PSText_Debugger ;

begin
    new( R, Init ) ;
    case Ordinal of
        0 : S := 'Allow_Cache = ' + Boolean_To_String( _Instance^.Allow_Cache ) ;
        1 : begin
                dispose( R, done ) ;
                R := PSText_Debugger( _Instance^.AT.Debugger ) ;
                if( R = nil ) then
                begin
                    new( R, Init ) ;
                end ;
                S := 'AT' ;
            end ;
        2 : S := 'Byte_Mapping = ' + Num1( _Instance^.Byte_Mapping ) ;
        3 : S := 'Device = ' + Num1( _Instance^.Device ) ;
        4 : S := 'Extension_Resolution = ' + Num1( _Instance^.Extension_Resolution ) ;
        5 : begin
                 dispose( R, done ) ;
                 if( _Instance^.FH_Store = nil ) then
                 begin
                     R := nil ;
                 end else
                 begin
                     R := PSText_Debugger( _Instance^.FH_Store^.Debugger ) ;
                 end ;
                 if( R = nil ) then
                 begin
                     new( R, Init ) ;
                 end ;
                 S := 'FH_Store = ' + Num1( longint( _Instance^.FH_Store ) ) ;
             end ;
        6 : S := 'Flags = ' + Num1( _Instance^.Flags ) ;
        7 : S := 'Lowest_Dynamic = ' + Num1( _Instance^.Lowest_Dynamic ) ;
        8 : begin
                 dispose( R, done ) ;
                 if( _Instance^.Memory = nil ) then
                 begin
                     R := nil ;
                 end else
                 begin
                     R := PSText_Debugger( _Instance^.Memory^.Debugger ) ;
                 end ;
                 if( R = nil ) then
                 begin
                     new( R, Init ) ;
                 end ;
                 S := 'Memory = ' + Num1( longint( _Instance^.Memory ) ) ;
             end ;
        9 : S := 'Own_Allocator = ' + Boolean_To_String( _Instance^.Own_Allocator ) ;
        10 : S := 'Resolution = ' + Num1( _Instance^.Resolution ) ;
        11 : S := 'Serial_Number = ' + Num1( _Instance^.Serial_Number ) ;
        12 : begin
                dispose( R, done ) ;
                new( R, Init ) ;
                S := 'Statistics' ;
            end ;
        13 : S := 'Track_FH = ' + Boolean_To_String( _Instance^.Track_FH ) ;
    end ;
    S := S + #0 ;
    R^.Set_Title( @S[ 1 ] ) ;
    Child := PSDebug_Interface( R ) ;
end ; { TFH_Debugger.Child }



{ TFH (store heap) methods... }

{ Internal utility routines }

procedure TFH.Log_Error( C, S : integer ) ;

var
{$IFDEF DELPHI}
    R : tUEC ;
{$ELSE}
    R : double ;
{$ENDIF}
    UEC : tUEC ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Last_Error.Facility := FHErr_Facility ;
    Last_Error.Code := C ;
    R := Error_Code( FHErr_Facility, C ) ;
    UEC := tUEC( R ) ;
    ErrMastr.Log_Error( UEC, '', S ) ;
    inc( Store_Statistics.Error_Count ) ;
end ;


{ API... }

procedure TFH.Determine_AT_Metrics( var UEC : TUEC ) ;

var Store_Size : TStore_Size ;
{    Dumm : byte ;}
{$IFNDEF DELPHI}
    Dummy : TRecord_Size ;
{$ENDIF}
    Header : FH_Header ;
(*
    Size : TRecord_Size ;
    Temp : TStore_Size ;
*)

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    fillchar( UEC, sizeof( UEC ), 0 ) ;
    Store_Size := FH_Store^.Max_Storage ;
    if( ( Flags and FHF_NoExtend ) <> 0 ) then { Static store }
    begin
        AT.Set_Offset( Store_Size - ( Store_Size div ( Byte_Mapping + 1 ) ) ) ;
{$IFDEF DELPHI}
        AT.Set_Size( Store_Size - AT.Get_Offset ) ; { Size of allocation table }
{$ELSE}
        AT.Set_Size( trunc( Store_Size - AT.Get_Offset ) ) ; { Size of allocation table }
{$ENDIF}
        AT.Set_Modified( False ) ; { We are matching the in-memory AT to the on-store AT, so nothing is modified }
{$IFNDEF DELPHI}
        Dummy :=
{$ENDIF}
            FH_Store^.Read_Data( Header, Resolution, Sizeof( Header ), UEC ) ;
        if( UEC.Code <> 0 ) then
        begin
            exit ;
        end ;
        if(
            ( Header.AT_Size = 0 ) { Invalid AT size }
            or
            ( Header.AT_Size >= Store_Size ) { Ditto }
            or
            ( Header.AT_Offset < Resolution * 2 ) { Invalid AT offset }
            or
            ( ( Header.AT_Offset_High <> 0 ) and ( sizeof( Store_Size ) = 4 ) )
            or
            ( Header.ID <> -1 ) { Not a valid header }
            or
            ( Header.ID1 <> 1 ) { Ditto }
          ) then
        begin
            AT.Set_Size( Store_Size div Byte_Mapping ) ;
            AT.Set_Offset( Store_Size div 2 ) ; { Center the AT in the fixed store }
{$IFDEF DELPHI}
            AT.Set_Offset( ( ( AT.Get_Offset + Resolution - 1 ) div Resolution ) * Resolution ) ; { Align on resolution }
{            Temp := AT.Get_Offset ;}
{$ELSE}
            AT.Set_Offset( ( trunc( AT.Get_Offset + Resolution - 1 ) div Resolution ) * Resolution ) ; { Align on resolution }
{            Temp := trunc( AT.Get_Offset ) ;}
{$ENDIF}
(* EAC, 10-Aug-2002 - Shouldn't write to store yet - we may just be seeing if this IS a valid file system
            Size := Resolution ;
            if( Size > sizeof( SH_Buffer ) ) then
            begin
                Size := sizeof( SH_Buffer ) ;
            end ;
            for Dummy := 0 to AT.Get_Size div Size do
            begin
                FH_Store^.Write_Data( SH_Buffer, Temp, Size, UEC ) ;
                Temp := Temp + Size ;
            end ;
            Dumm := 3 ;
{$IFDEF DELPHI}
            Dummy := FH_Store^.Write_Data( Dumm, AT.Get_Offset, 1, UEC ) ; { Root and header are preallocated }
{$ELSE}
            Dummy := FH_Store^.Write_Data( Dumm, trunc( AT.Get_Offset ), 1, UEC ) ; { Root and header are preallocated }
{$ENDIF}
            Update_Header ;
*)
        end else
        begin
            AT.Set_Offset( Header.AT_Offset ) ;
            AT.Set_Size( Header.AT_Size ) ;
        end ;
    end else
    begin
        { Dynamic store... }
        AT.Set_Offset( Store_Size - ( Store_Size div ( Byte_Mapping + 1 ) ) ) ;
{$IFDEF DELPHI}
        AT.Set_Size( Store_Size - AT.Get_Offset ) ; { Size of allocation table }
{$ELSE}
        AT.Set_Size( trunc( Store_Size - AT.Get_Offset ) ) ; { Size of allocation table }
{$ENDIF}
        if( AT.Get_Offset > 0 ) then { Heap was previously written to }
        begin
{$IFNDEF DELPHI}
            Dummy :=
{$ENDIF}
                FH_Store^.Read_Data( Header, Resolution, Sizeof( Header ), UEC ) ;
            AT.Set_Offset( Header.AT_Offset ) ;
            AT.Set_Size( Header.AT_Size ) ;
            if( Store_Size > AT.Get_Offset + AT.Get_Size ) then
            begin
{$IFDEF DELPHI}
                FH_Store^.Set_Max_Storage( AT.Get_Offset + AT.Get_Size, UEC ) ;
{$ELSE}
                FH_Store^.Set_Max_Storage( trunc( AT.Get_Offset + AT.Get_Size ), UEC ) ;
{$ENDIF}
            end ;
        end ;
    end ;
end ; { TFH.Determine_AT_Metrics }


constructor TFH.Init( Minimum, F : integer ; const Name : string ;
    var _Result : tUEC ; Mem : PSManaged_Store ) ;
{ Initialize for FH access.  Note that the requested file MUST exist.  _Result
  is result of initialization.  if the file is non-zero-length or a device is
  being used, the file/device must have already been initialized with FH format.
  F is the initialization flags.  Name is the file name.  Minimum is the
  allocation resolution.  F is flags. }

begin
    inherited Init ;
    AT.Init( 16 ) ;

    { Open file/device }
    if( ( length( Name ) = 2 ) and ( Name[ 2 ] = ':' ) ) then { Device }
    begin
        FH_Store := System.New( PBlock_Device, Init ) ;
        PBlock_Device( FH_Store )^.Reset( Name ) ;
        if( PBlock_Device( FH_Store )^.IO_Error.Code <> 0 ) then
        begin
            _Result := PBlock_Device( FH_Store )^.IO_Error ;
            Last_Error := _Result ;
            exit ;
        end ;
        Minimum := FH_Store^.Min_Storage ;
        Device := -1 ; { Flag this as a device rather than a file }
    end else
    begin
        FH_Store := System.New( PSFile, Init( nil, false, 1 ) ) ;
        if( ( F and FHF_Read_Only ) <> 0 ) then
        begin
            PSFile( FH_Store )^.File_Mode := FSOM_Read_Only ;
        end ;
        PSFile( FH_Store )^.Reset( uppercase( Explicit( Name ) ) ) ;
        if( PSFile( FH_Store )^.IO_Error.Code  <> 0 ) then
        begin
            _Result := PSFile( FH_Store )^.IO_Error ;
            Last_Error := _Result ;
            exit ;
        end ;
        if( ( F and FHF_NoExtend ) <> 0 ) then
        begin
            PSFile( FH_Store )^.Extendable := False ;
        end ;
        Add_To_FH_List( @self ) ;
    end ;
    Device := 0 ;
    if( length( Name ) > 1 ) then
    begin
        if( Name[ 2 ] = ':' ) then
        begin
            Device := ord( Name[ 1 ] ) - 64 ;
        end ;
    end ;

    Common_Init( Minimum, F, FH_Store, _Result, Mem ) ;
end ; { TFH.Init }


constructor TFH.Init_On_Store( Minimum, F : integer ; S : PSStore ;
    var _Result : tUEC ; Mem : PSManaged_Store ) ;

begin
    inherited Init ;

    AT.Init( 16 ) ;
    Device := 0 ;
    FH_Store := nil ;
    Common_Init( Minimum, F, S, _Result, Mem ) ;
end ;


procedure TFH.Common_Init( Minimum, F : integer ; S : PSStore ;
    var _Result : tUEC ; Mem : PSManaged_Store ) ;

var Dummy : integer ;
    Found : boolean ;

begin
{$IFDEF Debug}
    Sanity := 'TFH' ;
{$ENDIF}

{$IFDEF Verify}
    Address_List := TInteger_List.Create ;
    Size_List := TInteger_List.Create ;
{$ENDIF}
    Memory := Mem ;
    Own_Allocator := False ;
    fillchar( Statistics, sizeof( Statistics ), 0 ) ;
    Allow_Cache := True ;
    Extension_Resolution := 0 ; { Extend only by amount necessary }
    Lowest_Dynamic := 0 ; { Allow write access to entire heap }

    Flags := F ;
    _Result.Facility := FHErr_Facility ;
    _Result.Code := FHErr_Bad_Minimum ;
    Last_Error := _Result ;
    Found := false ;
    FH_Store := nil ;
    if( Minimum = 0 ) then
    begin
        Minimum := 16 ; { Default resolution }
    end ;
    for Dummy := 0 to 15 do
    begin
        if( Minimum = Bit_Values[ Dummy ] ) then
        begin
            Found := true ;
        end ;
    end ;
    { See if a power of 2 from 1 to 32768. }

    if( not Found ) then
    begin
        exit ;
    end ;

    FH_Store := S ;
    FH_Store^.Attach ;

    if( FH_Store^.Extend( 0 ) = -1 ) then
    begin
        Flags := Flags or FHF_NoExtend ; { Store is not expandable }
    end ;
    if( FH_Store^.Read_Only ) then
    begin
        Flags := Flags or FHF_Read_Only ; { Store is read-only }
    end ;

    { Now set up various variables and read in AT segment 0 }

    Resolution := Minimum ; { Number of bytes allocated per AT bit }
    Byte_Mapping := 8 * Resolution ; { How much one AT byte maps }

    { Allocate AT memory }
    AT.Init( Resolution ) ;
    AT.Allow_Cache := Allow_Cache ;
    AT.Set_Store( FH_Store ) ;

    Determine_AT_Metrics( _Result ) ;
    if( _Result.Code <> 0 ) then
    begin
        exit ;
    end ;
    At.Load_Table ;

    _Result.Facility := FHErr_Facility ;
    _Result.Code := 0 ; { Setup was OK }
    Last_Error := _Result ;
    Serial_Number := Get_Serial_Number ;

    Initialize ;
end ; { TFH.Common_Init }


destructor TFH.Done ; { Finish up FH access }

begin
{$IFDEF Verify}
    Address_List.Free ;
    Size_List.Free ;
{$ENDIF}
    AT.Done ;
    if( ( ( Flags and FHF_NoClose ) = 0 ) and ( FH_Store <> nil ) ) then { Not prevented from closing the store heap }
    begin
        FH_Store^.Detach ;
        FH_Store := nil ;
    end ;

    { Release memory }
    if( Own_Allocator ) then
    begin
        dispose( Memory, Done ) ;
        Memory := nil ;
    end ;

    { Notify FH list }
    if( Device > 0 ) then
    begin
        Remove_From_FH_List( @self ) ;
    end ;

{$IFDEF Debug}
    Sanity := ' ' ;
{$ENDIF}

    { Inherited destructor }
    inherited Done ;
end ; { TFH.Done }


procedure TFH.Update_Header ;

var Header : FH_Header ;
{$IFNDEF DELPHI}
    Dummy : TRecord_Size ;
{$ENDIF}
    UEC : TUEC ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    { Now write FH header... }
    fillchar( Header, sizeof( Header ), 0 ) ;
    Header.ID := -1 ;
    Header.ID1 := 1 ;
{$IFDEF DELPHI}
    Header.AT_Offset := AT.Get_Offset ;
{$ELSE}
    Header.AT_Offset := trunc( AT.Get_Offset ) ;
{$ENDIF}
    Header.AT_Size := AT.Get_Size ;
{$IFNDEF DELPHI}
    Dummy :=
{$ENDIF}
        FH_Store^.Write_Data( Header, Resolution, sizeof( Header ), UEC ) ;
end ;


function TFH.Path : string ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Path := FH_Store^.Get_Name ;
end ;


procedure TFH.New( _Size : TStore_Address ) ; { Create new FH, if empty }

var Counter : longint ;
{    Dummy : TRecord_Size ;}
    M : TStore_Address ;
    P : TStore_Address ;
{    UEC : TUEC ;}

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    { Determine initial file _Size and _Size of FAT }
    AT.Set_Offset( ( _Size + Byte_Mapping + Resolution * 2 ) and
        ( not ( Byte_Mapping - 1 ) ) ) ;
    { Round up to multiple of FH byte map (plus root and header) }
{$IFDEF DELPHI}
    AT.Set_Size( AT.Get_Offset div Byte_Mapping ) ; { Allocation table _Size }
{$ELSE}
    AT.Set_Size( trunc( AT.Get_Offset ) div Byte_Mapping ) ; { Allocation table _Size }
{$ENDIF}

    { Extend file up to, and including, start of AT... }
    FH_Store^.Extend( AT.Get_Offset + 1 ) ;

    { Write out allocation table... }
    AT.Allocate_At( 0, 2 * Resolution ) ;

    P := AT.Get_Offset + 1 ;
    Counter := AT.Get_Size - 1 ;
    M := FH_Store^.Max_Storage ;
    if( P + Counter > M ) then
    begin
        FH_Store^.Extend( ( P + Counter ) - M ) ; { Extend the store to include entire AT }
    end ;
(*
    while( Counter > 0 ) do
    begin
        if( Counter > sizeof( SH_Buffer ) ) then
        begin
            Dummy := FH_Store^.Write_Data( SH_Buffer, P, sizeof( SH_Buffer ), UEC ) ;
        end else
        begin
            Dummy := FH_Store^.Write_Data( SH_Buffer, P, Counter, UEC ) ;
        end ;
        P := P + Dummy ;
        Counter := Counter - sizeof( SH_Buffer ) ;
    end ;
*)
    inc( Statistics.Extensions ) ;
    AT.Flush ;
    Update_Header ;
end ; { TFH.New }


function TFH.Allocate( _Size : TStore_Address ) : TStore_Address ;
{ Allocate _Size bytes in store heap and return pointer.  Return 0 if error. }

var Actual_Size : longint ;
{    Dumm : TRecord_Size ;}
    Dummy, Dummy1{, Dummy2} : longint ; { Dummy variables }
    File_Size : longint ;
    Offset : TStore_Address ;
{    P : TStore_Address ;}
{    UEC : TUEC ;}
{    Write_Count : longint ; { Count of 128 byte sections to write at EOF }

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Allocate := 0 ; { Return zero (0) if error }
    if( _Size < 1 ) then { Cannot allocate nothing }
    begin
        exit ;
    end ;
    _Size := ( _Size + Resolution - 1 ) and ( not ( Resolution - 1 ) ) ;
    { Allocate minimum multiple bytes at a time }

    { Determine _Size for file extensions... }
    Actual_Size := _Size ;
    if( Extension_Resolution > 0 ) then
    begin
        File_Size := FH_Store^.Max_Storage ;
        Actual_Size := Extension_Resolution * File_Size div 100 ;
        if( Actual_Size <= 0 ) then { Handle overflow for big files }
        begin
            Actual_Size := Extension_Resolution * ( File_Size div 100 ) ;
        end ;
        if( Actual_Size < _Size ) then
        begin
            Actual_Size := _Size ;
        end ;
        if( Actual_Size > Max_Memory - File_Size ) then
        begin
            Actual_Size := Max_Memory - File_Size ;
        end ;
    end ; { if( Extension_Resolution > 0 ) }

    if( AT.Get_Offset = 0 ) then { File is empty }
    begin
        { Setup... }
        if( ( Flags and FHF_NoExtend ) <> 0 ) then { Cannot extend file }
        begin
            Log_Error( FHErr_No_Room_On_Device, 2 ) ;
            exit ;
        end ;
        New( Actual_Size ) ;
    end ; { if AT_Offset = 0 }

    Offset := AT.Allocate( _Size ) ;
    if( Offset = 0 ) then { No room in AT }
    begin
        { Need to expand file... }
        if( ( Flags and FHF_NoExtend ) <> 0 ) then { Cannot extend file }
        begin
            exit ;
        end ;

        { We always extend by at least the AT _Size so that we don't
          overwrite the original AT when we copy it (so we can recover
          if something goes wrong during the extend operation). }
        if( Actual_Size < AT.Get_Size ) then
        begin
            Actual_Size := AT.Get_Size ;
        end ;

        Dummy := ( Actual_Size + Byte_Mapping - 1 ) div Byte_Mapping ;
        { How big next AT segment will need to be (minimum) to provide space for
          current allocation request (and any additional extension amount).  It
          must be rounded up to a full byte. }

        { Physically extend the FH file... }
        Dummy1 := Dummy + Dummy * Byte_Mapping ; { Additional space needed in the store }
        if( FH_Store^.Extend( Dummy1 ) <> Dummy1 ) then { Failed to extend by required amount }
        begin
            Log_Error( FHErr_No_Room_On_Device, 2 ) ;
            exit ;
        end ;
(*
        P := FH_Store^.Max_Storage ; { Go to EOS }
        Dummy1 := Dummy ;
        while( Dummy1 > 0 ) do { Extend file, as needed }
        begin
            dec( Dummy1 ) ;
            { One less AT byte worth of space to extend }
            FH_Store^.Extend( 1 ) ;
            Dumm := FH_Store^.Write_Data( SH_Buffer, P, 1, UEC ) ;
            { Need a byte for AT }
            P := P + Dumm ;

            if( Byte_Mapping > sizeof( SH_Buffer ) ) then
            { Can only write out SH_Buffer zeroes at a time }
            begin
                Dummy2 := sizeof( SH_Buffer ) ;
            end else
            begin
                Dummy2 := Byte_Mapping ;
            end ;
            Write_Count := Byte_Mapping div Dummy2 ;
            while( Write_Count > 0 ) do
            begin
                FH_Store^.Extend( Dummy2 ) ;
                Dumm := FH_Store^.Write_Data( SH_Buffer, P, Dummy2, UEC ) ;
                { Need space for what AT byte can map }
                P := P + Dumm ;
                dec( Write_Count ) ;
            end ;
        end ; { while( Dummy1 > 0 ) }
*)

        AT.Set_Size( AT.Get_Size + Dummy ) ; { New AT Size }
        AT.Set_Offset( FH_Store^.Max_Storage - AT.Get_Size ) ; { New AT position }

        inc( Statistics.Extensions ) ;
        Update_Header ;
        Offset := AT.Allocate( _Size ) ;
    end ; { if( Offset = 0 ) }

    if( Offset < 2 * Resolution ) then
    begin
        { The only way we should get here is if the allocation table didn't have
          the first 2 Resolution bytes allocated, which indicates corruption of
          the AT. }
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( FHErr_Corrupted_File_Structure, 3 ) ;
        if( Application_Manager <> nil ) then
        begin
            Application_Manager^.Fatal_Exception( 'Corrupted File Structure' ) ;
        end ;
        halt ;
    end ;
    Allocate := Offset ; { Return offset of allocation }
    AT.Flush ;
    Log_Allocation( Offset, _Size ) ;
end ; { TFH.Allocate }


{ Allocate space at a specific location }
function TFH.Allocate_At( _Position, _Size : TStore_Address ) : boolean ;

var Actual_Size : longint ;
{    Dumm : TRecord_Size ;}
    Dummy, Dummy1{, Dummy2} : longint ; { Dummy variables }
    File_Size : longint ;
{    Offset : TStore_Address ;}
{    P : TStore_Address ;}
{    UEC : TUEC ;}
{    Write_Count : longint ; { Count of 128 byte sections to write at EOF }

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Allocate_At := False ; { Return false if error }
    if( _Size < 1 ) then { Cannot allocate nothing }
    begin
        exit ;
    end ;
    _Size := ( _Size + Resolution - 1 ) and ( not ( Resolution - 1 ) ) ;
    { Allocate minimum multiple bytes at a time }

    { Determine _Size for file extensions... }
    Actual_Size := _Size ;
    if( Extension_Resolution > 0 ) then
    begin
        File_Size := FH_Store^.Max_Storage ;
        Actual_Size := Extension_Resolution * File_Size div 100 ;
        if( Actual_Size <= 0 ) then { Handle overflow for big files }
        begin
            Actual_Size := Extension_Resolution * ( File_Size div 100 ) ;
        end ;
        if( Actual_Size < _Size ) then
        begin
            Actual_Size := _Size ;
        end ;
        if( Actual_Size > Max_Memory - File_Size ) then
        begin
            Actual_Size := Max_Memory - File_Size ;
        end ;
    end ; { if( Extension_Resolution > 0 ) }

    if( AT.Get_Offset = 0 ) then { File is empty }
    begin
        { Setup... }
        if( ( Flags and FHF_NoExtend ) <> 0 ) then { Cannot extend file }
        begin
            Log_Error( FHErr_No_Room_On_Device, 2 ) ;
            exit ;
        end ;
        New( Actual_Size ) ;
    end ; { if AT_Offset = 0 }

    if( not AT.Allocate_At( _Position, _Size ) ) then { No room in AT }
    begin
        { Need to expand file... }
        if( ( Flags and FHF_NoExtend ) <> 0 ) then { Cannot extend file }
        begin
            exit ;
        end ;
        if( _Position + _Size < AT.Get_Offset ) then { Not due to store being too small }
        begin
            exit ; { Exit, otherwise we needlessly expand the store }
        end ;

        { We always extend by at least the AT _Size so that we don't
          overwrite the original AT when we copy it (so we can recover
          if something goes wrong during the extend operation). }
        if( Actual_Size < AT.Get_Size ) then
        begin
            Actual_Size := AT.Get_Size ;
        end ;

        Dummy := ( Actual_Size + Byte_Mapping - 1 ) div Byte_Mapping ;
        { How big next AT segment will need to be (minimum) to provide space for
          current allocation request (and any additional extension amount).  It
          must be rounded up to a full byte. }

        { Physically extend the FH file... }
        Dummy1 := Dummy + Dummy * Byte_Mapping ; { Additional space needed in the store }
        if( FH_Store^.Extend( Dummy1 ) <> Dummy1 ) then { Failed to extend by required amount }
        begin
            exit ;
        end ;
(*
        P := FH_Store^.Max_Storage ; { Go to EOS }
        Dummy1 := Dummy ;
        while( Dummy1 > 0 ) do { Extend file, as needed }
        begin
            dec( Dummy1 ) ;
            { One less AT byte worth of space to extend }
            FH_Store^.Extend( 1 ) ;
            Dumm := FH_Store^.Write_Data( SH_Buffer, P, 1, UEC ) ;
            { Need a byte for AT }
            P := P + Dumm ;

            if( Byte_Mapping > sizeof( SH_Buffer ) ) then
            { Can only write out SH_Buffer zeroes at a time }
            begin
                Dummy2 := sizeof( SH_Buffer ) ;
            end else
            begin
                Dummy2 := Byte_Mapping ;
            end ;
            Write_Count := Byte_Mapping div Dummy2 ;
            while( Write_Count > 0 ) do
            begin
                FH_Store^.Extend( Dummy2 ) ;
                Dumm := FH_Store^.Write_Data( SH_Buffer, P, Dummy2, UEC ) ;
                { Need space for what AT byte can map }
                P := P + Dumm ;
                dec( Write_Count ) ;
            end ;
        end ; { while( Dummy1 > 0 ) }
*)

        AT.Set_Size( AT.Get_Size + Dummy ) ; { New AT Size }
        AT.Set_Offset( FH_Store^.Max_Storage - AT.Get_Size ) ; { New AT position }

        inc( Statistics.Extensions ) ;
        Update_Header ;
        if( not AT.Allocate_At( _Position, _Size ) ) then
        begin
            exit ;
        end ;
    end ; { if( Offset = 0 ) }

    Allocate_At := True ; { Return offset of allocation }
    AT.Flush ;
    Log_Allocation( _Position, _Size ) ;
end ; { TFH.Allocate_At }


procedure TFH.Deallocate( PTR, _Size : TStore_Address ) ; { Deallocate space in FH }

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    if( ( Flags and FHF_Fast_Deallocate ) <> 0 ) then
    begin
        _Size := 0 ; { Don't deallocate }
    end ;
    if( _Size = 0 ) then
    begin
        inc( Statistics.Deallocations ) ;
        exit ;
    end ;
    _Size := ( _Size + Resolution - 1 ) and ( not ( Resolution - 1 ) ) ;
    { Allocate minimum multiple bytes at a time }

    if(
        ( PTR < Resolution )
        or
        ( PTR < Lowest_Dynamic )
        or
        ( PTR + _Size > AT.Get_Offset )
        or
        ( _Size < 1 )
      ) then
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( FHErr_Invalid_Pointer_Operation, 3 ) ;
        if( Application_Manager <> nil ) then
        begin
            Application_Manager^.Fatal_Exception( 'Invalid Pointer Operation' ) ;
        end ;
        halt ;
    end ;
    Log_Deallocation( PTR, _Size ) ;
    AT.Deallocate( PTR, _Size ) ;
end ; { FH.Deallocate }


procedure TFH.Read( PTR : TStore_Address ; _Size : TStore_Address ;
    var Buffer ) ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Read_Heap( PTR, _Size, Buffer ) ;
end ;


procedure TFH.Read_Heap( PTR, _Size : TStore_Address ; var Buffer ) ; { Read data from FH }

var Dummy : TRecord_Size ;
    Len : longint ;
    UEC : TUEC ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    if( _Size = 0 ) then
    begin
        inc( Store_Statistics.Reads ) ;
        exit ; { Do nothing }
    end ;
    if(
        ( PTR < Resolution )
        or
        ( PTR + _Size > AT.Get_Offset )
        or
        ( _Size < 1 )
      ) then
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( FHErr_Invalid_Pointer_Operation, 3 ) ;
        if( Application_Manager <> nil ) then
        begin
            Application_Manager^.Fatal_Exception( 'Invalid Pointer Operation' ) ;
        end ;
        halt ;
    end ;
    Log_Read( PTR, _Size ) ;
    if( _Cache <> nil ) then
    begin
        Len := _Cache^.Read_Cache( Buffer, PTR, _Size ) ;
        if( Len = _Size ) then { Data was found in cache }
        begin
            Store_Statistics.Bytes_Read := Store_Statistics.Bytes_Read + Len ;
            exit ;
        end ;
    end ;
    Dummy := FH_Store^.Read_Data( Buffer, PTR, _Size, UEC ) ;
    Store_Statistics.Bytes_Read := Store_Statistics.Bytes_Read + Dummy ;
end ; { TFH.Read_Heap }


function TFH.Reallocate( PTR : TStore_Address ; Old, _New : TStore_Address ) : TStore_Address ;
{ Reallocate a pointer }

var Dummy : longint ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Dummy := not ( Resolution - 1 ) ;
    if( ( ( Old - 1 ) and Dummy ) = ( ( _New - 1 ) and Dummy ) ) then
    { No actual reallocation is needed: new allocation fits within the old one
      due to minimum resolution. }
    begin
        Reallocate := PTR ;
        inc( Statistics.Reallocations_In_Place ) ;
    end else
    begin
        Reallocate := inherited Reallocate( PTR, Old, _New ) ;
    end ;
end ;


procedure TFH.Write( PTR : TStore_Address ; _Size : TStore_Address ;
    var Buffer ) ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Write_Heap( PTR, _Size, Buffer ) ;
end ;


procedure TFH.Write_Heap( PTR, _Size : TStore_Address ; var Buffer ) ;
{ Write data to FH }

var Dummy : TRecord_Size ;
    UEC : TUEC ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    if( _Size = 0 ) then
    begin
        inc( Store_Statistics.Writes ) ;
        exit ;
    end ;
    if(
        ( PTR < Resolution )
        or
        ( PTR < Lowest_Dynamic )
        or
        ( PTR + _Size > AT.Get_Offset )
        or
        ( _Size < 1 )
      ) then
    begin
        Set_Master_Error_Operation( TMEO_Show ) ;
        Log_Error( FHErr_Invalid_Pointer_Operation, 3 ) ;
        if( Application_Manager <> nil ) then
        begin
            Application_Manager^.Fatal_Exception( 'Invalid Pointer Operation' ) ;
        end ;
        halt ;
    end ;
    Log_Write( PTR, _Size ) ;
    Dummy := FH_Store^.Write_Data( Buffer, PTR, _Size, UEC ) ;
    Store_Statistics.Bytes_Written := Store_Statistics.Bytes_Written + Dummy ;

    if( _Cache <> nil ) then
    begin
        _Cache^.Write_Cache( Buffer, PTR, _Size ) ;
    end ;
end ; { TFH.Write_Heap }


procedure TFH.Set_Root( var Buffer ) ; { Write data to root block of FH }

var Dumm : TRecord_Size ;
    UEC : TUEC ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    if( AT.Get_Offset = 0 ) then { File is empty }
    begin
        New( Resolution ) ;
    end ;
    if( Device = -1 ) then { Reserve first block of device for bootstrap }
    begin
        Dumm := FH_Store^.Write_Data( Buffer, Resolution, Resolution, UEC ) ;
    end else
    begin
        Dumm := FH_Store^.Write_Data( Buffer, 0, Resolution, UEC ) ;
    end ;
    if( Dumm <> Resolution ) then
    begin
        Log_Error( FHErr_Write_Failure, SEL_Fatal ) ;
    end ;
    inc( Statistics.Root_Writes ) ;
end ;


procedure TFH.Get_Root( var Buffer ) ; {Read data from root block of FH}

var
{$IFNDEF DELPHI}
    Dummy : TRecord_Size ;
{$ENDIF}
    UEC : TUEC ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    if( AT.Get_Offset = 0 ) then { File is empty }
    begin
        fillchar( Buffer, Resolution, 0 ) ;
    end else
    begin
        if( Device = -1 ) then
        { First block of device is reserved for bootstrap }
        begin
{$IFNDEF DELPHI}
            Dummy :=
{$ENDIF}
                FH_Store^.Read_Data( Buffer, Resolution, Resolution, UEC ) ;
        end else
        begin
{$IFNDEF DELPHI}
            Dummy :=
{$ENDIF}
                FH_Store^.Read_Data( Buffer, 0, Resolution, UEC ) ;
        end ;
        inc( Statistics.Root_Reads ) ;
    end ;
end ;


function TFH.XSpaceAvail : TStore_Address ;
{ Return available space TStore_Address store heap including file expansion }

var R, R1 : TStore_Address ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    R := SpaceAvail ; { Space available within store heap }
    if( FH_Store^.Extend( 0 ) = 0 ) then { Can extend }
    begin
        if( Diskfree( Device ) + R > OS.Max_File_Size ) then
        begin
            R1 := OS.Max_File_Size - FH_Store^.Max_Storage ; { Amount file can expand }
            R1 := R1 div ( ( Resolution * 8 ) + 1 ) * Resolution * 8 ; { Subtract overhead }
            R := R1 + R ;
        end else
        begin
            R := R + Diskfree( Device ) ;
        end ;
    end ;
{$WARNINGS OFF}
    if( R > $7FFFFFFF ) then { Larger value than fits in longint function result }
{$WARNINGS ON}
    begin
        R := $7FFFFFFF ;
    end ;
    XSpaceAvail := R ;
end ;


function TFH.SpaceAvail : TStore_Address ;
{ Return available space within store heap without expanding file }

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    SpaceAvail := AT.SpaceAvail ;
end ;


function TFH.MaxSpace : TStore_Address ;
{ Return largest contiguous space without extending file. }

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    MaxSpace := AT.MaxSpace ;
end ; { TFH.MaxSpace }


function TFH.Debugger : PSDebug_Interface ;

var R : PSFH_Debugger ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    system.new( R, Init( @Self ) ) ;
    Debugger := R ;
end ;


procedure TFH.Set_Cache_Flag( Flag : boolean ) ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    if( Flag <> Allow_Cache ) then
    begin
        Allow_Cache := Flag ;
        if( not Flag ) then
        begin
            if( AT.Get_Modified ) then
            begin
                AT.Flush ;
            end ;
        end ;
        AT.Allow_Cache := Flag ;
    end ;
end ;


function TFH.Classname : string ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TFH' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Classname := 'TFH' ;
end ;


function TFH.Is_Class( const Name : string ) : boolean ;

begin
{$IFDEF Debug}
    if( Sanity <> 'TFH' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    if( Name = Classname ) then
    begin
        Is_Class := True ;
    end else
    begin
        Is_Class := inherited Is_Class( Name ) ;
    end ;
end ;


function TFH.Extended_Size : TStore_Address ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Extended_Size := XSpaceAvail ;
end ;


function TFH.Valid_Data_Address( Address : TStore_Address ) : boolean ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    Valid_Data_Address := ( Address >= Resolution ) and ( Address < AT.Get_Offset ) ;
end ;


procedure TFH.Write_Cache ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    AT.Flush ;
    inc( FH_Statistics.Index_Writes ) ;
end ;


procedure TFH.Log_Allocation( Address, _Size : TStore_Address ) ;

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    inc( Statistics.Allocations ) ;
    Statistics.Bytes_Allocated := Statistics.Bytes_Allocated + _Size ;
{$IFDEF Verify}
    if( Track_FH ) then
    begin
        Global_Address_List.Add( Address ) ;
        Global_Size_List.Add( _Size ) ;
    end ;
{$ENDIF}
end ;


procedure TFH.Log_Deallocation( Address, _Size : TStore_Address ) ;

{$IFDEF Verify}
var Dummy : integer ;
{$ENDIF}

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    inc( Statistics.Deallocations ) ;
    Statistics.Bytes_Deallocated := Statistics.Bytes_Deallocated + _Size ;
{$IFDEF Verify}
    if( Track_FH ) then
    begin
        Dummy := Global_Address_List.Indexof( Address ) ;
        if( Dummy = -1 ) then
        begin
            Log_Error( FHErr_Deallocation_Error, 2 ) ;
        end ;
        if( _Size <> Size_List[ Dummy ] ) then
        begin
            Log_Error( FHErr_Size_Mismatch, 2 ) ;
        end ;
    end ;
{$ENDIF}
end ; { TFH.Log_Deallocation }


procedure TFH.Log_Read( Address, _Size : TStore_Address ) ;

{$IFDEF Verify}
var Dummy : integer ;
{$ENDIF}

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    inc( Store_Statistics.Reads ) ;
{$IFDEF Verify}
    if( Track_FH ) then
    begin
        Dummy := Global_Address_List.Indexof( Address ) ;
        if( Dummy = -1 ) then
        begin
            Log_Error( FHErr_Read_Out_Of_Range, 2 ) ;
        end ;
        if( _Size <> Size_List[ Dummy ] ) then
        begin
            Log_Error( FHErr_Read_Size_Error, 2 ) ;
        end ;
    end ;
{$ENDIF}
end ;


procedure TFH.Log_Write( Address, _Size : TStore_Address ) ;

{$IFDEF Verify}
var Dummy : integer ;
{$ENDIF}

begin
{$IFDEF Debug}
    Check_FH( self ) ;
{$ENDIF}
    inc( Store_Statistics.Writes ) ;
{$IFDEF Verify}
    if( Track_FH ) then
    begin
        Dummy := Global_Address_List.Indexof( Address ) ;
        if( Dummy = -1 ) then
        begin
            Log_Error( FHErr_Write_Out_Of_Range, 2 ) ;
        end ;
        if( _Size <> Size_List[ Dummy ] ) then
        begin
            Log_Error( FHErr_Write_Size_Error, 2 ) ;
        end ;
    end ;
{$ENDIF}
end ;



var SH_List : TCollection ;

procedure Add_To_FH_List( X : pointer ) ;

begin
    SH_List.Insert( X ) ;
end ;


procedure Remove_From_FH_List( X : pointer ) ;

begin
    SH_List.Delete( X ) ;
end ;


procedure Flush_FH( T : pointer ; const X : string ) ;

var Dummy : integer ;
    P : PFH ;

begin
    for Dummy := 0 to SH_List.Count - 1 do
    begin
        P := PFH( SH_List.At( Dummy ) ) ;
        if( P <> T ) then { Don't recurse }
        begin
            if( P^.AT.Get_Modified ) then
            begin
                if( P^.FH_Store^.Get_Name = X ) then
                begin
                    P^.Write_Cache ;
                end ;
            end ;
        end ;
    end ;
end ;


{ Unit initialization and finalization... }

{$IFDEF DELPHI}
var FSEI : pFS_EI = nil ;
    FHEI : pFH_EI = nil ;
{$ELSE}
var FSEI : pFS_EI ;
    FHEI : pFH_EI ;
{$ENDIF}

type TXFile_System_Server = object( TFile_System_Server )
                               public { API... }
                                   function Get_File_System_For_Store( Store : PSStore ;
                                       Own : boolean ) : PFile_System ;
                                       virtual ;
                                   { Creates and returns a file system for the
                                     passed specification.  On return, Name has
                                     the file system ID stripped from it.  If
                                     there is no file system available, nil is
                                     returned. }

                                   procedure Initialize( const Nam, ID : string ;
                                       Flags : integer ; var UEC : TUEC ) ;
                                       virtual ;
                                   { Initializes a file system on the specified
                                     device, with the specified ID and flags.
                                     Returns status in UEC. }
                           end ; { TXFile_System_Server }
     PXFile_System_Server = ^TXFile_System_Server ;

{ API... }

function TXFile_System_Server.Get_File_System_For_Store( Store : PSStore ;
    Own : boolean ) : PFile_System ;
{ Creates and returns a file system for the passed specification.  On return,
  Name has the file system ID stripped from it.  If there is no file system
  available, nil is returned. }

var Flags : integer ;
    FS : PXFile_System ;
    Res : TUEC ;

begin
    if( Own ) then
    begin
        Flags := 0 ;
    end else
    begin
        Flags := FSF_NoOwnStore ;
    end ;
    FS := new( PXFile_System, Init_From_Store( Store, Flags, Res ) ) ;
    if( Res.Code <> 0 ) then
    begin
        dispose( FS, Done ) ;
        FS := nil ;
    end ;
    Get_File_System_For_Store := FS ;
end ; { TXFile_System_Server.Get_File_System }


procedure TXFile_System_Server.Initialize( const Nam, ID : string ;
    Flags : integer ; var UEC : TUEC ) ;

var FS : PXFile_System ;

begin
    FS := new( PXFile_System, Init( '', 0, UEC ) ) ;
    if( UEC.Code = 0 ) then
    begin
        FS^.Initialize( Nam, ID, Flags, UEC ) ;
    end ;
    dispose( FS, Done ) ;
end ;


var XFSS : TXFile_System_Server ;

procedure Unit_Initialization ;

begin
    { Set up SH_Buffer with zeroes.  This is so we can quickly blockwrite
      zeroes to the store heap. }
    fillchar( SH_Buffer, sizeof( SH_Buffer ), 0 ) ;
    fillchar( FS_Buffer, sizeof( FS_Buffer ), 0 ) ;

{$IFDEF Verify}
    Global_Address_List := TInteger_List.Create ;
    Global_Size_List := TInteger_List.Create ;
{$ENDIF}

    { Create store heap list. }
    SH_List.Init( 4, 4 ) ;

    { Create error interfaces... }
{$IFDEF DELPHI}
    FSEI := tFS_EI.Create ;
    FHEI := tFH_EI.Create ;
    Register_Error_Interface( PSError_Interface( FSEI ) ) ;
    Register_Error_Interface( PSError_Interface( FHEI ) ) ;
{$ELSE}
    new( FSEI, Init ) ;
    new( FHEI, Init ) ;
    Register_Error_Interface( FSEI ) ;
    Register_Error_Interface( FHEI ) ;
{$ENDIF}

    XFSS.Init ;
    Register_File_System_Server( @XFSS ) ;
end ; { Unit_Initialization }


procedure Unit_Deallocation ;

begin
    SH_List.Done ;
{$IFDEF Verify}
    Global_Address_List.Free ;
    Global_Size_List.Free ;
{$ENDIF}
{$IFDEF DELPHI}
    Unregister_Error_Interface( FSEI.Get_Facility ) ;
    Unregister_Error_Interface( FHEI.Get_Facility ) ;
    FSEI.Free ;
    FSEI := nil ;
    FHEI.Free ;
    FHEI := nil ;
{$ELSE}
    Unregister_Error_Interface( FSEI^.Get_Facility ) ;
    Unregister_Error_Interface( FHEI^.Get_Facility ) ;
    dispose( FSEI ) ;
    FSEI := nil ;
    dispose( FHEI ) ;
    FHEI := nil ;
{$ENDIF}
end ; { Unit_Deallocation }


{$IFDEF Test}
procedure Test_FS ;

var F : PSFile ;
    UEC : TUEC ;

    procedure Check_UEC ;

    begin
        if( UEC.Code <> 0 ) then
        begin
            ShowMessage( PChar( 'Error: call failed with error ' + inttostr( UEC.Facility ) + ',' + inttostr( UEC.Code ) ) ) ;
            exit ;
        end ;
    end ;

    procedure Check_Not_UEC ;

    begin
        if( UEC.Code = 0 ) then
        begin
            ShowMessage( 'Error: call did not fail' ) ;
            exit ;
        end ;
    end ;

var Buffer : array[ 0..1023 ] of byte ;
    Dummy : integer ;
    Find : PFind_record ;
    FS, FS1 : PFile_System ;
    Header : FS_Header_Information ;
    S : string ;
    WDummy : word ;

    procedure Cleanup ;

    var Temp : file ;

    begin
        if( FileExists( 'C:\TESTFS.DIR' ) ) then
        begin
            assignfile( Temp, 'C:\TESTFS.DIR' ) ;
            erase( Temp ) ;
        end ;
        if( FileExists( 'C:\TESTFS1.DIR' ) ) then
        begin
            assignfile( Temp, 'C:\TESTFS1.DIR' ) ;
            erase( Temp ) ;
        end ;
        if( FileExists( 'C:\TEST.TXT' ) ) then
        begin
            assignfile( Temp, 'C:\TEST.TXT' ) ;
            erase( Temp ) ;
        end ;
        {$I-}
        Rmdir( 'C:\FS' ) ;
        {$I+}
        IOResult ;
    end ;


    procedure Show_Error ;

    begin
        ShowMessage( 'Error' ) ;
    end ;

var Handle : integer ;

begin { Test_FS }
    { Prepare clean slate for test... }
    Cleanup ;
    New( Find ) ;

    Initialize_Find_Record( Find^ ) ;

    { DOS file tests... }
    S := Files.Lookup( 'C:\*.*', Find^ ) ;
    if( S = '' ) then
    begin
        Show_Error ;
    end ;
    S := Files.Lookup( '', Find^ ) ;
    if( S = '' ) then
    begin
        Show_Error ;
    end ;

    Handle := DOS_FS.Open( 'C:\AUTOEXEC.BAT', FSOM_INPUT ) ;
    DOS_FS.Last_Error( UEC ) ;
    Check_UEC ;
    DOS_FS.Close( Handle ) ;
    Handle := DOS_FS.Create( 'C:\TEST.TXT', FSOM_TEXT ) ;
    DOS_FS.Last_Error( UEC ) ;
    Check_UEC ;
    DOS_FS.Close( Handle ) ;
    DOS_FS.Get_Header( 'C:', Header, UEC ) ;
    Check_UEC ;
    DOS_FS.Set_Header( 'C:', Header, UEC ) ;
    Check_UEC ;
    if( DOS_FS.MkDir( 'C:\FS' ) <> 0 ) then
    begin
        Show_Error ;
    end ;
    if( DOS_FS.DeviceFree( '' ) <> -1 ) then
    begin
        Show_Error ;
    end ;
    if( DOS_FS.DeviceFree( 'C:' ) = -1 ) then
    begin
        Show_Error ;
    end ;
    if( DOS_FS.DeviceSize( '' ) = -1 ) then
    begin
        Show_Error ;
    end ;
    if( DOS_FS.DeviceSize( 'C:' ) = -1 ) then
    begin
        Show_Error ;
    end ;
{
    S := Lookup( '*:\A.A', Find ) ;
    S := Lookup( '', Find ) ;
}

    { Test FS initialization... }
    Initialize_File_System( 'C:\?.DIR', 'Test FS', 0, UEC ) ;
    Check_Not_UEC ;
    Initialize_File_System( 'C:\TESTFS.DIR', 'Test FS', 0, UEC ) ;
    Check_UEC ;

    FS := Open_File_System( 'C:\TESTFS.DIR', 0, UEC ) ;
    Check_UEC ;
    if( FS = nil ) then
    begin
        Show_Error ;
    end ;
    FS.Verify( 'C:\TESTFS.DIR', UEC ) ;
    Check_UEC ;
    if( Files.Lookup( 'C:\TESTFS.DIR\*', Find^ ) <> '' ) then
    begin
        Show_Error ;
    end ;
    if( Files.Lookup( '', Find^ ) <> '' ) then
    begin
        Show_Error ;
    end ;
    if( Files.Lookup( 'C:\TESTFS.DIR\...', Find^ ) <> '' ) then
    begin
        Show_Error ;
    end ;
    FS1 := Open_File_System( 'C:\TESTFS.DIR', 0, UEC ) ;

    { Basic text file tests... }
    F := Open( 'C:\TESTFS.DIR\TEST.TXT', FSOM_OUTPUT or FSOM_TEXT, UEC ) ;
    Check_UEC ;
    Initialize_File_System( 'C:\TESTFS.DIR', 'Test FS', 0, UEC ) ;
    Check_Not_UEC ; { Should fail since it is open }
    Initialize_File_System( 'C:\TESTFS1.DIR', 'Test FS1', 0, UEC ) ;
    Check_UEC ;
    if( FS1.MkDir( '\D' ) <> 0 ) then
    begin
        Show_Error ;
    end ;
    S := Files.Lookup( 'C:\TESTFS.DIR\...\*.*', Find^ ) ;
    while( length( S ) > 0 ) do
    begin
        S := Files.Lookup( '', Find^ ) ;
    end ;
    if( FS1.MkDir( '\B' ) <> 0 ) then
    begin
        Show_Error ;
    end ;
    if( FS1.MkDir( '\E' ) <> 0 ) then
    begin
        Show_Error ;
    end ;
    if( FS1.MkDir( '\A' ) <> 0 ) then
    begin
        Show_Error ;
    end ;
    if( FS1.MkDir( '\C' ) = -1 ) then
    begin
        Show_Error ;
    end ;
    if( FS1.MkDir( '\F' ) <> 0 ) then
    begin
        Show_Error ;
    end ;
    if( FS1.MkDir( '\G' ) <> 0 ) then
    begin
        Show_Error ;
    end ;
    FS1.Erase( '\F' ) ;
    FS1.Last_Error( UEC ) ;
    Check_UEC ;
    FS1.Erase( '\B' ) ;
    FS1.Last_Error( UEC ) ;
    Check_UEC ;
    FS1.Erase( '\D' ) ;
    FS1.Last_Error( UEC ) ;
    Check_UEC ;
    Close_File_System( 'C:\TESTFS.DIR' ) ;
    if( not F^.EOF ) then
    begin
        Show_Error ;
    end ;
    if( F^.Get_Size <> 0 ) then
    begin
        Show_Error ;
    end ;
    F^.Writeln( 'Test line 1' ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Writeln( 'Test line 2' ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Seek( 10240 ) ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.Seek( 0 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( S <> 'Test line 1' ) then
    begin
        Show_Error ;
    end ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( S <> 'Test line 2' ) then
    begin
        Show_Error ;
    end ;
    F^.SetTextBuf( Buffer, sizeof( Buffer ) ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Seek( 0 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( F^.Get_Size = 0 ) then
    begin
        Show_Error ;
    end ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( S <> 'Test line 1' ) then
    begin
        Show_Error ;
    end ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( S <> 'Test line 2' ) then
    begin
        Show_Error ;
    end ;
    F^.SeekEOF ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Writeln( 'Test line 3' ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Reset( '' ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( S <> 'Test line 1' ) then
    begin
        Show_Error ;
    end ;
    if( F^.EOF ) then
    begin
        Show_Error ;
    end ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( S <> 'Test line 2' ) then
    begin
        Show_Error ;
    end ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( S <> 'Test line 3' ) then
    begin
        Show_Error ;
    end ;
    F^.Readln( S ) ; { Read at EOF }
    UEC := F^.IO_Error ;
    CHeck_Not_UEC ;
    WDummy := F^.Get_Attributes ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( WDummy <> 0 ) then
    begin
        Show_Error ;
    end ;
    WDummy := F^.Get_Attributes ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( WDummy <> 0 ) then
    begin
        Show_Error ;
    end ;
    if( F^.FilePos <> 39 ) then
    begin
        Show_Error ;
    end ;
    F^.Seek( 0 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Append( '' ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Writeln( 'Test line 4' ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( F^.Get_Size <> 52 ) then
    begin
        Show_Error ;
    end ;
    F^.Set_Size( F^.Get_Size - 2 ) ; { Trim CRLF off of last line }
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Seek( 0 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Readln( S ) ; { Line 1 }
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Readln( S ) ; { Line 2 }
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Readln( S ) ; { Line 3 }
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Readln( S ) ; { Line 4 (sans CRLF) }
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Readln( S ) ; { Read past EOF }
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    dispose( F, Done ) ;
    F := Open( 'C:\TESTFS.DIR\TEST.TXT', FSOM_APPEND or FSOM_TEXT, UEC ) ;
    Check_UEC ;
    F^.Writeln( '...Test line 4 extension' ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    dispose( F, Done ) ;
    F := Open( 'C:\TESTFS.DIR\TEST.TXT', FSOM_OUTPUT, UEC ) ;
    Check_UEC ;
    S := 'Test line 1' + LF ; { Non-CR EOL }
    move( S[ 1 ], Buffer, length( S ) ) ;
    F^.Blockwrite( Buffer, length( S ), Dummy ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( Dummy <> length( S ) ) then
    begin
        Show_Error ;
    end ;
    dispose( F, Done ) ;
    F := Open( 'C:\TESTFS.DIR\TEST.TXT', FSOM_INPUT or FSOM_TEXT, UEC ) ;
    Check_UEC ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    dispose( F, Done ) ;
    F := Open( 'C:\TESTFS.DIR\TEST.TXT', FSOM_OUTPUT, UEC ) ;
    Check_UEC ;
    S := 'Test line 1' + CR ; { CR-only EOL }
    move( S[ 1 ], Buffer, length( S ) ) ;
    F^.Blockwrite( Buffer, length( S ), Dummy ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( Dummy <> length( S ) ) then
    begin
        Show_Error ;
    end ;
    dispose( F, Done ) ;
    F := Open( 'C:\TESTFS.DIR\TEST.TXT', FSOM_INPUT or FSOM_TEXT, UEC ) ;
    Check_UEC ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    dispose( F, Done ) ;
    F := Open( 'C:\TESTFS.DIR\TEST.TXT', FSOM_TEXT, UEC ) ;
    Check_UEC ;
    dispose( F, Done ) ;
    F := Open( 'C:\TESTFS.DIR\TEST1.TXT', FSOM_TEXT, UEC ) ;
    Check_UEC ;
    System.setlength( S, 127 ) ;
    fillchar( S[ 1 ], 127, 65 ) ;
    F^.Writeln( S ) ; { Make CR/LF split across 128-byte buffers }
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Writeln( #26 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Seek( 0 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( not F^.EOF ) then
    begin
        Show_Error ;
    end ;
    dispose( F, Done ) ;

    { Basic non-text-file tests... }
    F := Open( 'C:\TESTFS.DIR\TEST.DAT', FSOM_OUTPUT, UEC ) ;
    Check_UEC ;
    fillchar( Buffer, 1024, 101 ) ;
    if( F^.FilePos <> 0 ) then
    begin
        Show_Error ;
    end ;
    F^.Blockwrite( Buffer, 1023, Dummy ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( Dummy <> 1023 ) then
    begin
        Show_Error ;
    end ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( F^.FilePos <> 1023 ) then
    begin
        Show_Error ;
    end ;
    F^.EOLN ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.Set_Size( 513 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( F^.Get_Size <> 513 ) then
    begin
        Show_Error ;
    end ;
    F^.Seek( 509 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Set_Size( -1 ) ; { Truncate to current position }
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( F^.Get_Size <> 509 ) then
    begin
        Show_Error ;
    end ;
    if( not F^.EOF ) then
    begin
        Show_Error ;
    end ;
    F^.Seek( 0 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( F^.EOF ) then
    begin
        Show_Error ;
    end ;
    fillchar( Buffer, sizeof( Buffer ), 0 ) ;
    F^.BlockRead( Buffer, 509, Dummy ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( Dummy <> 509 ) then
    begin
        Show_Error ;
    end ;
    for Dummy := 0 to 508 do
    begin
        if( Buffer[ Dummy ] <> 101 ) then
        begin
            Show_Error ;
            break ;
        end ;
    end ;
    for Dummy := 509 to 1023 do
    begin
        if( Buffer[ Dummy ] <> 0 ) then
        begin
            Show_Error ;
            break ;
        end ;
    end ;
    F^.Set_Size( 0 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( F^.Get_Size <> 0 ) then
    begin
        Show_Error ;
    end ;
    F^.Set_Size( -2 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( F^.Get_Size <> 0 ) then
    begin
        Show_Error ;
    end ;
    F^.Set_Size( 1492 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( F^.Get_Size <> 1492 ) then
    begin
        Show_Error ;
    end ;
    F^.Set_Attributes( FSFF_Statistics ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.SetFTime( 1020304 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.GetFTime( Dummy ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( Dummy <> 1020304 ) then
    begin
        Show_Error ;
    end ;
    F^.Set_Attributes( FSFF_Statistics ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    dispose( F, Done ) ;
    F := Open( 'C:\TESTFS.DIR\TEST.DAT', FSOM_OUTPUT, UEC ) ; { Open for output an existing file }
    Check_UEC ;
    F^.Writeln( 'Test' ) ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    Buffer[ 0 ] := 121 ;
    F^.Write( Buffer ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Seek( 0 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Read( Buffer ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( Buffer[ 0 ] <> 121 ) then
    begin
        Show_Error ;
    end ;
    F^.Seek( 0 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Write( Buffer ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    Dummy := F^.Get_Size ;
    F^.Set_Size( 127 * 256 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Seek( 65 * 256 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Blockread( Buffer, 1, Dummy ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.SeekEOF ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.Blockread( Buffer, 1, Dummy ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Set_Size( 64 * 256 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Set_Size( 127 * 256 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Set_Size( 0 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Set_Size( 63 * 256 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Set_Size( 256 ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    dispose( F, Done ) ;

    { General file tests... }
    F := Open( 'C:\TESTFS.DIR\DELETE.FIL', FSOM_OUTPUT, UEC ) ;
    Check_UEC ;
    F^.Erase( '' ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    dispose( F, Done ) ;

    { Test directories... }
    F := Open( 'C:\TESTFS.DIR\TEST.DIR', FSOM_OUTPUT, UEC ) ;
    Check_UEC ;
    F^.Set_Attributes( FSFF_Directory ) ; { This closes the file }
    UEC := F^.IO_Error ;
    Check_UEC ;
    if( FS.FindFirst( '\*.*', 0, Find^ ) = -1 ) then
    begin
        Show_Error ;
    end ;
    if( FS.FindNext( Find^ ) = -1 ) then
    begin
        Show_Error ;
    end ;

    { Test operations on closed file... }
    F^.Close ;
    F^.SeekEOLn ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.SeekEOF ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.GetFTime( Dummy ) ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.SetFTime( 0 ) ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.Set_Size( 0 ) ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    if( F^.Get_Size <> -1 ) then
    begin
        Show_Error ;
    end ;
    if( F^.Filepos <> -1 ) then
    begin
        Show_Error ;
    end ;
    F^.EOLN ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.EOF ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.Blockwrite( Buffer, 1023, Dummy ) ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.BlockRead( Buffer, 509, Dummy ) ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.Writeln( 'Test' ) ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
    F^.Readln( S ) ;
    UEC := F^.IO_Error ;
    Check_Not_UEC ;
{$IFNDEF DELPHI}
    WDummy :=
{$ENDIF}
        FS.Get_File_Attributes( '\TEST.DIR' ) ;
    FS.Last_Error( UEC ) ;
    Check_UEC ;

    { Other tests... }
    FS.Set_File_Attributes( '\TEST.DIR', 0 ) ; { Convert directory to normal file }
    FS.Last_Error( UEC ) ;
    Check_UEC ;
    FS.Set_File_Attributes( '\TEST.DIR', FSFF_Directory ) ; { and back for next test }
    FS.Last_Error( UEC ) ;
    Check_UEC ;
    dispose( F, Done ) ;
    FS.Close( 0 ) ;
    F := Open( 'C:\TESTFS.DIR\TEST.DIR\SUB.DAT', FSOM_OUTPUT, UEC ) ;
    Check_UEC ;
    FS.Set_File_Attributes( '\TEST.DIR', 0 ) ; { Convert non-emptry directory }
    dispose( F, Done ) ;
    F := Open( 'C:\TESTFS.DIR\TEST.DIR\SUB.DAT\FAIL.TST', FSOM_OUTPUT, UEC ) ;
    if( F <> nil ) then
    begin
        Show_Error ;
    end ;
    Check_Not_UEC ;
    FS.Erase( '\TEST.DIR' ) ; { Try killing non-empty directory }
    FS.Last_Error( UEC ) ;
    Check_Not_UEC ;
    FS.Erase( '\TEST.DAT' ) ;
    FS.Last_Error( UEC ) ;
    Check_UEC ;
    S := Files.Lookup( 'C:\TESTFS.DIR\', Find^ ) ;
    if( length( S ) > 0 ) then
    begin
        Show_Error ;
    end ;
    S := Files.Lookup( 'C:\TESTFS.DIR\...\TESTFS.DIR\', Find^ ) ;
    if( length( S ) > 0 ) then
    begin
        Show_Error ;
    end ;
    S := Files.Lookup( 'C:\TESTFS.DIR\*\SUB.DAT', Find^ ) ;
    if( length( S ) = 0 ) then
    begin
        Show_Error ;
    end ;
    S := Files.Lookup( 'C:\TESTFS.DIR\*\A\SUB.DAT', Find^ ) ;
    if( length( S ) > 0 ) then
    begin
        Show_Error ;
    end ;
    S := Files.Lookup( 'C:\TESTFS.DIR\Z*\SUB.DAT', Find^ ) ;
    if( length( S ) > 0 ) then
    begin
        Show_Error ;
    end ;
    S := Files.Lookup( 'C:\TESTFS.DIR\TEST.DIR\Z*\*.*', Find^ ) ;
    if( length( S ) > 0 ) then
    begin
        Show_Error ;
    end ;
    S := Files.Lookup( 'C:\TESTFS.DIR\TEST.*\SUB.DAT', Find^ ) ;
    if( length( S ) = 0 ) then
    begin
        Show_Error ;
    end ;
    S := Files.Lookup( '', Find^ ) ;
    if( length( S ) > 0 ) then
    begin
        Show_Error ;
    end ;
    FS.Rename( '\TEST.DIR\SUB.DAT', 'SUBFILE.DAT' ) ;
    FS.Last_Error( UEC ) ;
    Check_UEC ;
    F := Open( 'C:\TESTFS.DIR\TEST.DIR\SUB.DAT', FSOM_OUTPUT, UEC ) ;
    Check_UEC ;
    FS.Rename( '\SUBFILE1.DAT', '\TEST.DIR' ) ;
    FS.Last_Error( UEC ) ;
    if( UEC.Code = 0 ) then
    begin
        Show_Error ;
    end ;
    dispose( F, Done ) ;
    FS.Rename( '\TEST.DIR\SUB.DAT', '\TEST.DIR\SUBFILE.DAT' ) ;
    FS.Last_Error( UEC ) ;
    Check_UEC ;
    FS.Rename( '\TEST.DIR\SUBFILE.DAT', '\SUBFILE1.DAT' ) ;
    FS.Last_Error( UEC ) ;
    Check_UEC ;
    FS.Rename( '\SUBFILE1.DAT', '\NODIR.DIR\SUBFILE1.DAT' ) ; { Rename to invalid location }
    FS.Last_Error( UEC ) ;
    Check_Not_UEC ;
    FS.Rename( '\S.DAT', '\NODIR.DIR\SUBFILE1.DAT' ) ; { Rename non-existent file }
    FS.Last_Error( UEC ) ;
    Check_Not_UEC ;
    FS.Rename( '\SUBFILE1.DAT', '\TEST.DIR' ) ; { Rename to existent file }
    FS.Last_Error( UEC ) ;
    Check_Not_UEC ;
    if( FS.MkDir( '\DIR.DIR' ) <> 0 ) then
    begin
        Show_Error ;
    end ;
    if( FS.MkDir( '\DIR.DIR' ) = 0 ) then { Try creating one that already exists }
    begin
        Show_Error ;
    end ;
    if( FS.MkDir( '\NONEXISTENT.FILE\DIR.DIR' ) = 0 ) then { Use invalid path }
    begin
        Show_Error ;
    end ;
    if( FS.DeviceSize( 'C:\TESTFS.DIR' ) = -1 ) then
    begin
        Show_Error ;
    end ;
    if( FS.DeviceFree( 'C:\TESTFS.DIR' ) = -1 ) then
    begin
        Show_Error ;
    end ;
    if( FS.DeviceSize( '' ) = -1 ) then
    begin
        Show_Error ;
    end ;
    if( FS.DeviceFree( '' ) = -1 ) then
    begin
        Show_Error ;
    end ;
    if( FS.DeviceSize( 'C:' ) <> -1 ) then
    begin
        Show_Error ;
    end ;
    if( FS.DeviceFree( 'C:' ) <> -1 ) then
    begin
        Show_Error ;
    end ;

    { Test invalid filename handling... }
    FS.Erase( '*?' ) ;
    FS.Last_Error( UEC ) ;
    Check_Not_UEC ;
    FS.Rename( '*?', '?*' ) ;
    FS.Last_Error( UEC ) ;
    Check_Not_UEC ;
    F := Open( '', FSOM_OUTPUT, UEC ) ;
    if( F <> nil ) then
    begin
        Show_Error ;
    end ;
    Check_Not_UEC ;

    { Test file system header get/set... }
    FS.Get_Header( 'C:\TESTFS.DIR', Header, UEC ) ;
    Check_UEC ;
    FS.Set_Header( 'C:\TESTFS.DIR', Header, UEC ) ;
    Check_UEC ;

    { Test lookups... }
    if( FS.MkDir( '\EMPTY.DIR' ) <> 0 ) then
    begin
        Show_Error ;
    end ;
    S := Files.Lookup( 'C:\TESTFS.DIR\...\*.*', Find^ ) ;
    while( length( S ) > 0 ) do
    begin
        S := Files.Lookup( '', Find^ ) ;
    end ;
    F := Open( 'C:\TESTFS.DIR\TEST.TXT', FSOM_INPUT or FSOM_TEXT, UEC ) ;
    Check_UEC ;
    FS.Set_File_Attributes( '\SUB.DAT', FSFF_VolumeID ) ; { Non-existent file }
    FS.Last_Error( UEC ) ;
    Check_Not_UEC ;
    dispose( F, Done ) ;
    F := Open( 'C:\TESTFS.DIR\A.LABEL', FSOM_OUTPUT, UEC ) ;
    Check_UEC ;
    F^.Set_Attributes( FSFF_VolumeID ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    dispose( F, Done ) ;
    S := Files.Lookup( 'C:\TESTFS.DIR\...\*.*', Find^ ) ;
    while( length( S ) > 0 ) do
    begin
        S := Files.Lookup( '', Find^ ) ;
    end ;
{$IFNDEF DELPHI}
    F :=
{$ENDIF}
        Open( 'C:\TESTFS.DIR\TESTT.TXT', FSOM_INPUT, UEC ) ;
    Check_Not_UEC ;
    F := Open( 'C:\TESTFS.DIR\TEST.TXT', FSOM_INPUT, UEC ) ;
    Check_UEC ;
    S := 'A' ;
    F^.Blockread( S[ 1 ], 1, Dummy ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    F^.Blockwrite( S[ 1 ], 1, Dummy ) ;
    UEC := F^.IO_Error ;
    Check_UEC ;
    S := Files.Lookup( 'C:\TESTFS.DIR\*\*\*.*', Find^ ) ;
    if( length( S ) > 0 ) then
    begin
        Show_Error ;
    end ;
    dispose( F, Done ) ;

    { Final tests/close... }
    Close_File_System( 'C:\TESTFS.DIR' ) ;

    FS := Open_File_System( 'C:\TESTFS.DIR', 0, UEC ) ;
    if( FS.DeviceFree( 'C:\TESTFS.DIR' ) = -1 ) then
    begin
        Show_Error ;
    end ;
    FS.Verify( 'C:\TESTFS.DIR', UEC ) ;
    Check_UEC ;
    Close_File_System( 'C:\TESTFS.DIR' ) ;

    Uninitialize_Find_Record( Find^ ) ;
    freemem( Find ) ;
    Cleanup ;
end ; { Test_FS }
{$ENDIF}

{$IFDEF DELPHI}
initialization
{$ELSE}
begin
{$ENDIF}
    Unit_Initialization ;
{$IFDEF DELPHI}

finalization
    Unit_Deallocation ;
{$ENDIF}
end.
